//{{NO_DEPENDENCIES}}
// Microsoft Visual C++ generated include file.
// Used by daedalus.rc
//
#define icon                            101
#define icon2                           102
#define icon3                           103
#define icon4                           104
#define menu                            111
#define accelerator                     121
#define dlgAbort                        201
#define dlgFile                         202
#define dlgAbout                        203
#define dlgCommand                      204
#define dlgEvent                        205
#define dlgDisplay                      206
#define dlgColor                        207
#define dlgRandom                       208
#define dlgDot                          209
#define dlgInside                       210
#define dlgSize                         211
#define dlgZoom                         212
#define dlgCubeFlip                     213
#define dlgBright                       214
#define dlgColorReplace                 215
#define dlgSizeMaze                     216
#define dlgBias                         217
#define dlgMaze                         218
#define dlgCreate                       219
#define dlgLabyrinth                    220
#define dlgDraw                         221
#define dlgDraw2                        222
#define dlgString                       223
#define dlgString2                      224
#define dlgString3                      225
#define cmd2nd                          1001
#define cmd2ndAnd                       1002
#define cmd2ndBlock                     1003
#define cmd2ndBox                       1004
#define cmd2ndCircle                    1005
#define cmd2ndDisk                      1006
#define cmd2ndGet                       1007
#define cmd2ndLine                      1008
#define cmd2ndOr                        1009
#define cmd2ndPut                       1010
#define cmd2ndXor                       1011
#define cmd3D                           1012
#define cmdAbout                        1013
#define cmdAccentBoundary               1014
#define cmdAddDetachment                1015
#define cmdAddIn                        1016
#define cmdAddIsolation                 1017
#define cmdAddOut                       1018
#define cmdAddX                         1019
#define cmdAddX2                        1020
#define cmdAddY                         1021
#define cmdAddY2                        1022
#define cmdAllClear                     1023
#define cmdAllInvert                    1024
#define cmdAllRandom                    1025
#define cmdAllSet                       1026
#define cmdAnalyze                      1027
#define cmdAnalyze2                     1028
#define cmdBias                         1029
#define cmdCellDec1                     1030
#define cmdCellDec10                    1031
#define cmdCellInc1                     1032
#define cmdCellInc10                    1033
#define cmdClarify3D                    1034
#define cmdClarify4D                    1035
#define cmdClarifyWeave                 1036
#define cmdCollapse                     1037
#define cmdColmap                       1038
#define cmdColmapAnd                    1039
#define cmdColmapAntialias2             1040
#define cmdColmapAntialias3             1041
#define cmdColmapAntialias4             1042
#define cmdColmapAntialias5             1043
#define cmdColmapAntialias6             1044
#define cmdColmapAntialias7             1045
#define cmdColmapAntialias8             1046
#define cmdColmapBlend                  1047
#define cmdColmapBlend2                 1048
#define cmdColmapBrightness             1049
#define cmdColmapDel                    1050
#define cmdColmapDistance               1051
#define cmdColmapDistance2              1052
#define cmdColmapDotAdd                 1053
#define cmdColmapDotAnd                 1054
#define cmdColmapDotBlend               1055
#define cmdColmapDotMax                 1056
#define cmdColmapDotMin                 1057
#define cmdColmapDotMul                 1058
#define cmdColmapDotOr                  1059
#define cmdColmapDotSet                 1060
#define cmdColmapDotSub                 1061
#define cmdColmapDotXor                 1062
#define cmdColmapGet                    1063
#define cmdColmapMono                   1064
#define cmdColmapOr                     1065
#define cmdColmapPut                    1066
#define cmdColmapPutDither              1067
#define cmdColmapPutNearest             1068
#define cmdColmapReplace                1069
#define cmdColmapXor                    1070
#define cmdCommand                      1071
#define cmdCopyBitmap                   1072
#define cmdCopyPicture                  1073
#define cmdCopyText                     1074
#define cmdCorner                       1075
#define cmdCount                        1076
#define cmdCountPossible                1077
#define cmdCrackBlindAlley              1078
#define cmdCrackBlock                   1079
#define cmdCrackBoundary                1080
#define cmdCrackCell                    1081
#define cmdCrackDetach                  1082
#define cmdCrackEntrance                1083
#define cmdCrackIsolation               1084
#define cmdCrackPassage                 1085
#define cmdCrackPole                    1086
#define cmdCrackRoom                    1087
#define cmdCrackTube                    1088
#define cmdCreate3D                     1089
#define cmdCreate4D                     1090
#define cmdCreateAldous                 1091
#define cmdCreateArrow                  1092
#define cmdCreateBinary                 1093
#define cmdCreateBraid                  1094
#define cmdCreateCavern                 1095
#define cmdCreateCrack                  1096
#define cmdCreateDelta                  1097
#define cmdCreateDiagonal               1098
#define cmdCreateDivision               1099
#define cmdCreateEller                  1100
#define cmdCreateForest                 1101
#define cmdCreateFractal                1102
#define cmdCreateFractal2               1103
#define cmdCreateGamma                  1104
#define cmdCreateHilbert                1105
#define cmdCreateHyper                  1106
#define cmdCreateInfinite0              1107
#define cmdCreateInfinite1              1108
#define cmdCreateInfinite2              1109
#define cmdCreateInfinite3              1110
#define cmdCreateKruskal                1111
#define cmdCreateLabyrinth              1112
#define cmdCreatePatchOverview          1113
#define cmdCreatePerfect                1114
#define cmdCreatePlanair                1115
#define cmdCreatePrim                   1116
#define cmdCreatePrim2                  1117
#define cmdCreateRandom                 1118
#define cmdCreateRecursive              1119
#define cmdCreateSegment                1120
#define cmdCreateSidewinder             1121
#define cmdCreateSigma                  1122
#define cmdCreateSpiral                 1123
#define cmdCreateTheta                  1124
#define cmdCreateTilt                   1125
#define cmdCreateTree                   1126
#define cmdCreateUnicursal              1127
#define cmdCreateUpsilon                1128
#define cmdCreateWeave                  1129
#define cmdCreateWilson                 1130
#define cmdCreateWireOverview           1131
#define cmdCreateZeta                   1132
#define cmdCubemapFlip                  1133
#define cmdDeletePole                   1134
#define cmdDelX                         1135
#define cmdDelX2                        1136
#define cmdDelY                         1137
#define cmdDelY2                        1138
#define cmdDlgColor                     1139
#define cmdDlgCreate                    1140
#define cmdDlgDisplay                   1141
#define cmdDlgDot                       1142
#define cmdDlgDraw                      1143
#define cmdDlgDraw2                     1144
#define cmdDlgEvent                     1145
#define cmdDlgFile                      1146
#define cmdDlgInside                    1147
#define cmdDlgLabyrinth                 1148
#define cmdDlgMaze                      1149
#define cmdDlgRandom                    1150
#define cmdDlgSize                      1151
#define cmdDlgSizeMaze                  1152
#define cmdDlgZoom                      1153
#define cmdDotBig                       1154
#define cmdDotBoth                      1155
#define cmdDotDrag                      1156
#define cmdDotOn                        1157
#define cmdDotSet                       1158
#define cmdDotShow                      1159
#define cmdDotTwo                       1160
#define cmdDotWall                      1161
#define cmdDotZap                       1162
#define cmdDotZapTrans                  1163
#define cmdDotZapUntrans                1164
#define cmdDrawDel                      1165
#define cmdDrawGet                      1166
#define cmdDrawPut                      1167
#define cmdDrawSwap                     1168
#define cmdEraseCorner                  1169
#define cmdExit                         1170
#define cmdExpandSet                    1171
#define cmdFlipX                        1172
#define cmdFlipY                        1173
#define cmdFlipZ                        1174
#define cmdFloodDot                     1175
#define cmdFloodDot2                    1176
#define cmdFloodPassage                 1177
#define cmdFloodWallL                   1178
#define cmdFloodWallR                   1179
#define cmdGo2nd                        1180
#define cmdGoCorner1                    1181
#define cmdGoCorner3                    1182
#define cmdGoCorner7                    1183
#define cmdGoCorner9                    1184
#define cmdGoIn                         1185
#define cmdGoInY                        1186
#define cmdGoMiddle                     1187
#define cmdGoOut                        1188
#define cmdGoOutY                       1189
#define cmdGoRandom                     1190
#define cmdGroundVariable               1191
#define cmdInsideMap                    1192
#define cmdInsideView                   1193
#define cmdJump1                        1194
#define cmdJump2                        1195
#define cmdJump3                        1196
#define cmdJump4                        1197
#define cmdJump6                        1198
#define cmdJump7                        1199
#define cmdJump8                        1200
#define cmdJump9                        1201
#define cmdLabyrinthChartres            1202
#define cmdLabyrinthChartresF           1203
#define cmdLabyrinthChartresR           1204
#define cmdLabyrinthClassical           1205
#define cmdLabyrinthClassicalF          1206
#define cmdLabyrinthCretan              1207
#define cmdLabyrinthCustom              1208
#define cmdLabyrinthMan                 1209
#define cmdLife                         1210
#define cmdMacro01                      1211
#define cmdMacro02                      1212
#define cmdMacro03                      1213
#define cmdMacro04                      1214
#define cmdMacro05                      1215
#define cmdMacro06                      1216
#define cmdMacro07                      1217
#define cmdMacro08                      1218
#define cmdMacro09                      1219
#define cmdMacro10                      1220
#define cmdMacro11                      1221
#define cmdMacro12                      1222
#define cmdMacro13                      1223
#define cmdMacro14                      1224
#define cmdMacro15                      1225
#define cmdMacro16                      1226
#define cmdMacro17                      1227
#define cmdMacro18                      1228
#define cmdMacro19                      1229
#define cmdMacro20                      1230
#define cmdMacro21                      1231
#define cmdMacro22                      1232
#define cmdMacro23                      1233
#define cmdMacro24                      1234
#define cmdMacro25                      1235
#define cmdMacro26                      1236
#define cmdMacro27                      1237
#define cmdMacro28                      1238
#define cmdMacro29                      1239
#define cmdMacro30                      1240
#define cmdMacro31                      1241
#define cmdMacro32                      1242
#define cmdMacro33                      1243
#define cmdMacro34                      1244
#define cmdMacro35                      1245
#define cmdMacro36                      1246
#define cmdMacro37                      1247
#define cmdMacro38                      1248
#define cmdMacro39                      1249
#define cmdMacro40                      1250
#define cmdMacro41                      1251
#define cmdMacro42                      1252
#define cmdMacro43                      1253
#define cmdMacro44                      1254
#define cmdMacro45                      1255
#define cmdMacro46                      1256
#define cmdMacro47                      1257
#define cmdMacro48                      1258
#define cmdMark                         1259
#define cmdMarkBlindAlley               1260
#define cmdMarkBlock                    1261
#define cmdMarkBlock2                   1262
#define cmdMarkCollision                1263
#define cmdMarkCuldsac                  1264
#define cmdMessage                      1265
#define cmdMove1                        1266
#define cmdMove2                        1267
#define cmdMove3                        1268
#define cmdMove4                        1269
#define cmdMove6                        1270
#define cmdMove7                        1271
#define cmdMove8                        1272
#define cmdMove9                        1273
#define cmdMoveAround                   1274
#define cmdMoveBack                     1275
#define cmdMoveD                        1276
#define cmdMoveEast                     1277
#define cmdMoveForward                  1278
#define cmdMoveLeft                     1279
#define cmdMoveNorth                    1280
#define cmdMoveRandom                   1281
#define cmdMoveRight                    1282
#define cmdMoveSouth                    1283
#define cmdMoveU                        1284
#define cmdMoveWest                     1285
#define cmdNormalize                    1286
#define cmdNormalize2                   1287
#define cmdOpen                         1288
#define cmdOpen3D                       1289
#define cmdOpenBitmap                   1290
#define cmdOpenColmapTarga              1291
#define cmdOpenPatch                    1292
#define cmdOpenScript                   1293
#define cmdOpenText                     1294
#define cmdOpenWire                     1295
#define cmdOpenXbm                      1296
#define cmdPartBraid                    1297
#define cmdPartCavern                   1298
#define cmdPartCrack                    1299
#define cmdPartKruskal                  1300
#define cmdPartPerfect                  1301
#define cmdPartPrim                     1302
#define cmdPartRecursive                1303
#define cmdPartTree                     1304
#define cmdPartUnicursal                1305
#define cmdPartWeave                    1306
#define cmdPartZeta                     1307
#define cmdPaste                        1308
#define cmdPatchToWire                  1309
#define cmdPolishMaze                   1310
#define cmdPrint                        1311
#define cmdPrintSetup                   1312
#define cmdRandomize                    1313
#define cmdRedrawNow                    1314
#define cmdRenderAerial                 1315
#define cmdRenderOverview               1316
#define cmdRenderPatchPerspective       1317
#define cmdRenderPyramid                1318
#define cmdRenderWirePerspective        1319
#define cmdRepaintNow                   1320
#define cmdRepeat                       1321
#define cmdRoomThin                     1322
#define cmdRotateA                      1323
#define cmdRotateL                      1324
#define cmdRotateR                      1325
#define cmdSave3DN                      1326
#define cmdSave3DS                      1327
#define cmdSaveBitmap                   1328
#define cmdSaveColmapTarga              1329
#define cmdSaveDOS                      1330
#define cmdSaveDOS2                     1331
#define cmdSaveDOS3                     1332
#define cmdSavePatch                    1333
#define cmdSavePicture                  1334
#define cmdSaveText                     1335
#define cmdSaveWallCenter               1336
#define cmdSaveWallFill                 1337
#define cmdSaveWallFit                  1338
#define cmdSaveWallStretch              1339
#define cmdSaveWallTile                 1340
#define cmdSaveWire                     1341
#define cmdSaveXbmC                     1342
#define cmdSaveXbmN                     1343
#define cmdSaveXbmS                     1344
#define cmdScript01                     1345
#define cmdScript02                     1346
#define cmdScript03                     1347
#define cmdScript04                     1348
#define cmdScript05                     1349
#define cmdScript06                     1350
#define cmdScript07                     1351
#define cmdScript08                     1352
#define cmdScript09                     1353
#define cmdScript10                     1354
#define cmdScript11                     1355
#define cmdScript12                     1356
#define cmdScript13                     1357
#define cmdScript14                     1358
#define cmdScript15                     1359
#define cmdScript16                     1360
#define cmdScript17                     1361
#define cmdScript18                     1362
#define cmdScript19                     1363
#define cmdScript21                     1364
#define cmdScript22                     1365
#define cmdScript23                     1366
#define cmdScript24                     1367
#define cmdScript25                     1368
#define cmdScript26                     1369
#define cmdScript27                     1370
#define cmdScript28                     1371
#define cmdScript29                     1372
#define cmdScrollDown                   1373
#define cmdScrollEnd                    1374
#define cmdScrollHome                   1375
#define cmdScrollUp                     1376
#define cmdSetupAll                     1377
#define cmdSetupDesktop                 1378
#define cmdSetupExtension               1379
#define cmdSetupUser                    1380
#define cmdShellChange                  1381
#define cmdShellHelp                    1382
#define cmdShellLicense                 1383
#define cmdShellScript                  1384
#define cmdShellWeb                     1385
#define cmdShowBottleneck               1386
#define cmdSize01                       1387
#define cmdSize02                       1388
#define cmdSize03                       1389
#define cmdSize04                       1390
#define cmdSize05                       1391
#define cmdSize06                       1392
#define cmdSize07                       1393
#define cmdSize08                       1394
#define cmdSize09                       1395
#define cmdSize10                       1396
#define cmdSize11                       1397
#define cmdSize12                       1398
#define cmdSize13                       1399
#define cmdSize14                       1400
#define cmdSize15                       1401
#define cmdSize16                       1402
#define cmdSize17                       1403
#define cmdSize18                       1404
#define cmdSize19                       1405
#define cmdSizeWindow                   1406
#define cmdSlide                        1407
#define cmdSmooth                       1408
#define cmdSolve4D                      1409
#define cmdSolveBlindAlley              1410
#define cmdSolveBlock                   1411
#define cmdSolveChain                   1412
#define cmdSolveCollision               1413
#define cmdSolveCuldsac                 1414
#define cmdSolveFollowL                 1415
#define cmdSolveFollowR                 1416
#define cmdSolveFractal2                1417
#define cmdSolveHyper                   1418
#define cmdSolvePledgeL                 1419
#define cmdSolvePledgeR                 1420
#define cmdSolveRecursive               1421
#define cmdSolveShortest                1422
#define cmdSolveShortest2               1423
#define cmdSolveTremaux                 1424
#define cmdSolveWeave                   1425
#define cmdSpree                        1426
#define cmdSymmetric                    1427
#define cmdTempAdd                      1428
#define cmdTempAlpha                    1429
#define cmdTempAnd                      1430
#define cmdTempBlend                    1431
#define cmdTempDel                      1432
#define cmdTempGet                      1433
#define cmdTempOr                       1434
#define cmdTempPut                      1435
#define cmdTempSub                      1436
#define cmdTempSwap                     1437
#define cmdTempTessellate               1438
#define cmdTempXor                      1439
#define cmdTextureBackground            1440
#define cmdTextureBlock1                1441
#define cmdTextureBlock2                1442
#define cmdTextureBlock3                1443
#define cmdTextureCeiling               1444
#define cmdTextureDel                   1445
#define cmdTextureGround1               1446
#define cmdTextureGround2               1447
#define cmdTextureGround3               1448
#define cmdTextureWall1                 1449
#define cmdTextureWall2                 1450
#define cmdTextureWall3                 1451
#define cmdThicken                      1452
#define cmdThinner                      1453
#define cmdTimeGet                      1454
#define cmdTimePause                    1455
#define cmdTimeReset                    1456
#define cmdTweakEndpoint                1457
#define cmdTweakPassage                 1458
#define cmdUnmark                       1459
#define cmdUnsetup                      1460
#define cmdWallVariable                 1461
#define cmdWeave3D                      1462
#define cmdWeaveInside                  1463
#define cmdWindowBitmap                 1464
#define cmdWindowFull                   1465
#define dbCl_g                          10001
#define dbCl_p                          10002
#define dcDo_gch                        10003
#define dcDo_gcl                        10004
#define dcDo_oc                         10005
#define dcDo_sch                        10006
#define dcDo_scl                        10007
#define dcDo_tc                         10008
#define dcRc_1                          10009
#define dcRc_2                          10010
#define dcRc_3                          10011
#define dcRc_4                          10012
#define dc_Dot                          10013
#define dc_Edge                         10014
#define dc_In3D                         10015
#define dc_InCeil                       10016
#define dc_InCld                        10017
#define dc_InDirt                       10018
#define dc_InEdge                       10019
#define dc_InFog                        10020
#define dc_InMtn                        10021
#define dc_InSky                        10022
#define dc_InWall                       10023
#define dc_Map                          10024
#define dc_Off                          10025
#define dc_On                           10026
#define de1                             10027
#define de2                             10028
#define de3                             10029
#define deBi_x0                         10030
#define deBi_x1                         10031
#define deBi_y0                         10032
#define deBi_y1                         10033
#define deBr_v                          10034
#define deCl_cl                         10035
#define deCl_mt                         10036
#define deCl_n                          10037
#define deCr_cl                         10038
#define deCr_cp                         10039
#define deCr_cs                         10040
#define deCr_fa                         10041
#define deCr_fi                         10042
#define deCr_fn                         10043
#define deCr_fx                         10044
#define deCr_fy                         10045
#define deCr_j                          10046
#define deCr_l                          10047
#define deCr_o                          10048
#define deCr_o2                         10049
#define deCr_of                         10050
#define deCr_os                         10051
#define deCr_ra                         10052
#define deCr_ss                         10053
#define deCr_sw                         10054
#define deCr_t                          10055
#define deCr_ts                         10056
#define deCr_vs                         10057
#define deDo_bs                         10058
#define deDo_hs                         10059
#define deDo_s                          10060
#define deDo_vx                         10061
#define deDo_vy                         10062
#define deDo_vz                         10063
#define deDo_xs                         10064
#define deDo_ys                         10065
#define deDo_zs                         10066
#define deDr_bx                         10067
#define deDr_by                         10068
#define deDr_mx                         10069
#define deDr_my                         10070
#define deDr_mz                         10071
#define deDr_ph                         10072
#define deDr_sm                         10073
#define deDr_sr                         10074
#define deDr_th                         10075
#define deDr_vx                         10076
#define deDr_vy                         10077
#define deDr_x                          10078
#define deDr_y                          10079
#define deDr_z                          10080
#define deDs_d                          10081
#define deDs_ds                         10082
#define deDs_mc                         10083
#define deDs_ra                         10084
#define deDs_x                          10085
#define deDs_x2                         10086
#define deDs_y                          10087
#define deDs_y2                         10088
#define deDs_z                          10089
#define deDs_z2                         10090
#define deEv_c                          10091
#define deEv_cl                         10092
#define deEv_cm                         10093
#define deEv_cn                         10094
#define deEv_cp                         10095
#define deEv_cr                         10096
#define deEv_ex                         10097
#define deEv_f                          10098
#define deEv_i                          10099
#define deEv_i2                         10100
#define deEv_m                          10101
#define deEv_m2                         10102
#define deEv_me                         10103
#define deEv_mm                         10104
#define deEv_mw                         10105
#define deFs_2x                         10106
#define deFs_2y                         10107
#define deFs_3                          10108
#define deFs_ic                         10109
#define deFs_ig                         10110
#define deFs_in                         10111
#define deIn_cl                         10112
#define deIn_cp                         10113
#define deIn_fd                         10114
#define deIn_fm                         10115
#define deIn_fr                         10116
#define deIn_fu                         10117
#define deIn_lf                         10118
#define deIn_mr                         10119
#define deIn_mt                         10120
#define deIn_od                         10121
#define deIn_ox                         10122
#define deIn_oy                         10123
#define deIn_oz                         10124
#define deIn_s                          10125
#define deIn_si                         10126
#define deIn_sn                         10127
#define deIn_ss                         10128
#define deIn_sw                         10129
#define deIn_sz                         10130
#define deIn_vm                         10131
#define deIn_vr                         10132
#define deIn_vs                         10133
#define deIn_vu                         10134
#define deIn_z                          10135
#define deLa_aa                         10136
#define deLa_cc                         10137
#define deLa_cp                         10138
#define deLa_cs                         10139
#define deLa_fa                         10140
#define deLa_pc                         10141
#define deLa_pr                         10142
#define deMa_mc                         10143
#define deMa_tp                         10144
#define deMa_wp                         10145
#define deMa_wr                         10146
#define deMa_ww                         10147
#define deRa_b                          10148
#define deRa_r                          10149
#define deRa_s                          10150
#define deRc_d                          10151
#define deRc_s                          10152
#define deRe_sx                         10153
#define deRe_sy                         10154
#define deRe_td                         10155
#define deRe_ti                         10156
#define deRe_wx                         10157
#define deRe_wxo                        10158
#define deRe_wy                         10159
#define deRe_wyo                        10160
#define deRe_x                          10161
#define deRe_y                          10162
#define deSi_w3                         10163
#define deSi_x                          10164
#define deSi_x3                         10165
#define deSi_y                          10166
#define deSi_y3                         10167
#define deSi_z3                         10168
#define deSm_w3                         10169
#define deSm_x                          10170
#define deSm_x3                         10171
#define deSm_y                          10172
#define deSm_y3                         10173
#define deSm_z3                         10174
#define deZo_x                          10175
#define deZo_y                          10176
#define dr1                             10177
#define dr2                             10178
#define dr3                             10179
#define dr4                             10180
#define dr5                             10181
#define dr6                             10182
#define dr7                             10183
#define dr8                             10184
#define dr9                             10185
#define dra                             10186
#define drb                             10187
#define drc                             10188
#define drd                             10189
#define dre                             10190
#define drf                             10191
#define ds1                             10192
#define ds2                             10193
#define ds3                             10194
#define dxCo_ms                         10195
#define dxCr_c                          10196
#define dxCr_fi                         10197
#define dxCr_kp                         10198
#define dxCr_r                          10199
#define dxCr_re                         10200
#define dxCr_rf                         10201
#define dxCr_t                          10202
#define dxCr_tr                         10203
#define dxCr_tw                         10204
#define dxCr_wc                         10205
#define dxDo_e                          10206
#define dxDo_gs                         10207
#define dxDo_r                          10208
#define dxDo_s                          10209
#define dxDo_ss                         10210
#define dxDo_t                          10211
#define dxDr_ak                         10212
#define dxDr_mc                         10213
#define dxDr_ri                         10214
#define dxDs_3                          10215
#define dxDs_cm                         10216
#define dxDs_da                         10217
#define dxDs_db                         10218
#define dxDs_dc                         10219
#define dxDs_de                         10220
#define dxDs_dm                         10221
#define dxDs_dt                         10222
#define dxDs_et                         10223
#define dxDs_fp                         10224
#define dxDs_gb                         10225
#define dxDs_mc                         10226
#define dxDs_mt                         10227
#define dxDs_nc                         10228
#define dxDs_sd                         10229
#define dxDs_tw                         10230
#define dxDs_wi                         10231
#define dxFs_b                          10232
#define dxFs_e                          10233
#define dxFs_id                         10234
#define dxFs_if                         10235
#define dxFs_is                         10236
#define dxFs_l                          10237
#define dxFs_t                          10238
#define dxIn_3                          10239
#define dxIn_ca                         10240
#define dxIn_ci                         10241
#define dxIn_li                         10242
#define dxIn_mo                         10243
#define dxIn_mt                         10244
#define dxIn_nw                         10245
#define dxIn_rb                         10246
#define dxIn_se                         10247
#define dxIn_si                         10248
#define dxIn_sm                         10249
#define dxIn_tm                         10250
#define dxLa_ca                         10251
#define dxLa_re                         10252
#define dxLa_sc                         10253
#define dxMa_ad                         10254
#define dxMa_cp                         10255
#define dxMa_cs                         10256
#define dxMa_de                         10257
#define dxMa_ep                         10258
#define dxMa_pm                         10259
#define dxMa_rp                         10260
#define dxMa_te                         10261
#define dxRc_x                          10262
#define dxRe_ar                         10263
#define dxRe_c                          10264
#define dxRe_d                          10265
#define dxRe_ec                         10266
#define dxRe_hg                         10267
#define dxRe_pt                         10268
#define dxRe_sm                         10269
#define dxRe_sp                         10270
#define dxRe_td                         10271
#define dxRe_u                          10272
#define dxRe_wc                         10273
#define dxRe_wm                         10274
#define dxRe_ws                         10275
#define dxRe_wt                         10276
#define dxSi_c                          10277
#define dxSm_c                          10278
#define IDC_STATIC                      -1

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE        301
#define _APS_NEXT_COMMAND_VALUE         1466
#define _APS_NEXT_CONTROL_VALUE         10279
#define _APS_NEXT_SYMED_VALUE           301
#endif
#endif
