/* Generated by Cython 0.29.10 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "daedalus/pdconfig.h",
            "daedalus/wrapper.h"
        ],
        "extra_compile_args": [
            "-std=c++11",
            "-Wno-error=format-security"
        ],
        "include_dirs": [
            "daedalus",
            "daedalus/src"
        ],
        "language": "c++",
        "name": "daedalus._maze",
        "sources": [
            "daedalus/_maze.pyx",
            "daedalus/wrapper.cpp",
            "daedalus/src/daedalus.cpp",
            "daedalus/src/maze.cpp",
            "daedalus/src/solve.cpp",
            "daedalus/src/util.cpp",
            "daedalus/src/graphics.cpp",
            "daedalus/src/color.cpp",
            "daedalus/src/create.cpp",
            "daedalus/src/create2.cpp",
            "daedalus/src/create3.cpp",
            "daedalus/src/labyrnth.cpp",
            "daedalus/src/threed.cpp",
            "daedalus/src/draw.cpp",
            "daedalus/src/inside.cpp",
            "daedalus/src/command.cpp",
            "daedalus/src/solids.cpp",
            "daedalus/src/draw2.cpp"
        ]
    },
    "module_name": "daedalus._maze"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_10"
#define CYTHON_HEX_VERSION 0x001D0AF0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX < 0x030800A4
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__daedalus___maze
#define __PYX_HAVE_API__daedalus___maze
/* Early includes */
#include "pdconfig.h"
#include "wrapper.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "daedalus/_maze.pyx",
  "stringsource",
};

/*--- Type declarations ---*/
struct __pyx_obj_8daedalus_5_maze_Maze;
struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__;
struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count;
struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr;

/* "daedalus/_maze.pyx":234
 *     pass
 * 
 * cdef class Maze(object):             # <<<<<<<<<<<<<<
 *     '''
 *     A monochrome daedalus Maze object.
 */
struct __pyx_obj_8daedalus_5_maze_Maze {
  PyObject_HEAD
  CMaz *_maze;
  int _xEntrance;
  int _yEntrance;
  int _xExit;
  int _yExit;
};


/* "daedalus/_maze.pyx":298
 *         return False
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         i = 0
 *         try:
 */
struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__ {
  PyObject_HEAD
  PyObject *__pyx_v_i;
  struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self;
  PyObject *__pyx_v_v;
  PyObject *__pyx_t_0;
  PyObject *__pyx_t_1;
  PyObject *__pyx_t_2;
};


/* "daedalus/_maze.pyx":329
 *         raise ValueError
 * 
 *     def count(self, value):             # <<<<<<<<<<<<<<
 *         '''
 *         S.count(value) -> integer -- return number of occurrences of
 */
struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count {
  PyObject_HEAD
  struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self;
  PyObject *__pyx_v_value;
};


/* "daedalus/_maze.pyx":338
 *         :rtype: int
 *         '''
 *         return sum(1 for v in self if v == value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr {
  PyObject_HEAD
  struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count *__pyx_outer_scope;
  PyObject *__pyx_v_v;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* IncludeStringH.proto */
#include <string.h>

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_KV(KV value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned long __Pyx_PyInt_As_unsigned_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE KV __Pyx_PyInt_As_KV(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libcpp' */

/* Module declarations from 'daedalus._maze' */
static PyTypeObject *__pyx_ptype_8daedalus_5_maze_Maze = 0;
static PyTypeObject *__pyx_ptype_8daedalus_5_maze___pyx_scope_struct____iter__ = 0;
static PyTypeObject *__pyx_ptype_8daedalus_5_maze___pyx_scope_struct_1_count = 0;
static PyTypeObject *__pyx_ptype_8daedalus_5_maze___pyx_scope_struct_2_genexpr = 0;
#define __Pyx_MODULE_NAME "daedalus._maze"
extern int __pyx_module_is_main_daedalus___maze;
int __pyx_module_is_main_daedalus___maze = 0;

/* Implementation of 'daedalus._maze' */
static PyObject *__pyx_builtin_staticmethod;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_map;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_sum;
static PyObject *__pyx_builtin_TypeError;
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k_c2[] = "c2";
static const char __pyx_k_c3[] = "c3";
static const char __pyx_k_rv[] = "rv";
static const char __pyx_k__30[] = "";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_map[] = "map";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_Maze[] = "Maze";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_iter[] = "__iter__";
static const char __pyx_k_kvOn[] = "kvOn";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_seed[] = "seed";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_fWall[] = "fWall";
static const char __pyx_k_kvOff[] = "kvOff";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_width[] = "width";
static const char __pyx_k_Maze_x[] = "<Maze: {}x{}>";
static const char __pyx_k_compat[] = "compat";
static const char __pyx_k_fClear[] = "fClear";
static const char __pyx_k_fRiver[] = "fRiver";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_height[] = "height";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_cSpiral[] = "cSpiral";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_nRndRun[] = "nRndRun";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_fTextTab[] = "fTextTab";
static const char __pyx_k_filename[] = "filename";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_nForsAdd[] = "nForsAdd";
static const char __pyx_k_nRndBias[] = "nRndBias";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_MazeError[] = "MazeError";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_fLineChar[] = "fLineChar";
static const char __pyx_k_fTextClip[] = "fTextClip";
static const char __pyx_k_fTreeWall[] = "fTreeWall";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_nForsInit[] = "nForsInit";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_cRandomAdd[] = "cRandomAdd";
static const char __pyx_k_nTreeRiver[] = "nTreeRiver";
static const char __pyx_k_COLOR_BLACK[] = "COLOR_BLACK";
static const char __pyx_k_COLOR_WHITE[] = "COLOR_WHITE";
static const char __pyx_k_Maze___iter[] = "Maze.__iter__";
static const char __pyx_k_cSpiralWall[] = "cSpiralWall";
static const char __pyx_k_fKruskalPic[] = "fKruskalPic";
static const char __pyx_k_fTreeRandom[] = "fTreeRandom";
static const char __pyx_k_init_random[] = "init_random";
static const char __pyx_k_fTiltDiamond[] = "fTiltDiamond";
static const char __pyx_k_nEntrancePos[] = "nEntrancePos";
static const char __pyx_k_staticmethod[] = "staticmethod";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_daedalus__maze[] = "daedalus._maze";
static const char __pyx_k_negative_width[] = "negative width";
static const char __pyx_k_ENTRANCE_CORNER[] = "ENTRANCE_CORNER";
static const char __pyx_k_ENTRANCE_MIDDLE[] = "ENTRANCE_MIDDLE";
static const char __pyx_k_ENTRANCE_RANDOM[] = "ENTRANCE_RANDOM";
static const char __pyx_k_negative_height[] = "negative height";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_py3_str_to_bytes[] = "py3_str_to_bytes";
static const char __pyx_k_ENTRANCE_BALANCED[] = "ENTRANCE_BALANCED";
static const char __pyx_k_No_filename_given[] = "No filename given.";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_daedalus__maze_pyx[] = "daedalus/_maze.pyx";
static const char __pyx_k_handle_save_retval[] = "_handle_save_retval";
static const char __pyx_k_store_back_globals[] = "_store_back_globals";
static const char __pyx_k_Could_not_open_file[] = "Could not open file.";
static const char __pyx_k_Unknown_save_command[] = "Unknown save command.";
static const char __pyx_k_count_locals_genexpr[] = "count.<locals>.genexpr";
static const char __pyx_k_Could_not_resize_the_Maze[] = "Could not resize the Maze.";
static const char __pyx_k_x_coordinate_out_of_range[] = "x coordinate out of range";
static const char __pyx_k_y_coordinate_out_of_range[] = "y coordinate out of range";
static const char __pyx_k_Could_not_create_Prim_Maze[] = "Could not create Prim Maze.";
static const char __pyx_k_Could_not_create_Tree_Maze[] = "Could not create Tree Maze.";
static const char __pyx_k_Could_not_create_Braid_Maze[] = "Could not create Braid Maze.";
static const char __pyx_k_Could_not_create_Eller_Maze[] = "Could not create Eller Maze.";
static const char __pyx_k_Could_not_create_Prim2_Maze[] = "Could not create Prim2 Maze.";
static const char __pyx_k_Could_not_create_Binary_Maze[] = "Could not create Binary Maze.";
static const char __pyx_k_Could_not_create_Forest_Maze[] = "Could not create Forest Maze.";
static const char __pyx_k_Could_not_create_Spiral_Maze[] = "Could not create Spiral Maze.";
static const char __pyx_k_Could_not_create_Wilson_Maze[] = "Could not create Wilson Maze.";
static const char __pyx_k_Could_not_create_Kruskal_Maze[] = "Could not create Kruskal Maze.";
static const char __pyx_k_Could_not_create_Perfect_Maze[] = "Could not create Perfect Maze.";
static const char __pyx_k_Maze_not_properly_constructed[] = "Maze not properly constructed.";
static const char __pyx_k_Could_not_create_Diagonal_Maze[] = "Could not create Diagonal Maze.";
static const char __pyx_k_Could_not_create_Division_Maze[] = "Could not create Division Maze.";
static const char __pyx_k_Could_not_create_Perfect2_Maze[] = "Could not create Perfect2 Maze.";
static const char __pyx_k_Could_not_construct_Maze_object[] = "Could not construct Maze object.";
static const char __pyx_k_Could_not_create_BraidTilt_Maze[] = "Could not create BraidTilt Maze.";
static const char __pyx_k_Could_not_create_Recursive_Maze[] = "Could not create Recursive Maze.";
static const char __pyx_k_Could_not_create_Unicursal_Maze[] = "Could not create Unicursal Maze.";
static const char __pyx_k_Wrapper_for_the_daedalus_monoch[] = "\nWrapper for the daedalus monochrome maze functionality.\n\nmaze = Maze(63, 63)\nmaze.create_perfect()\nmaze.resize(31, 61)\nmaze.create_perfect()\nmaze.save_bitmap('test.bmp')\nmaze.save_text('test.txt')\n";
static const char __pyx_k_An_error_while_performing_an_act[] = "An error while performing an action on a Maze object.";
static const char __pyx_k_Could_not_create_AldousBroder_Ma[] = "Could not create AldousBroder Maze.";
static const char __pyx_k_Could_not_create_BraidEller_Maze[] = "Could not create BraidEller Maze.";
static const char __pyx_k_Could_not_create_Sidewinder_Maze[] = "Could not create Sidewinder Maze.";
static const char __pyx_k_Unknown_return_value_from_Maze_s[] = "Unknown return value from Maze save function.";
static const char __pyx_k_key_must_be_an_integer_or_a_tupl[] = "key must be an integer or a tuple of two integers";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static PyObject *__pyx_kp_s_An_error_while_performing_an_act;
static PyObject *__pyx_n_s_COLOR_BLACK;
static PyObject *__pyx_n_s_COLOR_WHITE;
static PyObject *__pyx_kp_s_Could_not_construct_Maze_object;
static PyObject *__pyx_kp_s_Could_not_create_AldousBroder_Ma;
static PyObject *__pyx_kp_s_Could_not_create_Binary_Maze;
static PyObject *__pyx_kp_s_Could_not_create_BraidEller_Maze;
static PyObject *__pyx_kp_s_Could_not_create_BraidTilt_Maze;
static PyObject *__pyx_kp_s_Could_not_create_Braid_Maze;
static PyObject *__pyx_kp_s_Could_not_create_Diagonal_Maze;
static PyObject *__pyx_kp_s_Could_not_create_Division_Maze;
static PyObject *__pyx_kp_s_Could_not_create_Eller_Maze;
static PyObject *__pyx_kp_s_Could_not_create_Forest_Maze;
static PyObject *__pyx_kp_s_Could_not_create_Kruskal_Maze;
static PyObject *__pyx_kp_s_Could_not_create_Perfect2_Maze;
static PyObject *__pyx_kp_s_Could_not_create_Perfect_Maze;
static PyObject *__pyx_kp_s_Could_not_create_Prim2_Maze;
static PyObject *__pyx_kp_s_Could_not_create_Prim_Maze;
static PyObject *__pyx_kp_s_Could_not_create_Recursive_Maze;
static PyObject *__pyx_kp_s_Could_not_create_Sidewinder_Maze;
static PyObject *__pyx_kp_s_Could_not_create_Spiral_Maze;
static PyObject *__pyx_kp_s_Could_not_create_Tree_Maze;
static PyObject *__pyx_kp_s_Could_not_create_Unicursal_Maze;
static PyObject *__pyx_kp_s_Could_not_create_Wilson_Maze;
static PyObject *__pyx_kp_s_Could_not_open_file;
static PyObject *__pyx_kp_s_Could_not_resize_the_Maze;
static PyObject *__pyx_n_s_ENTRANCE_BALANCED;
static PyObject *__pyx_n_s_ENTRANCE_CORNER;
static PyObject *__pyx_n_s_ENTRANCE_MIDDLE;
static PyObject *__pyx_n_s_ENTRANCE_RANDOM;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_n_s_Maze;
static PyObject *__pyx_n_s_MazeError;
static PyObject *__pyx_n_s_Maze___iter;
static PyObject *__pyx_kp_s_Maze_not_properly_constructed;
static PyObject *__pyx_kp_s_Maze_x;
static PyObject *__pyx_kp_s_No_filename_given;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unknown_return_value_from_Maze_s;
static PyObject *__pyx_kp_s_Unknown_save_command;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s__30;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_c2;
static PyObject *__pyx_n_s_c3;
static PyObject *__pyx_n_s_cRandomAdd;
static PyObject *__pyx_n_s_cSpiral;
static PyObject *__pyx_n_s_cSpiralWall;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_compat;
static PyObject *__pyx_n_s_count_locals_genexpr;
static PyObject *__pyx_n_s_daedalus__maze;
static PyObject *__pyx_kp_s_daedalus__maze_pyx;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_fClear;
static PyObject *__pyx_n_s_fKruskalPic;
static PyObject *__pyx_n_s_fLineChar;
static PyObject *__pyx_n_s_fRiver;
static PyObject *__pyx_n_s_fTextClip;
static PyObject *__pyx_n_s_fTextTab;
static PyObject *__pyx_n_s_fTiltDiamond;
static PyObject *__pyx_n_s_fTreeRandom;
static PyObject *__pyx_n_s_fTreeWall;
static PyObject *__pyx_n_s_fWall;
static PyObject *__pyx_n_s_filename;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_handle_save_retval;
static PyObject *__pyx_n_s_height;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_init_random;
static PyObject *__pyx_n_s_iter;
static PyObject *__pyx_kp_s_key_must_be_an_integer_or_a_tupl;
static PyObject *__pyx_n_s_kvOff;
static PyObject *__pyx_n_s_kvOn;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_map;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_nEntrancePos;
static PyObject *__pyx_n_s_nForsAdd;
static PyObject *__pyx_n_s_nForsInit;
static PyObject *__pyx_n_s_nRndBias;
static PyObject *__pyx_n_s_nRndRun;
static PyObject *__pyx_n_s_nTreeRiver;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_kp_s_negative_height;
static PyObject *__pyx_kp_s_negative_width;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_py3_str_to_bytes;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_rv;
static PyObject *__pyx_n_s_seed;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_staticmethod;
static PyObject *__pyx_n_s_store_back_globals;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_width;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_kp_s_x_coordinate_out_of_range;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_kp_s_y_coordinate_out_of_range;
static PyObject *__pyx_pf_8daedalus_5_maze_init_random(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_seed); /* proto */
static int __pyx_pf_8daedalus_5_maze_4Maze___cinit__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_width, PyObject *__pyx_v_height); /* proto */
static void __pyx_pf_8daedalus_5_maze_4Maze_2__dealloc__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_8daedalus_5_maze_4Maze_4__len__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_6__getitem__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static int __pyx_pf_8daedalus_5_maze_4Maze_8__contains__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_10__iter__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_13__str__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_15__repr__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_17index(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_5count_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_19count(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_5width___get__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_6height___get__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_8entrance___get__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_4exit___get__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_21get(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_x, PyObject *__pyx_v_y); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_23create_perfect(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fRiver, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_25create_perfect2(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fRiver, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_27create_braid(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_29create_braid_tilt(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fTiltDiamond, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_31create_spiral(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_cRandomAdd, PyObject *__pyx_v_cSpiral, PyObject *__pyx_v_cSpiralWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_33create_diagonal(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_cRandomAdd, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_35create_recursive(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_37create_prim(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_39create_prim2(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fTreeRandom, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_41create_kruskal(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fClear, CYTHON_UNUSED PyObject *__pyx_v_c2, CYTHON_UNUSED PyObject *__pyx_v_c3, PyObject *__pyx_v_fKruskalPic, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_43create_tree(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fTreeRandom, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun, PyObject *__pyx_v_nTreeRiver); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_45create_forest(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fWall, PyObject *__pyx_v_fTreeRandom, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nForsAdd, PyObject *__pyx_v_nForsInit, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun, PyObject *__pyx_v_nTreeRiver); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_47create_aldous_broder(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_49create_wilson(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_51create_eller(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_53create_braid_eller(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_55create_division(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_57create_binary(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_cRandomAdd, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_59create_sidewinder(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_61create_unicursal(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_cRandomAdd, PyObject *__pyx_v_fRiver, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_63resize(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_width, PyObject *__pyx_v_height); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_65_handle_save_retval(PyObject *__pyx_v_rv); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_67_store_back_globals(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_69save_bitmap(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_filename, PyObject *__pyx_v_kvOn, PyObject *__pyx_v_kvOff); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_71save_text(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_filename, PyObject *__pyx_v_fTextClip, PyObject *__pyx_v_fLineChar, PyObject *__pyx_v_fTextTab); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_73__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_75__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_8daedalus_5_maze_Maze(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8daedalus_5_maze___pyx_scope_struct____iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8daedalus_5_maze___pyx_scope_struct_1_count(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8daedalus_5_maze___pyx_scope_struct_2_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_15;
static PyObject *__pyx_int_neg_100;
static PyObject *__pyx_k__6;
static PyObject *__pyx_k__7;
static PyObject *__pyx_k__8;
static PyObject *__pyx_k__9;
static PyObject *__pyx_k__10;
static PyObject *__pyx_k__11;
static PyObject *__pyx_k__12;
static PyObject *__pyx_k__13;
static PyObject *__pyx_k__14;
static PyObject *__pyx_k__15;
static PyObject *__pyx_k__16;
static PyObject *__pyx_k__17;
static PyObject *__pyx_k__18;
static PyObject *__pyx_k__19;
static PyObject *__pyx_k__20;
static PyObject *__pyx_k__21;
static PyObject *__pyx_k__22;
static PyObject *__pyx_k__23;
static PyObject *__pyx_k__24;
static PyObject *__pyx_k__25;
static PyObject *__pyx_k__26;
static PyObject *__pyx_k__27;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__34;
/* Late includes */

/* "daedalus/_maze.pyx":219
 * cpp_Init()
 * 
 * def init_random(seed):             # <<<<<<<<<<<<<<
 *     '''
 *     Seeds the random number generator used internally by Daedalus to
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_1init_random(PyObject *__pyx_self, PyObject *__pyx_v_seed); /*proto*/
static char __pyx_doc_8daedalus_5_maze_init_random[] = "\n    Seeds the random number generator used internally by Daedalus to\n    generate mazes.\n\n    :param int seed: The new seed value for the random number\n        generator.\n    :rtype: None\n    ";
static PyMethodDef __pyx_mdef_8daedalus_5_maze_1init_random = {"init_random", (PyCFunction)__pyx_pw_8daedalus_5_maze_1init_random, METH_O, __pyx_doc_8daedalus_5_maze_init_random};
static PyObject *__pyx_pw_8daedalus_5_maze_1init_random(PyObject *__pyx_self, PyObject *__pyx_v_seed) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("init_random (wrapper)", 0);
  __pyx_r = __pyx_pf_8daedalus_5_maze_init_random(__pyx_self, ((PyObject *)__pyx_v_seed));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_init_random(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_seed) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  unsigned long __pyx_t_1;
  __Pyx_RefNannySetupContext("init_random", 0);

  /* "daedalus/_maze.pyx":228
 *     :rtype: None
 *     '''
 *     InitRndL(seed)             # <<<<<<<<<<<<<<
 * 
 * class MazeError(Exception):
 */
  __pyx_t_1 = __Pyx_PyInt_As_unsigned_long(__pyx_v_seed); if (unlikely((__pyx_t_1 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 228, __pyx_L1_error)
  InitRndL(__pyx_t_1);

  /* "daedalus/_maze.pyx":219
 * cpp_Init()
 * 
 * def init_random(seed):             # <<<<<<<<<<<<<<
 *     '''
 *     Seeds the random number generator used internally by Daedalus to
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("daedalus._maze.init_random", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":250
 *     cdef int _yExit
 * 
 *     def __cinit__(self, width, height):             # <<<<<<<<<<<<<<
 *         '''
 *         Maze constructor.
 */

/* Python wrapper */
static int __pyx_pw_8daedalus_5_maze_4Maze_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8daedalus_5_maze_4Maze_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_width = 0;
  PyObject *__pyx_v_height = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_width,&__pyx_n_s_height,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_width)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_height)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, 1); __PYX_ERR(0, 250, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 250, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_width = values[0];
    __pyx_v_height = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 250, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze___cinit__(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_width, __pyx_v_height);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8daedalus_5_maze_4Maze___cinit__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_width, PyObject *__pyx_v_height) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "daedalus/_maze.pyx":257
 *         :param int height: Odd integer, 3 or greater
 *         '''
 *         if width < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError('negative width')
 *         if height < 0:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_width, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "daedalus/_maze.pyx":258
 *         '''
 *         if width < 0:
 *             raise ValueError('negative width')             # <<<<<<<<<<<<<<
 *         if height < 0:
 *             raise ValueError('negative height')
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 258, __pyx_L1_error)

    /* "daedalus/_maze.pyx":257
 *         :param int height: Odd integer, 3 or greater
 *         '''
 *         if width < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError('negative width')
 *         if height < 0:
 */
  }

  /* "daedalus/_maze.pyx":259
 *         if width < 0:
 *             raise ValueError('negative width')
 *         if height < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError('negative height')
 *         # if width % 2 != 1:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_height, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "daedalus/_maze.pyx":260
 *             raise ValueError('negative width')
 *         if height < 0:
 *             raise ValueError('negative height')             # <<<<<<<<<<<<<<
 *         # if width % 2 != 1:
 *         #     raise ValueError('width must be odd')
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 260, __pyx_L1_error)

    /* "daedalus/_maze.pyx":259
 *         if width < 0:
 *             raise ValueError('negative width')
 *         if height < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError('negative height')
 *         # if width % 2 != 1:
 */
  }

  /* "daedalus/_maze.pyx":265
 *         # if height % 2 != 1:
 *         #     raise ValueError('height must be odd')
 *         self._xEntrance = 0             # <<<<<<<<<<<<<<
 *         self._yEntrance = 0
 *         self._xExit = 0
 */
  __pyx_v_self->_xEntrance = 0;

  /* "daedalus/_maze.pyx":266
 *         #     raise ValueError('height must be odd')
 *         self._xEntrance = 0
 *         self._yEntrance = 0             # <<<<<<<<<<<<<<
 *         self._xExit = 0
 *         self._yExit = 0
 */
  __pyx_v_self->_yEntrance = 0;

  /* "daedalus/_maze.pyx":267
 *         self._xEntrance = 0
 *         self._yEntrance = 0
 *         self._xExit = 0             # <<<<<<<<<<<<<<
 *         self._yExit = 0
 *         self._maze = cpp_Constructor(width, height)
 */
  __pyx_v_self->_xExit = 0;

  /* "daedalus/_maze.pyx":268
 *         self._yEntrance = 0
 *         self._xExit = 0
 *         self._yExit = 0             # <<<<<<<<<<<<<<
 *         self._maze = cpp_Constructor(width, height)
 *         if self._maze is NULL:
 */
  __pyx_v_self->_yExit = 0;

  /* "daedalus/_maze.pyx":269
 *         self._xExit = 0
 *         self._yExit = 0
 *         self._maze = cpp_Constructor(width, height)             # <<<<<<<<<<<<<<
 *         if self._maze is NULL:
 *             raise MazeError('Could not construct Maze object.')
 */
  __pyx_t_3 = __Pyx_PyInt_As_unsigned_int(__pyx_v_width); if (unlikely((__pyx_t_3 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 269, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyInt_As_unsigned_int(__pyx_v_height); if (unlikely((__pyx_t_4 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 269, __pyx_L1_error)
  __pyx_v_self->_maze = cpp_Constructor(__pyx_t_3, __pyx_t_4);

  /* "daedalus/_maze.pyx":270
 *         self._yExit = 0
 *         self._maze = cpp_Constructor(width, height)
 *         if self._maze is NULL:             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not construct Maze object.')
 * 
 */
  __pyx_t_2 = ((__pyx_v_self->_maze == NULL) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "daedalus/_maze.pyx":271
 *         self._maze = cpp_Constructor(width, height)
 *         if self._maze is NULL:
 *             raise MazeError('Could not construct Maze object.')             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_kp_s_Could_not_construct_Maze_object) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_s_Could_not_construct_Maze_object);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 271, __pyx_L1_error)

    /* "daedalus/_maze.pyx":270
 *         self._yExit = 0
 *         self._maze = cpp_Constructor(width, height)
 *         if self._maze is NULL:             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not construct Maze object.')
 * 
 */
  }

  /* "daedalus/_maze.pyx":250
 *     cdef int _yExit
 * 
 *     def __cinit__(self, width, height):             # <<<<<<<<<<<<<<
 *         '''
 *         Maze constructor.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("daedalus._maze.Maze.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":273
 *             raise MazeError('Could not construct Maze object.')
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         '''Destructor.'''
 *         if self._maze is not NULL:
 */

/* Python wrapper */
static void __pyx_pw_8daedalus_5_maze_4Maze_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_8daedalus_5_maze_4Maze_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_8daedalus_5_maze_4Maze_2__dealloc__(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_8daedalus_5_maze_4Maze_2__dealloc__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "daedalus/_maze.pyx":275
 *     def __dealloc__(self):
 *         '''Destructor.'''
 *         if self._maze is not NULL:             # <<<<<<<<<<<<<<
 *             cpp_Destructor(self._maze)
 *             self._maze = NULL
 */
  __pyx_t_1 = ((__pyx_v_self->_maze != NULL) != 0);
  if (__pyx_t_1) {

    /* "daedalus/_maze.pyx":276
 *         '''Destructor.'''
 *         if self._maze is not NULL:
 *             cpp_Destructor(self._maze)             # <<<<<<<<<<<<<<
 *             self._maze = NULL
 * 
 */
    cpp_Destructor(__pyx_v_self->_maze);

    /* "daedalus/_maze.pyx":277
 *         if self._maze is not NULL:
 *             cpp_Destructor(self._maze)
 *             self._maze = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
    __pyx_v_self->_maze = NULL;

    /* "daedalus/_maze.pyx":275
 *     def __dealloc__(self):
 *         '''Destructor.'''
 *         if self._maze is not NULL:             # <<<<<<<<<<<<<<
 *             cpp_Destructor(self._maze)
 *             self._maze = NULL
 */
  }

  /* "daedalus/_maze.pyx":273
 *             raise MazeError('Could not construct Maze object.')
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         '''Destructor.'''
 *         if self._maze is not NULL:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "daedalus/_maze.pyx":279
 *             self._maze = NULL
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.height
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_8daedalus_5_maze_4Maze_5__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_8daedalus_5_maze_4Maze_5__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_4__len__(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_8daedalus_5_maze_4Maze_4__len__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "daedalus/_maze.pyx":280
 * 
 *     def __len__(self):
 *         return self.height             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, key):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_height); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyIndex_AsSsize_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "daedalus/_maze.pyx":279
 *             self._maze = NULL
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.height
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("daedalus._maze.Maze.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":282
 *         return self.height
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         if isinstance(key, int):
 *             y = key
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_7__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_7__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_6__getitem__(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_6__getitem__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_v_y = NULL;
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "daedalus/_maze.pyx":283
 * 
 *     def __getitem__(self, key):
 *         if isinstance(key, int):             # <<<<<<<<<<<<<<
 *             y = key
 *             return [int(self.get(x, y)) for x in range(self.width)]
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_key); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "daedalus/_maze.pyx":284
 *     def __getitem__(self, key):
 *         if isinstance(key, int):
 *             y = key             # <<<<<<<<<<<<<<
 *             return [int(self.get(x, y)) for x in range(self.width)]
 *         else:
 */
    __Pyx_INCREF(__pyx_v_key);
    __pyx_v_y = __pyx_v_key;

    /* "daedalus/_maze.pyx":285
 *         if isinstance(key, int):
 *             y = key
 *             return [int(self.get(x, y)) for x in range(self.width)]             # <<<<<<<<<<<<<<
 *         else:
 *             if not isinstance(key, tuple) and len(key) == 2 and list(set(map(type, key))) == [int]:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_width); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_4 = __pyx_t_5; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 285, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 285, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 285, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 285, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 285, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_7(__pyx_t_4);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 285, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_10 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_v_x, __pyx_v_y};
        __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 285, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_v_x, __pyx_v_y};
        __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 285, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else
      #endif
      {
        __pyx_t_11 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 285, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__pyx_t_9) {
          __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_9); __pyx_t_9 = NULL;
        }
        __Pyx_INCREF(__pyx_v_x);
        __Pyx_GIVEREF(__pyx_v_x);
        PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_v_x);
        __Pyx_INCREF(__pyx_v_y);
        __Pyx_GIVEREF(__pyx_v_y);
        PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_v_y);
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_11, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 285, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyNumber_Int(__pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "daedalus/_maze.pyx":283
 * 
 *     def __getitem__(self, key):
 *         if isinstance(key, int):             # <<<<<<<<<<<<<<
 *             y = key
 *             return [int(self.get(x, y)) for x in range(self.width)]
 */
  }

  /* "daedalus/_maze.pyx":287
 *             return [int(self.get(x, y)) for x in range(self.width)]
 *         else:
 *             if not isinstance(key, tuple) and len(key) == 2 and list(set(map(type, key))) == [int]:             # <<<<<<<<<<<<<<
 *                 raise ValueError('key must be an integer or a tuple of two integers')
 *             (x,y) = key
 */
  /*else*/ {
    __pyx_t_1 = PyTuple_Check(__pyx_v_key); 
    __pyx_t_12 = ((!(__pyx_t_1 != 0)) != 0);
    if (__pyx_t_12) {
    } else {
      __pyx_t_2 = __pyx_t_12;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_6 = PyObject_Length(__pyx_v_key); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 287, __pyx_L1_error)
    __pyx_t_12 = ((__pyx_t_6 == 2) != 0);
    if (__pyx_t_12) {
    } else {
      __pyx_t_2 = __pyx_t_12;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)(&PyType_Type)));
    __Pyx_GIVEREF(((PyObject *)(&PyType_Type)));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyType_Type)));
    __Pyx_INCREF(__pyx_v_key);
    __Pyx_GIVEREF(__pyx_v_key);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_key);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PySet_New(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PySequence_List(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)(&PyInt_Type)));
    __Pyx_GIVEREF(((PyObject *)(&PyInt_Type)));
    PyList_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&PyInt_Type)));
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_2 = __pyx_t_12;
    __pyx_L7_bool_binop_done:;
    if (unlikely(__pyx_t_2)) {

      /* "daedalus/_maze.pyx":288
 *         else:
 *             if not isinstance(key, tuple) and len(key) == 2 and list(set(map(type, key))) == [int]:
 *                 raise ValueError('key must be an integer or a tuple of two integers')             # <<<<<<<<<<<<<<
 *             (x,y) = key
 *             return self.get(x,y)
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(0, 288, __pyx_L1_error)

      /* "daedalus/_maze.pyx":287
 *             return [int(self.get(x, y)) for x in range(self.width)]
 *         else:
 *             if not isinstance(key, tuple) and len(key) == 2 and list(set(map(type, key))) == [int]:             # <<<<<<<<<<<<<<
 *                 raise ValueError('key must be an integer or a tuple of two integers')
 *             (x,y) = key
 */
    }

    /* "daedalus/_maze.pyx":289
 *             if not isinstance(key, tuple) and len(key) == 2 and list(set(map(type, key))) == [int]:
 *                 raise ValueError('key must be an integer or a tuple of two integers')
 *             (x,y) = key             # <<<<<<<<<<<<<<
 *             return self.get(x,y)
 * 
 */
    if ((likely(PyTuple_CheckExact(__pyx_v_key))) || (PyList_CheckExact(__pyx_v_key))) {
      PyObject* sequence = __pyx_v_key;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 289, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_3);
      #else
      __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      #endif
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_v_key); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_13 = Py_TYPE(__pyx_t_4)->tp_iternext;
      index = 0; __pyx_t_8 = __pyx_t_13(__pyx_t_4); if (unlikely(!__pyx_t_8)) goto __pyx_L10_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      index = 1; __pyx_t_3 = __pyx_t_13(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L10_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_4), 2) < 0) __PYX_ERR(0, 289, __pyx_L1_error)
      __pyx_t_13 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L11_unpacking_done;
      __pyx_L10_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_13 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 289, __pyx_L1_error)
      __pyx_L11_unpacking_done:;
    }
    __pyx_v_x = __pyx_t_8;
    __pyx_t_8 = 0;
    __pyx_v_y = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "daedalus/_maze.pyx":290
 *                 raise ValueError('key must be an integer or a tuple of two integers')
 *             (x,y) = key
 *             return self.get(x,y)             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, value):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_4 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_8)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_x, __pyx_v_y};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 290, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_x, __pyx_v_y};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 290, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 290, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_x);
      __Pyx_GIVEREF(__pyx_v_x);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_10, __pyx_v_x);
      __Pyx_INCREF(__pyx_v_y);
      __Pyx_GIVEREF(__pyx_v_y);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_10, __pyx_v_y);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 290, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "daedalus/_maze.pyx":282
 *         return self.height
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         if isinstance(key, int):
 *             y = key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("daedalus._maze.Maze.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":292
 *             return self.get(x,y)
 * 
 *     def __contains__(self, value):             # <<<<<<<<<<<<<<
 *         for v in self:
 *             if v == value:
 */

/* Python wrapper */
static int __pyx_pw_8daedalus_5_maze_4Maze_9__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_8daedalus_5_maze_4Maze_9__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_8__contains__(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8daedalus_5_maze_4Maze_8__contains__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_v = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "daedalus/_maze.pyx":293
 * 
 *     def __contains__(self, value):
 *         for v in self:             # <<<<<<<<<<<<<<
 *             if v == value:
 *                 return True
 */
  if (likely(PyList_CheckExact(((PyObject *)__pyx_v_self))) || PyTuple_CheckExact(((PyObject *)__pyx_v_self))) {
    __pyx_t_1 = ((PyObject *)__pyx_v_self); __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 293, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 293, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 293, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 293, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 293, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 293, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "daedalus/_maze.pyx":294
 *     def __contains__(self, value):
 *         for v in self:
 *             if v == value:             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
 */
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_v, __pyx_v_value, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 294, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {

      /* "daedalus/_maze.pyx":295
 *         for v in self:
 *             if v == value:
 *                 return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
      __pyx_r = 1;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "daedalus/_maze.pyx":294
 *     def __contains__(self, value):
 *         for v in self:
 *             if v == value:             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
 */
    }

    /* "daedalus/_maze.pyx":293
 * 
 *     def __contains__(self, value):
 *         for v in self:             # <<<<<<<<<<<<<<
 *             if v == value:
 *                 return True
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "daedalus/_maze.pyx":296
 *             if v == value:
 *                 return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "daedalus/_maze.pyx":292
 *             return self.get(x,y)
 * 
 *     def __contains__(self, value):             # <<<<<<<<<<<<<<
 *         for v in self:
 *             if v == value:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("daedalus._maze.Maze.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_8daedalus_5_maze_4Maze_12generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "daedalus/_maze.pyx":298
 *         return False
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         i = 0
 *         try:
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_11__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_11__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_10__iter__(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_10__iter__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self) {
  struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__ *)__pyx_tp_new_8daedalus_5_maze___pyx_scope_struct____iter__(__pyx_ptype_8daedalus_5_maze___pyx_scope_struct____iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 298, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_8daedalus_5_maze_4Maze_12generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_Maze___iter, __pyx_n_s_daedalus__maze); if (unlikely(!gen)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_8daedalus_5_maze_4Maze_12generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__ *__pyx_cur_scope = ((struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L12_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 298, __pyx_L1_error)

  /* "daedalus/_maze.pyx":299
 * 
 *     def __iter__(self):
 *         i = 0             # <<<<<<<<<<<<<<
 *         try:
 *             while True:
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __pyx_cur_scope->__pyx_v_i = __pyx_int_0;

  /* "daedalus/_maze.pyx":300
 *     def __iter__(self):
 *         i = 0
 *         try:             # <<<<<<<<<<<<<<
 *             while True:
 *                 v = self[i]
 */
  {
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "daedalus/_maze.pyx":301
 *         i = 0
 *         try:
 *             while True:             # <<<<<<<<<<<<<<
 *                 v = self[i]
 *                 yield v
 */
      while (1) {

        /* "daedalus/_maze.pyx":302
 *         try:
 *             while True:
 *                 v = self[i]             # <<<<<<<<<<<<<<
 *                 yield v
 *                 i += 1
 */
        __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_cur_scope->__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 302, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
        __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        __pyx_t_4 = 0;

        /* "daedalus/_maze.pyx":303
 *             while True:
 *                 v = self[i]
 *                 yield v             # <<<<<<<<<<<<<<
 *                 i += 1
 *         except IndexError:
 */
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_v);
        __pyx_r = __pyx_cur_scope->__pyx_v_v;
        __Pyx_XGIVEREF(__pyx_t_1);
        __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
        __Pyx_XGIVEREF(__pyx_t_2);
        __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
        __Pyx_XGIVEREF(__pyx_t_3);
        __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
        __Pyx_XGIVEREF(__pyx_r);
        __Pyx_RefNannyFinishContext();
        __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
        /* return from generator, yielding value */
        __pyx_generator->resume_label = 1;
        return __pyx_r;
        __pyx_L12_resume_from_yield:;
        __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
        __pyx_cur_scope->__pyx_t_0 = 0;
        __Pyx_XGOTREF(__pyx_t_1);
        __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
        __pyx_cur_scope->__pyx_t_1 = 0;
        __Pyx_XGOTREF(__pyx_t_2);
        __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
        __pyx_cur_scope->__pyx_t_2 = 0;
        __Pyx_XGOTREF(__pyx_t_3);
        if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 303, __pyx_L4_error)

        /* "daedalus/_maze.pyx":304
 *                 v = self[i]
 *                 yield v
 *                 i += 1             # <<<<<<<<<<<<<<
 *         except IndexError:
 *             return
 */
        __pyx_t_4 = __Pyx_PyInt_AddObjC(__pyx_cur_scope->__pyx_v_i, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 304, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_i);
        __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_i, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        __pyx_t_4 = 0;
      }

      /* "daedalus/_maze.pyx":300
 *     def __iter__(self):
 *         i = 0
 *         try:             # <<<<<<<<<<<<<<
 *             while True:
 *                 v = self[i]
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "daedalus/_maze.pyx":305
 *                 yield v
 *                 i += 1
 *         except IndexError:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_IndexError);
    if (__pyx_t_5) {
      __Pyx_AddTraceback("daedalus._maze.Maze.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 305, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "daedalus/_maze.pyx":306
 *                 i += 1
 *         except IndexError:
 *             return             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L7_except_return;
    }
    goto __pyx_L6_except_error;
    __pyx_L6_except_error:;

    /* "daedalus/_maze.pyx":300
 *     def __iter__(self):
 *         i = 0
 *         try:             # <<<<<<<<<<<<<<
 *             while True:
 *                 v = self[i]
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L9_try_end:;
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "daedalus/_maze.pyx":298
 *         return False
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         i = 0
 *         try:
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":308
 *             return
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return '<Maze: {}x{}>'.format(self.width, self.height)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_14__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_14__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_13__str__(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_13__str__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "daedalus/_maze.pyx":309
 * 
 *     def __str__(self):
 *         return '<Maze: {}x{}>'.format(self.width, self.height)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Maze_x, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_width); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_height); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 309, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 309, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "daedalus/_maze.pyx":308
 *             return
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return '<Maze: {}x{}>'.format(self.width, self.height)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("daedalus._maze.Maze.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":311
 *         return '<Maze: {}x{}>'.format(self.width, self.height)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_16__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_16__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_15__repr__(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_15__repr__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "daedalus/_maze.pyx":312
 * 
 *     def __repr__(self):
 *         return str(self)             # <<<<<<<<<<<<<<
 * 
 *     def index(self, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "daedalus/_maze.pyx":311
 *         return '<Maze: {}x{}>'.format(self.width, self.height)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return str(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("daedalus._maze.Maze.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":314
 *         return str(self)
 * 
 *     def index(self, value):             # <<<<<<<<<<<<<<
 *         '''
 *         S.index(value) -> integer -- return first index of value.  Raises
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_18index(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_17index[] = "\n        S.index(value) -> integer -- return first index of value.  Raises\n        ValueError if the value is not present.\n\n        :param value: A value to search for in this Maze\n        :return: The index of the location in this Maze where `value` is first found\n        :rtype: int\n        :raises ValueError: If `value` is not found in this Maze\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_18index(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("index (wrapper)", 0);
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_17index(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_17index(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("index", 0);

  /* "daedalus/_maze.pyx":324
 *         :raises ValueError: If `value` is not found in this Maze
 *         '''
 *         for i, v in enumerate(self):             # <<<<<<<<<<<<<<
 *             if v == value:
 *                 return i
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_1 = __pyx_int_0;
  if (likely(PyList_CheckExact(((PyObject *)__pyx_v_self))) || PyTuple_CheckExact(((PyObject *)__pyx_v_self))) {
    __pyx_t_2 = ((PyObject *)__pyx_v_self); __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 324, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 324, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 324, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 324, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 324, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 324, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "daedalus/_maze.pyx":325
 *         '''
 *         for i, v in enumerate(self):
 *             if v == value:             # <<<<<<<<<<<<<<
 *                 return i
 *         raise ValueError
 */
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_v, __pyx_v_value, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 325, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_6) {

      /* "daedalus/_maze.pyx":326
 *         for i, v in enumerate(self):
 *             if v == value:
 *                 return i             # <<<<<<<<<<<<<<
 *         raise ValueError
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_i);
      __pyx_r = __pyx_v_i;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "daedalus/_maze.pyx":325
 *         '''
 *         for i, v in enumerate(self):
 *             if v == value:             # <<<<<<<<<<<<<<
 *                 return i
 *         raise ValueError
 */
    }

    /* "daedalus/_maze.pyx":324
 *         :raises ValueError: If `value` is not found in this Maze
 *         '''
 *         for i, v in enumerate(self):             # <<<<<<<<<<<<<<
 *             if v == value:
 *                 return i
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "daedalus/_maze.pyx":327
 *             if v == value:
 *                 return i
 *         raise ValueError             # <<<<<<<<<<<<<<
 * 
 *     def count(self, value):
 */
  __Pyx_Raise(__pyx_builtin_ValueError, 0, 0, 0);
  __PYX_ERR(0, 327, __pyx_L1_error)

  /* "daedalus/_maze.pyx":314
 *         return str(self)
 * 
 *     def index(self, value):             # <<<<<<<<<<<<<<
 *         '''
 *         S.index(value) -> integer -- return first index of value.  Raises
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("daedalus._maze.Maze.index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":329
 *         raise ValueError
 * 
 *     def count(self, value):             # <<<<<<<<<<<<<<
 *         '''
 *         S.count(value) -> integer -- return number of occurrences of
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_20count(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_19count[] = "\n        S.count(value) -> integer -- return number of occurrences of\n        value.\n\n        :param value: A value to count in this Maze\n        :return: The number of times `value` is found in this Maze.\n        :rtype: int\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_20count(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("count (wrapper)", 0);
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_19count(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_8daedalus_5_maze_4Maze_5count_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "daedalus/_maze.pyx":338
 *         :rtype: int
 *         '''
 *         return sum(1 for v in self if v == value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_5count_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr *)__pyx_tp_new_8daedalus_5_maze___pyx_scope_struct_2_genexpr(__pyx_ptype_8daedalus_5_maze___pyx_scope_struct_2_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 338, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_8daedalus_5_maze_4Maze_5count_2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_count_locals_genexpr, __pyx_n_s_daedalus__maze); if (unlikely(!gen)) __PYX_ERR(0, 338, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.count.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_8daedalus_5_maze_4Maze_5count_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr *__pyx_cur_scope = ((struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 338, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 338, __pyx_L1_error) }
  if (likely(PyList_CheckExact(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self))) || PyTuple_CheckExact(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self))) {
    __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self); __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 338, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 338, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 338, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 338, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 338, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 338, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_value)) { __Pyx_RaiseClosureNameError("value"); __PYX_ERR(0, 338, __pyx_L1_error) }
    __pyx_t_4 = PyObject_RichCompare(__pyx_cur_scope->__pyx_v_v, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_value, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 338, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 338, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {
      __Pyx_INCREF(__pyx_int_1);
      __pyx_r = __pyx_int_1;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 338, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":329
 *         raise ValueError
 * 
 *     def count(self, value):             # <<<<<<<<<<<<<<
 *         '''
 *         S.count(value) -> integer -- return number of occurrences of
 */

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_19count(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_value) {
  struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("count", 0);
  __pyx_cur_scope = (struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count *)__pyx_tp_new_8daedalus_5_maze___pyx_scope_struct_1_count(__pyx_ptype_8daedalus_5_maze___pyx_scope_struct_1_count, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 329, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_value = __pyx_v_value;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_value);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_value);

  /* "daedalus/_maze.pyx":338
 *         :rtype: int
 *         '''
 *         return sum(1 for v in self if v == value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_pf_8daedalus_5_maze_4Maze_5count_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "daedalus/_maze.pyx":329
 *         raise ValueError
 * 
 *     def count(self, value):             # <<<<<<<<<<<<<<
 *         '''
 *         S.count(value) -> integer -- return number of occurrences of
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("daedalus._maze.Maze.count", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":341
 * 
 *     @property
 *     def width(self):             # <<<<<<<<<<<<<<
 *         '''
 *         Returns this Maze's width.
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_5width_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_5width_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_5width___get__(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_5width___get__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "daedalus/_maze.pyx":345
 *         Returns this Maze's width.
 *         '''
 *         if self._maze is NULL:             # <<<<<<<<<<<<<<
 *             return 0
 *         return self._maze.m_x
 */
  __pyx_t_1 = ((__pyx_v_self->_maze == NULL) != 0);
  if (__pyx_t_1) {

    /* "daedalus/_maze.pyx":346
 *         '''
 *         if self._maze is NULL:
 *             return 0             # <<<<<<<<<<<<<<
 *         return self._maze.m_x
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_int_0);
    __pyx_r = __pyx_int_0;
    goto __pyx_L0;

    /* "daedalus/_maze.pyx":345
 *         Returns this Maze's width.
 *         '''
 *         if self._maze is NULL:             # <<<<<<<<<<<<<<
 *             return 0
 *         return self._maze.m_x
 */
  }

  /* "daedalus/_maze.pyx":347
 *         if self._maze is NULL:
 *             return 0
 *         return self._maze.m_x             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->_maze->m_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "daedalus/_maze.pyx":341
 * 
 *     @property
 *     def width(self):             # <<<<<<<<<<<<<<
 *         '''
 *         Returns this Maze's width.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("daedalus._maze.Maze.width.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":350
 * 
 *     @property
 *     def height(self):             # <<<<<<<<<<<<<<
 *         '''
 *         Returns this Maze's height.
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_6height_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_6height_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_6height___get__(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_6height___get__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "daedalus/_maze.pyx":354
 *         Returns this Maze's height.
 *         '''
 *         if self._maze is NULL:             # <<<<<<<<<<<<<<
 *             return 0
 *         return self._maze.m_y
 */
  __pyx_t_1 = ((__pyx_v_self->_maze == NULL) != 0);
  if (__pyx_t_1) {

    /* "daedalus/_maze.pyx":355
 *         '''
 *         if self._maze is NULL:
 *             return 0             # <<<<<<<<<<<<<<
 *         return self._maze.m_y
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_int_0);
    __pyx_r = __pyx_int_0;
    goto __pyx_L0;

    /* "daedalus/_maze.pyx":354
 *         Returns this Maze's height.
 *         '''
 *         if self._maze is NULL:             # <<<<<<<<<<<<<<
 *             return 0
 *         return self._maze.m_y
 */
  }

  /* "daedalus/_maze.pyx":356
 *         if self._maze is NULL:
 *             return 0
 *         return self._maze.m_y             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->_maze->m_y); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "daedalus/_maze.pyx":350
 * 
 *     @property
 *     def height(self):             # <<<<<<<<<<<<<<
 *         '''
 *         Returns this Maze's height.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("daedalus._maze.Maze.height.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":359
 * 
 *     @property
 *     def entrance(self):             # <<<<<<<<<<<<<<
 *         '''
 *         Returns this Maze's entrance location as a tuple (x,y).
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_8entrance_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_8entrance_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_8entrance___get__(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_8entrance___get__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "daedalus/_maze.pyx":363
 *         Returns this Maze's entrance location as a tuple (x,y).
 *         '''
 *         return (self._xEntrance, self._yEntrance)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_xEntrance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->_yEntrance); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "daedalus/_maze.pyx":359
 * 
 *     @property
 *     def entrance(self):             # <<<<<<<<<<<<<<
 *         '''
 *         Returns this Maze's entrance location as a tuple (x,y).
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("daedalus._maze.Maze.entrance.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":366
 * 
 *     @property
 *     def exit(self):             # <<<<<<<<<<<<<<
 *         '''
 *         Returns this Maze's exit location as a tuple (x,y).
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_4exit_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_4exit_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_4exit___get__(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_4exit___get__(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "daedalus/_maze.pyx":370
 *         Returns this Maze's exit location as a tuple (x,y).
 *         '''
 *         return (self._xExit, self._yExit)             # <<<<<<<<<<<<<<
 * 
 *     def get(self, x, y):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_xExit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->_yExit); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "daedalus/_maze.pyx":366
 * 
 *     @property
 *     def exit(self):             # <<<<<<<<<<<<<<
 *         '''
 *         Returns this Maze's exit location as a tuple (x,y).
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("daedalus._maze.Maze.exit.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":372
 *         return (self._xExit, self._yExit)
 * 
 *     def get(self, x, y):             # <<<<<<<<<<<<<<
 *         '''
 *         Gets the monochrome pixel value at the coordinate (x,y) on this
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_22get(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_21get[] = "\n        Gets the monochrome pixel value at the coordinate (x,y) on this\n        Maze.\n\n        :param int x: The x-coordinate of the pixel to get\n        :param int y: The y-coordinate of the pixel to get\n        :return: True if the given pixel is on, False if it is off.\n        :rtype: bool\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_22get(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_x = 0;
  PyObject *__pyx_v_y = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get", 1, 2, 2, 1); __PYX_ERR(0, 372, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get") < 0)) __PYX_ERR(0, 372, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_x = values[0];
    __pyx_v_y = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 372, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_21get(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_x, __pyx_v_y);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_21get(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_x, PyObject *__pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  unsigned int __pyx_t_6;
  __Pyx_RefNannySetupContext("get", 0);

  /* "daedalus/_maze.pyx":382
 *         :rtype: bool
 *         '''
 *         if x < 0 or self.width <= x:             # <<<<<<<<<<<<<<
 *             raise IndexError('x coordinate out of range')
 *         if y < 0 or self.height <= y:
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_x, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 382, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_width); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_v_x, Py_LE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "daedalus/_maze.pyx":383
 *         '''
 *         if x < 0 or self.width <= x:
 *             raise IndexError('x coordinate out of range')             # <<<<<<<<<<<<<<
 *         if y < 0 or self.height <= y:
 *             raise IndexError('y coordinate out of range')
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 383, __pyx_L1_error)

    /* "daedalus/_maze.pyx":382
 *         :rtype: bool
 *         '''
 *         if x < 0 or self.width <= x:             # <<<<<<<<<<<<<<
 *             raise IndexError('x coordinate out of range')
 *         if y < 0 or self.height <= y:
 */
  }

  /* "daedalus/_maze.pyx":384
 *         if x < 0 or self.width <= x:
 *             raise IndexError('x coordinate out of range')
 *         if y < 0 or self.height <= y:             # <<<<<<<<<<<<<<
 *             raise IndexError('y coordinate out of range')
 *         return cpp_Get(self._maze, x, y)
 */
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_y, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 384, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_height); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_4, __pyx_v_y, Py_LE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "daedalus/_maze.pyx":385
 *             raise IndexError('x coordinate out of range')
 *         if y < 0 or self.height <= y:
 *             raise IndexError('y coordinate out of range')             # <<<<<<<<<<<<<<
 *         return cpp_Get(self._maze, x, y)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 385, __pyx_L1_error)

    /* "daedalus/_maze.pyx":384
 *         if x < 0 or self.width <= x:
 *             raise IndexError('x coordinate out of range')
 *         if y < 0 or self.height <= y:             # <<<<<<<<<<<<<<
 *             raise IndexError('y coordinate out of range')
 *         return cpp_Get(self._maze, x, y)
 */
  }

  /* "daedalus/_maze.pyx":386
 *         if y < 0 or self.height <= y:
 *             raise IndexError('y coordinate out of range')
 *         return cpp_Get(self._maze, x, y)             # <<<<<<<<<<<<<<
 * 
 *     def create_perfect(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyInt_As_unsigned_int(__pyx_v_x); if (unlikely((__pyx_t_5 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 386, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyInt_As_unsigned_int(__pyx_v_y); if (unlikely((__pyx_t_6 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 386, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(cpp_Get(__pyx_v_self->_maze, __pyx_t_5, __pyx_t_6)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "daedalus/_maze.pyx":372
 *         return (self._xExit, self._yExit)
 * 
 *     def get(self, x, y):             # <<<<<<<<<<<<<<
 *         '''
 *         Gets the monochrome pixel value at the coordinate (x,y) on this
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("daedalus._maze.Maze.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":388
 *         return cpp_Get(self._maze, x, y)
 * 
 *     def create_perfect(self,             # <<<<<<<<<<<<<<
 *                        fRiver=True,
 *                        fTreeWall=False,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_24create_perfect(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_23create_perfect[] = "\n        Overwrites the contents of this Maze to create a perfect maze\n        using the Hunt and Kill algorithm.\n\n        Create a new perfect Maze in the bitmap using the Hunt and\n        Kill algorithm, by carving passages.\n\n        This algorithm creates the Maze by carving passages, where it\n        behaves similarly to Recursive Backtrack. The difference is\n        when it can't add onto the current passage, it will enter\n        \"hunting\" mode and search over the Maze until an unmade\n        section is found next to an already carved passage. By default\n        this results in Mazes with a high \"river\" factor, but not as\n        high as Recursive Backtrack. If fRiver is False, the Maze will\n        instead have a low \"river\" factor, about the same as Kruskal's\n        Algorithm. This method runs slightly slower than Kruskal's\n        Algorithm.\n\n        :param bool fRiver: If set, creates a Maze with a relatively\n            high \"river\" factor (with relatively longer passages between\n            junctions and fewer but longer dead ends).  Defaults to True.\n        :param bool fTreeWall: If true, creates the Maze by adding\n            walls; if false, creates the Maze by carving passages.\n            Defaults to False.\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value i""s positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_24create_perfect(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fRiver = 0;
  PyObject *__pyx_v_fTreeWall = 0;
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_perfect (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fRiver,&__pyx_n_s_fTreeWall,&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[5] = {0,0,0,0,0};

    /* "daedalus/_maze.pyx":389
 * 
 *     def create_perfect(self,
 *                        fRiver=True,             # <<<<<<<<<<<<<<
 *                        fTreeWall=False,
 *                        nEntrancePos=ENTRANCE_RANDOM,
 */
    values[0] = ((PyObject *)Py_True);

    /* "daedalus/_maze.pyx":390
 *     def create_perfect(self,
 *                        fRiver=True,
 *                        fTreeWall=False,             # <<<<<<<<<<<<<<
 *                        nEntrancePos=ENTRANCE_RANDOM,
 *                        nRndBias=0,
 */
    values[1] = ((PyObject *)Py_False);
    values[2] = __pyx_k__6;
    values[3] = ((PyObject *)__pyx_int_0);
    values[4] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fRiver);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTreeWall);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_perfect") < 0)) __PYX_ERR(0, 388, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fRiver = values[0];
    __pyx_v_fTreeWall = values[1];
    __pyx_v_nEntrancePos = values[2];
    __pyx_v_nRndBias = values[3];
    __pyx_v_nRndRun = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_perfect", 0, 0, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 388, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_perfect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_23create_perfect(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_fRiver, __pyx_v_fTreeWall, __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* "daedalus/_maze.pyx":388
 *         return cpp_Get(self._maze, x, y)
 * 
 *     def create_perfect(self,             # <<<<<<<<<<<<<<
 *                        fRiver=True,
 *                        fTreeWall=False,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_23create_perfect(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fRiver, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  bool __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("create_perfect", 0);

  /* "daedalus/_maze.pyx":445
 *         '''
 *         if not cpp_CreateMazePerfect(self._maze,
 *                                      fRiver,             # <<<<<<<<<<<<<<
 *                                      True,
 *                                      True,
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_fRiver); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 445, __pyx_L1_error)

  /* "daedalus/_maze.pyx":449
 *                                      True,
 *                                      False,
 *                                      fTreeWall,             # <<<<<<<<<<<<<<
 *                                      nEntrancePos,
 *                                      nRndBias,
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_fTreeWall); if (unlikely((__pyx_t_2 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 449, __pyx_L1_error)

  /* "daedalus/_maze.pyx":450
 *                                      False,
 *                                      fTreeWall,
 *                                      nEntrancePos,             # <<<<<<<<<<<<<<
 *                                      nRndBias,
 *                                      nRndRun):
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 450, __pyx_L1_error)

  /* "daedalus/_maze.pyx":451
 *                                      fTreeWall,
 *                                      nEntrancePos,
 *                                      nRndBias,             # <<<<<<<<<<<<<<
 *                                      nRndRun):
 *             raise MazeError('Could not create Perfect Maze.')
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 451, __pyx_L1_error)

  /* "daedalus/_maze.pyx":452
 *                                      nEntrancePos,
 *                                      nRndBias,
 *                                      nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Perfect Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 452, __pyx_L1_error)

  /* "daedalus/_maze.pyx":444
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazePerfect(self._maze,             # <<<<<<<<<<<<<<
 *                                      fRiver,
 *                                      True,
 */
  __pyx_t_6 = ((!(cpp_CreateMazePerfect(__pyx_v_self->_maze, __pyx_t_1, 1, 1, 0, __pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5) != 0)) != 0);
  if (unlikely(__pyx_t_6)) {

    /* "daedalus/_maze.pyx":453
 *                                      nRndBias,
 *                                      nRndRun):
 *             raise MazeError('Could not create Perfect Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_kp_s_Could_not_create_Perfect_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_kp_s_Could_not_create_Perfect_Maze);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 453, __pyx_L1_error)

    /* "daedalus/_maze.pyx":444
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazePerfect(self._maze,             # <<<<<<<<<<<<<<
 *                                      fRiver,
 *                                      True,
 */
  }

  /* "daedalus/_maze.pyx":454
 *                                      nRndRun):
 *             raise MazeError('Could not create Perfect Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_perfect2(self,
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "daedalus/_maze.pyx":388
 *         return cpp_Get(self._maze, x, y)
 * 
 *     def create_perfect(self,             # <<<<<<<<<<<<<<
 *                        fRiver=True,
 *                        fTreeWall=False,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_perfect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":456
 *         self._store_back_globals()
 * 
 *     def create_perfect2(self,             # <<<<<<<<<<<<<<
 *                         fRiver=True,
 *                         nEntrancePos=ENTRANCE_RANDOM,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_26create_perfect2(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_25create_perfect2[] = "\n        Overwrites the contents of this Maze to create a Perfect2 maze.\n\n        Create a new perfect Maze in the bitmap using the Hunt and\n        Kill algorithm, by adding walls.\n\n        :param bool fRiver: If set, creates a Maze with a relatively\n            high \"river\" factor (with relatively longer passages between\n            junctions and fewer but longer dead ends).  Defaults to True.\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher"" the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_26create_perfect2(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fRiver = 0;
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_perfect2 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fRiver,&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[4] = {0,0,0,0};

    /* "daedalus/_maze.pyx":457
 * 
 *     def create_perfect2(self,
 *                         fRiver=True,             # <<<<<<<<<<<<<<
 *                         nEntrancePos=ENTRANCE_RANDOM,
 *                         nRndBias=0,
 */
    values[0] = ((PyObject *)Py_True);
    values[1] = __pyx_k__7;
    values[2] = ((PyObject *)__pyx_int_0);
    values[3] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fRiver);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_perfect2") < 0)) __PYX_ERR(0, 456, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fRiver = values[0];
    __pyx_v_nEntrancePos = values[1];
    __pyx_v_nRndBias = values[2];
    __pyx_v_nRndRun = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_perfect2", 0, 0, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 456, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_perfect2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_25create_perfect2(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_fRiver, __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* "daedalus/_maze.pyx":456
 *         self._store_back_globals()
 * 
 *     def create_perfect2(self,             # <<<<<<<<<<<<<<
 *                         fRiver=True,
 *                         nEntrancePos=ENTRANCE_RANDOM,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_25create_perfect2(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fRiver, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("create_perfect2", 0);

  /* "daedalus/_maze.pyx":497
 *         '''
 *         if not cpp_CreateMazePerfect2(self._maze,
 *                                       fRiver,             # <<<<<<<<<<<<<<
 *                                       False,
 *                                       nEntrancePos,
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_fRiver); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 497, __pyx_L1_error)

  /* "daedalus/_maze.pyx":499
 *                                       fRiver,
 *                                       False,
 *                                       nEntrancePos,             # <<<<<<<<<<<<<<
 *                                       nRndBias,
 *                                       nRndRun):
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 499, __pyx_L1_error)

  /* "daedalus/_maze.pyx":500
 *                                       False,
 *                                       nEntrancePos,
 *                                       nRndBias,             # <<<<<<<<<<<<<<
 *                                       nRndRun):
 *             raise MazeError('Could not create Perfect2 Maze.')
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 500, __pyx_L1_error)

  /* "daedalus/_maze.pyx":501
 *                                       nEntrancePos,
 *                                       nRndBias,
 *                                       nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Perfect2 Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 501, __pyx_L1_error)

  /* "daedalus/_maze.pyx":496
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazePerfect2(self._maze,             # <<<<<<<<<<<<<<
 *                                       fRiver,
 *                                       False,
 */
  __pyx_t_5 = ((!(cpp_CreateMazePerfect2(__pyx_v_self->_maze, __pyx_t_1, 0, __pyx_t_2, __pyx_t_3, __pyx_t_4) != 0)) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "daedalus/_maze.pyx":502
 *                                       nRndBias,
 *                                       nRndRun):
 *             raise MazeError('Could not create Perfect2 Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_kp_s_Could_not_create_Perfect2_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_s_Could_not_create_Perfect2_Maze);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 502, __pyx_L1_error)

    /* "daedalus/_maze.pyx":496
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazePerfect2(self._maze,             # <<<<<<<<<<<<<<
 *                                       fRiver,
 *                                       False,
 */
  }

  /* "daedalus/_maze.pyx":503
 *                                       nRndRun):
 *             raise MazeError('Could not create Perfect2 Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_braid(self,
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "daedalus/_maze.pyx":456
 *         self._store_back_globals()
 * 
 *     def create_perfect2(self,             # <<<<<<<<<<<<<<
 *                         fRiver=True,
 *                         nEntrancePos=ENTRANCE_RANDOM,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_perfect2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":505
 *         self._store_back_globals()
 * 
 *     def create_braid(self,             # <<<<<<<<<<<<<<
 *                      nEntrancePos=ENTRANCE_RANDOM,
 *                      nRndBias=0,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_28create_braid(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_27create_braid[] = "\n        Overwrites the contents of this Maze to create a Braid maze.\n\n        Create a new braid Maze in the bitmap, i.e. a Maze without any\n        dead ends, using a wall adding algorithm.\n\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_28create_braid(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_braid (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[3] = {0,0,0};
    values[0] = __pyx_k__8;
    values[1] = ((PyObject *)__pyx_int_0);
    values[2] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_braid") < 0)) __PYX_ERR(0, 505, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_nEntrancePos = values[0];
    __pyx_v_nRndBias = values[1];
    __pyx_v_nRndRun = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_braid", 0, 0, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 505, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_braid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_27create_braid(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_27create_braid(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("create_braid", 0);

  /* "daedalus/_maze.pyx":543
 *         if not cpp_CreateMazeBraid(self._maze,
 *                                    False,
 *                                    nEntrancePos,             # <<<<<<<<<<<<<<
 *                                    nRndBias,
 *                                    nRndRun):
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 543, __pyx_L1_error)

  /* "daedalus/_maze.pyx":544
 *                                    False,
 *                                    nEntrancePos,
 *                                    nRndBias,             # <<<<<<<<<<<<<<
 *                                    nRndRun):
 *             raise MazeError('Could not create Braid Maze.')
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 544, __pyx_L1_error)

  /* "daedalus/_maze.pyx":545
 *                                    nEntrancePos,
 *                                    nRndBias,
 *                                    nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Braid Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 545, __pyx_L1_error)

  /* "daedalus/_maze.pyx":541
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeBraid(self._maze,             # <<<<<<<<<<<<<<
 *                                    False,
 *                                    nEntrancePos,
 */
  __pyx_t_4 = ((!(cpp_CreateMazeBraid(__pyx_v_self->_maze, 0, __pyx_t_1, __pyx_t_2, __pyx_t_3) != 0)) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "daedalus/_maze.pyx":546
 *                                    nRndBias,
 *                                    nRndRun):
 *             raise MazeError('Could not create Braid Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 546, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_kp_s_Could_not_create_Braid_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_s_Could_not_create_Braid_Maze);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 546, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 546, __pyx_L1_error)

    /* "daedalus/_maze.pyx":541
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeBraid(self._maze,             # <<<<<<<<<<<<<<
 *                                    False,
 *                                    nEntrancePos,
 */
  }

  /* "daedalus/_maze.pyx":547
 *                                    nRndRun):
 *             raise MazeError('Could not create Braid Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_braid_tilt(self,
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "daedalus/_maze.pyx":505
 *         self._store_back_globals()
 * 
 *     def create_braid(self,             # <<<<<<<<<<<<<<
 *                      nEntrancePos=ENTRANCE_RANDOM,
 *                      nRndBias=0,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_braid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":549
 *         self._store_back_globals()
 * 
 *     def create_braid_tilt(self,             # <<<<<<<<<<<<<<
 *                           fTiltDiamond=False,
 *                           nEntrancePos=ENTRANCE_RANDOM,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_30create_braid_tilt(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_29create_braid_tilt[] = "\n        Overwrites the contents of this Maze to create a BraidTilt maze.\n\n        Create a new braid Maze in the bitmap using a binary\n        algorithm, i.e.  starting with a template based on the Tilt\n        Maze pattern.\n\n        :param bool fTiltDiamond: When this is set, there will never\n            be any \"diamonds\" or places where 2x2 groups of diagonal\n            lines fit together to make a diamond shaped\n            space. Defaults to False.\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The"" more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_30create_braid_tilt(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fTiltDiamond = 0;
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_braid_tilt (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fTiltDiamond,&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[4] = {0,0,0,0};

    /* "daedalus/_maze.pyx":550
 * 
 *     def create_braid_tilt(self,
 *                           fTiltDiamond=False,             # <<<<<<<<<<<<<<
 *                           nEntrancePos=ENTRANCE_RANDOM,
 *                           nRndBias=0,
 */
    values[0] = ((PyObject *)Py_False);
    values[1] = __pyx_k__9;
    values[2] = ((PyObject *)__pyx_int_0);
    values[3] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTiltDiamond);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_braid_tilt") < 0)) __PYX_ERR(0, 549, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fTiltDiamond = values[0];
    __pyx_v_nEntrancePos = values[1];
    __pyx_v_nRndBias = values[2];
    __pyx_v_nRndRun = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_braid_tilt", 0, 0, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 549, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_braid_tilt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_29create_braid_tilt(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_fTiltDiamond, __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* "daedalus/_maze.pyx":549
 *         self._store_back_globals()
 * 
 *     def create_braid_tilt(self,             # <<<<<<<<<<<<<<
 *                           fTiltDiamond=False,
 *                           nEntrancePos=ENTRANCE_RANDOM,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_29create_braid_tilt(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fTiltDiamond, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("create_braid_tilt", 0);

  /* "daedalus/_maze.pyx":593
 *         if not cpp_CreateMazeBraidTilt(self._maze,
 *                                        False,
 *                                        fTiltDiamond,             # <<<<<<<<<<<<<<
 *                                        nEntrancePos,
 *                                        nRndBias,
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_fTiltDiamond); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 593, __pyx_L1_error)

  /* "daedalus/_maze.pyx":594
 *                                        False,
 *                                        fTiltDiamond,
 *                                        nEntrancePos,             # <<<<<<<<<<<<<<
 *                                        nRndBias,
 *                                        nRndRun):
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 594, __pyx_L1_error)

  /* "daedalus/_maze.pyx":595
 *                                        fTiltDiamond,
 *                                        nEntrancePos,
 *                                        nRndBias,             # <<<<<<<<<<<<<<
 *                                        nRndRun):
 *             raise MazeError('Could not create BraidTilt Maze.')
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 595, __pyx_L1_error)

  /* "daedalus/_maze.pyx":596
 *                                        nEntrancePos,
 *                                        nRndBias,
 *                                        nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create BraidTilt Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 596, __pyx_L1_error)

  /* "daedalus/_maze.pyx":591
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeBraidTilt(self._maze,             # <<<<<<<<<<<<<<
 *                                        False,
 *                                        fTiltDiamond,
 */
  __pyx_t_5 = ((!(cpp_CreateMazeBraidTilt(__pyx_v_self->_maze, 0, __pyx_t_1, __pyx_t_2, __pyx_t_3, __pyx_t_4) != 0)) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "daedalus/_maze.pyx":597
 *                                        nRndBias,
 *                                        nRndRun):
 *             raise MazeError('Could not create BraidTilt Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 597, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_kp_s_Could_not_create_BraidTilt_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_s_Could_not_create_BraidTilt_Maze);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 597, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 597, __pyx_L1_error)

    /* "daedalus/_maze.pyx":591
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeBraidTilt(self._maze,             # <<<<<<<<<<<<<<
 *                                        False,
 *                                        fTiltDiamond,
 */
  }

  /* "daedalus/_maze.pyx":598
 *                                        nRndRun):
 *             raise MazeError('Could not create BraidTilt Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_spiral(self,
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "daedalus/_maze.pyx":549
 *         self._store_back_globals()
 * 
 *     def create_braid_tilt(self,             # <<<<<<<<<<<<<<
 *                           fTiltDiamond=False,
 *                           nEntrancePos=ENTRANCE_RANDOM,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_braid_tilt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":600
 *         self._store_back_globals()
 * 
 *     def create_spiral(self,             # <<<<<<<<<<<<<<
 *                       cRandomAdd=0,
 *                       cSpiral=15,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_32create_spiral(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_31create_spiral[] = "\n        Overwrites the contents of this Maze to create a Spiral maze.\n\n        Create a new spiral Maze in the bitmap, formed of interlocking\n        spirals.\n\n        :param int cRandomAdd: The number of random walls and passages\n            to add to the maze.  A high number will make the maze more\n            random and less like its base design, while a large negative\n            number will make the maze be its base design with longer\n            passages and dead ends.  Defaults to 0.\n        :param int cSpiral: The maximum number of spirals that can be\n            drawn at once.  Larger numbers make the maze have many\n            smaller spirals; smaller number make the maze have fewer,\n            larger spirals.  Defaults to 15.\n        :param int cSpiralWall: The maximum number of walls that can\n            be drawn at once around each spiral.  Larger numbers make\n            spirals have many passages that rapidly spiral away from\n            the centre, while smaller values make spirals with fewer\n            passages that go around the centre more times. Defaults to\n            15.\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther"" the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_32create_spiral(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cRandomAdd = 0;
  PyObject *__pyx_v_cSpiral = 0;
  PyObject *__pyx_v_cSpiralWall = 0;
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_spiral (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cRandomAdd,&__pyx_n_s_cSpiral,&__pyx_n_s_cSpiralWall,&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    values[0] = ((PyObject *)__pyx_int_0);
    values[1] = ((PyObject *)__pyx_int_15);
    values[2] = ((PyObject *)__pyx_int_15);
    values[3] = __pyx_k__10;
    values[4] = ((PyObject *)__pyx_int_0);
    values[5] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cRandomAdd);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cSpiral);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cSpiralWall);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_spiral") < 0)) __PYX_ERR(0, 600, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cRandomAdd = values[0];
    __pyx_v_cSpiral = values[1];
    __pyx_v_cSpiralWall = values[2];
    __pyx_v_nEntrancePos = values[3];
    __pyx_v_nRndBias = values[4];
    __pyx_v_nRndRun = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_spiral", 0, 0, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 600, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_spiral", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_31create_spiral(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_cRandomAdd, __pyx_v_cSpiral, __pyx_v_cSpiralWall, __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_31create_spiral(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_cRandomAdd, PyObject *__pyx_v_cSpiral, PyObject *__pyx_v_cSpiralWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("create_spiral", 0);

  /* "daedalus/_maze.pyx":655
 *         '''
 *         if not cpp_CreateMazeSpiral(self._maze,
 *                                     cRandomAdd,             # <<<<<<<<<<<<<<
 *                                     cSpiral,
 *                                     cSpiralWall,
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_cRandomAdd); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 655, __pyx_L1_error)

  /* "daedalus/_maze.pyx":656
 *         if not cpp_CreateMazeSpiral(self._maze,
 *                                     cRandomAdd,
 *                                     cSpiral,             # <<<<<<<<<<<<<<
 *                                     cSpiralWall,
 *                                     False,
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_cSpiral); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 656, __pyx_L1_error)

  /* "daedalus/_maze.pyx":657
 *                                     cRandomAdd,
 *                                     cSpiral,
 *                                     cSpiralWall,             # <<<<<<<<<<<<<<
 *                                     False,
 *                                     nEntrancePos,
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_cSpiralWall); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 657, __pyx_L1_error)

  /* "daedalus/_maze.pyx":659
 *                                     cSpiralWall,
 *                                     False,
 *                                     nEntrancePos,             # <<<<<<<<<<<<<<
 *                                     nRndBias,
 *                                     nRndRun):
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 659, __pyx_L1_error)

  /* "daedalus/_maze.pyx":660
 *                                     False,
 *                                     nEntrancePos,
 *                                     nRndBias,             # <<<<<<<<<<<<<<
 *                                     nRndRun):
 *             raise MazeError('Could not create Spiral Maze.')
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 660, __pyx_L1_error)

  /* "daedalus/_maze.pyx":661
 *                                     nEntrancePos,
 *                                     nRndBias,
 *                                     nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Spiral Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 661, __pyx_L1_error)

  /* "daedalus/_maze.pyx":654
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeSpiral(self._maze,             # <<<<<<<<<<<<<<
 *                                     cRandomAdd,
 *                                     cSpiral,
 */
  __pyx_t_7 = ((!(cpp_CreateMazeSpiral(__pyx_v_self->_maze, __pyx_t_1, __pyx_t_2, __pyx_t_3, 0, __pyx_t_4, __pyx_t_5, __pyx_t_6) != 0)) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "daedalus/_maze.pyx":662
 *                                     nRndBias,
 *                                     nRndRun):
 *             raise MazeError('Could not create Spiral Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_kp_s_Could_not_create_Spiral_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_kp_s_Could_not_create_Spiral_Maze);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_Raise(__pyx_t_8, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_ERR(0, 662, __pyx_L1_error)

    /* "daedalus/_maze.pyx":654
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeSpiral(self._maze,             # <<<<<<<<<<<<<<
 *                                     cRandomAdd,
 *                                     cSpiral,
 */
  }

  /* "daedalus/_maze.pyx":663
 *                                     nRndRun):
 *             raise MazeError('Could not create Spiral Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_diagonal(self,
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "daedalus/_maze.pyx":600
 *         self._store_back_globals()
 * 
 *     def create_spiral(self,             # <<<<<<<<<<<<<<
 *                       cRandomAdd=0,
 *                       cSpiral=15,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_spiral", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":665
 *         self._store_back_globals()
 * 
 *     def create_diagonal(self,             # <<<<<<<<<<<<<<
 *                         cRandomAdd=0,
 *                         nEntrancePos=ENTRANCE_RANDOM,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_34create_diagonal(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_33create_diagonal[] = "\n        Overwrites the contents of this Maze to create a Diagonal maze.\n\n        Create a Maze with a diagonal bias, where many walls look like\n        stairs.\n\n        :param int cRandomAdd: The number of random walls and passages\n            to add to the maze.  A high number will make the maze more\n            random and less like its base design, while a large negative\n            number will make the maze be its base design with longer\n            passages and dead ends.  Defaults to 0.\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and ""will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_34create_diagonal(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cRandomAdd = 0;
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_diagonal (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cRandomAdd,&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[4] = {0,0,0,0};
    values[0] = ((PyObject *)__pyx_int_0);
    values[1] = __pyx_k__11;
    values[2] = ((PyObject *)__pyx_int_0);
    values[3] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cRandomAdd);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_diagonal") < 0)) __PYX_ERR(0, 665, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cRandomAdd = values[0];
    __pyx_v_nEntrancePos = values[1];
    __pyx_v_nRndBias = values[2];
    __pyx_v_nRndRun = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_diagonal", 0, 0, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 665, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_diagonal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_33create_diagonal(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_cRandomAdd, __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_33create_diagonal(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_cRandomAdd, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("create_diagonal", 0);

  /* "daedalus/_maze.pyx":708
 *         '''
 *         if not cpp_CreateMazeDiagonal(self._maze,
 *                                       cRandomAdd,             # <<<<<<<<<<<<<<
 *                                       False,
 *                                       nEntrancePos,
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_cRandomAdd); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 708, __pyx_L1_error)

  /* "daedalus/_maze.pyx":710
 *                                       cRandomAdd,
 *                                       False,
 *                                       nEntrancePos,             # <<<<<<<<<<<<<<
 *                                       nRndBias,
 *                                       nRndRun):
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 710, __pyx_L1_error)

  /* "daedalus/_maze.pyx":711
 *                                       False,
 *                                       nEntrancePos,
 *                                       nRndBias,             # <<<<<<<<<<<<<<
 *                                       nRndRun):
 *             raise MazeError('Could not create Diagonal Maze.')
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 711, __pyx_L1_error)

  /* "daedalus/_maze.pyx":712
 *                                       nEntrancePos,
 *                                       nRndBias,
 *                                       nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Diagonal Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 712, __pyx_L1_error)

  /* "daedalus/_maze.pyx":707
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeDiagonal(self._maze,             # <<<<<<<<<<<<<<
 *                                       cRandomAdd,
 *                                       False,
 */
  __pyx_t_5 = ((!(cpp_CreateMazeDiagonal(__pyx_v_self->_maze, __pyx_t_1, 0, __pyx_t_2, __pyx_t_3, __pyx_t_4) != 0)) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "daedalus/_maze.pyx":713
 *                                       nRndBias,
 *                                       nRndRun):
 *             raise MazeError('Could not create Diagonal Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 713, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_kp_s_Could_not_create_Diagonal_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_s_Could_not_create_Diagonal_Maze);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 713, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 713, __pyx_L1_error)

    /* "daedalus/_maze.pyx":707
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeDiagonal(self._maze,             # <<<<<<<<<<<<<<
 *                                       cRandomAdd,
 *                                       False,
 */
  }

  /* "daedalus/_maze.pyx":714
 *                                       nRndRun):
 *             raise MazeError('Could not create Diagonal Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_recursive(self,
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "daedalus/_maze.pyx":665
 *         self._store_back_globals()
 * 
 *     def create_diagonal(self,             # <<<<<<<<<<<<<<
 *                         cRandomAdd=0,
 *                         nEntrancePos=ENTRANCE_RANDOM,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_diagonal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":716
 *         self._store_back_globals()
 * 
 *     def create_recursive(self,             # <<<<<<<<<<<<<<
 *                          nEntrancePos=ENTRANCE_RANDOM,
 *                          nRndBias=0,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_36create_recursive(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_35create_recursive[] = "\n        Overwrites the contents of this Maze to create a Recursive maze.\n\n        Create a new perfect Maze in the bitmap using the Recursive\n        Backtracking algorithm. This carves passages.\n\n        This algorithm creates the Maze by carving passages, where it\n        always adds onto the most recently created passage whenever\n        possible, and only \"backs up\" to create other sections when\n        it's forced to. This results in Mazes with about as high a\n        \"river\" factor as possible, with fewer but longer dead ends,\n        and usually a very long and twisty solution. This command is\n        fast, although Prim's Algorithm is slightly faster.\n\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than thr""ee or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_36create_recursive(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_recursive (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[3] = {0,0,0};
    values[0] = __pyx_k__12;
    values[1] = ((PyObject *)__pyx_int_0);
    values[2] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_recursive") < 0)) __PYX_ERR(0, 716, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_nEntrancePos = values[0];
    __pyx_v_nRndBias = values[1];
    __pyx_v_nRndRun = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_recursive", 0, 0, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 716, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_recursive", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_35create_recursive(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_35create_recursive(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("create_recursive", 0);

  /* "daedalus/_maze.pyx":762
 *         if not cpp_CreateMazeRecursive(self._maze,
 *                                        False,
 *                                        nEntrancePos,             # <<<<<<<<<<<<<<
 *                                        nRndBias,
 *                                        nRndRun):
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 762, __pyx_L1_error)

  /* "daedalus/_maze.pyx":763
 *                                        False,
 *                                        nEntrancePos,
 *                                        nRndBias,             # <<<<<<<<<<<<<<
 *                                        nRndRun):
 *             raise MazeError('Could not create Recursive Maze.')
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 763, __pyx_L1_error)

  /* "daedalus/_maze.pyx":764
 *                                        nEntrancePos,
 *                                        nRndBias,
 *                                        nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Recursive Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 764, __pyx_L1_error)

  /* "daedalus/_maze.pyx":760
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeRecursive(self._maze,             # <<<<<<<<<<<<<<
 *                                        False,
 *                                        nEntrancePos,
 */
  __pyx_t_4 = ((!(cpp_CreateMazeRecursive(__pyx_v_self->_maze, 0, __pyx_t_1, __pyx_t_2, __pyx_t_3) != 0)) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "daedalus/_maze.pyx":765
 *                                        nRndBias,
 *                                        nRndRun):
 *             raise MazeError('Could not create Recursive Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_kp_s_Could_not_create_Recursive_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_s_Could_not_create_Recursive_Maze);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 765, __pyx_L1_error)

    /* "daedalus/_maze.pyx":760
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeRecursive(self._maze,             # <<<<<<<<<<<<<<
 *                                        False,
 *                                        nEntrancePos,
 */
  }

  /* "daedalus/_maze.pyx":766
 *                                        nRndRun):
 *             raise MazeError('Could not create Recursive Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_prim(self,
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "daedalus/_maze.pyx":716
 *         self._store_back_globals()
 * 
 *     def create_recursive(self,             # <<<<<<<<<<<<<<
 *                          nEntrancePos=ENTRANCE_RANDOM,
 *                          nRndBias=0,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_recursive", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":768
 *         self._store_back_globals()
 * 
 *     def create_prim(self,             # <<<<<<<<<<<<<<
 *                     fTreeWall=False,
 *                     nEntrancePos=ENTRANCE_RANDOM,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_38create_prim(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_37create_prim[] = "\n        Overwrites the contents of this Maze to create a Prim maze.\n\n        Create a new perfect Maze in the bitmap using a modified\n        version of Prim's algorithm. This can carve passages or add\n        walls.\n\n        This algorithm creates the Maze by carving passages, where it\n        attaches new passage segments onto the created portion of the\n        Maze at random points, and will rarely stay with any single\n        passage for any length of time. This results in Mazes with\n        about as low a \"river\" factor as possible, with many short\n        dead ends, and the solution is usually pretty direct too. This\n        command is faster than any of the others except Eller's\n        algorithm.\n\n        :param bool fTreeWall: If true, creates the Maze by adding\n            walls; if false, creates the Maze by carving passages.\n            Defaults to False.\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" fa""ctor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_38create_prim(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fTreeWall = 0;
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_prim (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fTreeWall,&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[4] = {0,0,0,0};

    /* "daedalus/_maze.pyx":769
 * 
 *     def create_prim(self,
 *                     fTreeWall=False,             # <<<<<<<<<<<<<<
 *                     nEntrancePos=ENTRANCE_RANDOM,
 *                     nRndBias=0,
 */
    values[0] = ((PyObject *)Py_False);
    values[1] = __pyx_k__13;
    values[2] = ((PyObject *)__pyx_int_0);
    values[3] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTreeWall);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_prim") < 0)) __PYX_ERR(0, 768, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fTreeWall = values[0];
    __pyx_v_nEntrancePos = values[1];
    __pyx_v_nRndBias = values[2];
    __pyx_v_nRndRun = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_prim", 0, 0, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 768, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_prim", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_37create_prim(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_fTreeWall, __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* "daedalus/_maze.pyx":768
 *         self._store_back_globals()
 * 
 *     def create_prim(self,             # <<<<<<<<<<<<<<
 *                     fTreeWall=False,
 *                     nEntrancePos=ENTRANCE_RANDOM,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_37create_prim(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("create_prim", 0);

  /* "daedalus/_maze.pyx":820
 *         if not cpp_CreateMazePrim(self._maze,
 *                                   False,
 *                                   fTreeWall,             # <<<<<<<<<<<<<<
 *                                   nEntrancePos,
 *                                   nRndBias,
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_fTreeWall); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 820, __pyx_L1_error)

  /* "daedalus/_maze.pyx":821
 *                                   False,
 *                                   fTreeWall,
 *                                   nEntrancePos,             # <<<<<<<<<<<<<<
 *                                   nRndBias,
 *                                   nRndRun):
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 821, __pyx_L1_error)

  /* "daedalus/_maze.pyx":822
 *                                   fTreeWall,
 *                                   nEntrancePos,
 *                                   nRndBias,             # <<<<<<<<<<<<<<
 *                                   nRndRun):
 *             raise MazeError('Could not create Prim Maze.')
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 822, __pyx_L1_error)

  /* "daedalus/_maze.pyx":823
 *                                   nEntrancePos,
 *                                   nRndBias,
 *                                   nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Prim Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 823, __pyx_L1_error)

  /* "daedalus/_maze.pyx":818
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazePrim(self._maze,             # <<<<<<<<<<<<<<
 *                                   False,
 *                                   fTreeWall,
 */
  __pyx_t_5 = ((!(cpp_CreateMazePrim(__pyx_v_self->_maze, 0, __pyx_t_1, __pyx_t_2, __pyx_t_3, __pyx_t_4) != 0)) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "daedalus/_maze.pyx":824
 *                                   nRndBias,
 *                                   nRndRun):
 *             raise MazeError('Could not create Prim Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 824, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_kp_s_Could_not_create_Prim_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_s_Could_not_create_Prim_Maze);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 824, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 824, __pyx_L1_error)

    /* "daedalus/_maze.pyx":818
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazePrim(self._maze,             # <<<<<<<<<<<<<<
 *                                   False,
 *                                   fTreeWall,
 */
  }

  /* "daedalus/_maze.pyx":825
 *                                   nRndRun):
 *             raise MazeError('Could not create Prim Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_prim2(self,
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "daedalus/_maze.pyx":768
 *         self._store_back_globals()
 * 
 *     def create_prim(self,             # <<<<<<<<<<<<<<
 *                     fTreeWall=False,
 *                     nEntrancePos=ENTRANCE_RANDOM,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_prim", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":827
 *         self._store_back_globals()
 * 
 *     def create_prim2(self,             # <<<<<<<<<<<<<<
 *                      fTreeRandom=True,
 *                      fTreeWall=False,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_40create_prim2(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_39create_prim2[] = "\n        Overwrites the contents of this Maze to create a Prim2 maze.\n\n        Create a new perfect Maze in the bitmap using full or\n        simplified versions of Prim's algorithm. This can carve\n        passages or add walls.\n\n        This is a full implementation of Prim's algorithm involving\n        random edge weights. It will produce a minimum spanning tree\n        like Kruskal's algorithm, and therefore the same random number\n        seed will produce identical Mazes when running these two\n        commands. If fTreeWall is True, then the Maze will be created\n        by adding walls instead of carving passages. If fTreeRandom is\n        True, then all edge weights will be the same instead of\n        uniquely different, resulting in a Maze that's generated\n        faster, however with a lower \"river\" factor, similar to the\n        Prim's Algorithm method.\n\n        :param bool fTreeRandom: If False, the algorithm will tend to\n            create Mazes with long and windy solutions; if True, the\n            solution will tend to be shorter and more direct.\n            Defaults to True.\n        :param bool fTreeWall: If true, creates the Maze by adding\n            walls; if false, creates the Maze by carving passages.\n            Defaults to False.\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles"". When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_40create_prim2(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fTreeRandom = 0;
  PyObject *__pyx_v_fTreeWall = 0;
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_prim2 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fTreeRandom,&__pyx_n_s_fTreeWall,&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[5] = {0,0,0,0,0};

    /* "daedalus/_maze.pyx":828
 * 
 *     def create_prim2(self,
 *                      fTreeRandom=True,             # <<<<<<<<<<<<<<
 *                      fTreeWall=False,
 *                      nEntrancePos=ENTRANCE_RANDOM,
 */
    values[0] = ((PyObject *)Py_True);

    /* "daedalus/_maze.pyx":829
 *     def create_prim2(self,
 *                      fTreeRandom=True,
 *                      fTreeWall=False,             # <<<<<<<<<<<<<<
 *                      nEntrancePos=ENTRANCE_RANDOM,
 *                      nRndBias=0,
 */
    values[1] = ((PyObject *)Py_False);
    values[2] = __pyx_k__14;
    values[3] = ((PyObject *)__pyx_int_0);
    values[4] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTreeRandom);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTreeWall);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_prim2") < 0)) __PYX_ERR(0, 827, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fTreeRandom = values[0];
    __pyx_v_fTreeWall = values[1];
    __pyx_v_nEntrancePos = values[2];
    __pyx_v_nRndBias = values[3];
    __pyx_v_nRndRun = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_prim2", 0, 0, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 827, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_prim2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_39create_prim2(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_fTreeRandom, __pyx_v_fTreeWall, __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* "daedalus/_maze.pyx":827
 *         self._store_back_globals()
 * 
 *     def create_prim2(self,             # <<<<<<<<<<<<<<
 *                      fTreeRandom=True,
 *                      fTreeWall=False,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_39create_prim2(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fTreeRandom, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  bool __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("create_prim2", 0);

  /* "daedalus/_maze.pyx":886
 *         if not cpp_CreateMazePrim2(self._maze,
 *                                    False,
 *                                    fTreeRandom,             # <<<<<<<<<<<<<<
 *                                    fTreeWall,
 *                                    nEntrancePos,
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_fTreeRandom); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 886, __pyx_L1_error)

  /* "daedalus/_maze.pyx":887
 *                                    False,
 *                                    fTreeRandom,
 *                                    fTreeWall,             # <<<<<<<<<<<<<<
 *                                    nEntrancePos,
 *                                    nRndBias,
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_fTreeWall); if (unlikely((__pyx_t_2 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 887, __pyx_L1_error)

  /* "daedalus/_maze.pyx":888
 *                                    fTreeRandom,
 *                                    fTreeWall,
 *                                    nEntrancePos,             # <<<<<<<<<<<<<<
 *                                    nRndBias,
 *                                    nRndRun):
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 888, __pyx_L1_error)

  /* "daedalus/_maze.pyx":889
 *                                    fTreeWall,
 *                                    nEntrancePos,
 *                                    nRndBias,             # <<<<<<<<<<<<<<
 *                                    nRndRun):
 *             raise MazeError('Could not create Prim2 Maze.')
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 889, __pyx_L1_error)

  /* "daedalus/_maze.pyx":890
 *                                    nEntrancePos,
 *                                    nRndBias,
 *                                    nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Prim2 Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 890, __pyx_L1_error)

  /* "daedalus/_maze.pyx":884
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazePrim2(self._maze,             # <<<<<<<<<<<<<<
 *                                    False,
 *                                    fTreeRandom,
 */
  __pyx_t_6 = ((!(cpp_CreateMazePrim2(__pyx_v_self->_maze, 0, __pyx_t_1, __pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5) != 0)) != 0);
  if (unlikely(__pyx_t_6)) {

    /* "daedalus/_maze.pyx":891
 *                                    nRndBias,
 *                                    nRndRun):
 *             raise MazeError('Could not create Prim2 Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 891, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_kp_s_Could_not_create_Prim2_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_kp_s_Could_not_create_Prim2_Maze);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 891, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 891, __pyx_L1_error)

    /* "daedalus/_maze.pyx":884
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazePrim2(self._maze,             # <<<<<<<<<<<<<<
 *                                    False,
 *                                    fTreeRandom,
 */
  }

  /* "daedalus/_maze.pyx":892
 *                                    nRndRun):
 *             raise MazeError('Could not create Prim2 Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_kruskal(self, fClear, c2, c3,
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "daedalus/_maze.pyx":827
 *         self._store_back_globals()
 * 
 *     def create_prim2(self,             # <<<<<<<<<<<<<<
 *                      fTreeRandom=True,
 *                      fTreeWall=False,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_prim2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":894
 *         self._store_back_globals()
 * 
 *     def create_kruskal(self, fClear, c2, c3,             # <<<<<<<<<<<<<<
 *                        fKruskalPic=False,
 *                        fTreeWall=False,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_42create_kruskal(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_41create_kruskal[] = "\n        Overwrites the contents of this Maze to create a Kruskal maze.\n\n        Create a new perfect Maze in the bitmap using Kruskal's\n        algorithm. This can carve passages or add walls.\n\n        This algorithm creates the Maze by carving passages, however\n        it doesn't \"grow\" the Maze like a tree, but rather carves\n        passage segments all over the Maze at random, while still\n        resulting in a perfect Maze when done. This results in Mazes\n        with a low \"river\" factor, but not as low as Prim's\n        algorithm. This method runs slightly slower than Recursive\n        Backtrack.\n\n        :param bool fKruskalPic: defaults to False\n        :param bool fTreeWall: If true, creates the Maze by adding\n            walls; if false, creates the Maze by carving passages.\n            Defaults to False.\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long ""straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_42create_kruskal(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fClear = 0;
  CYTHON_UNUSED PyObject *__pyx_v_c2 = 0;
  CYTHON_UNUSED PyObject *__pyx_v_c3 = 0;
  PyObject *__pyx_v_fKruskalPic = 0;
  PyObject *__pyx_v_fTreeWall = 0;
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_kruskal (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fClear,&__pyx_n_s_c2,&__pyx_n_s_c3,&__pyx_n_s_fKruskalPic,&__pyx_n_s_fTreeWall,&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};

    /* "daedalus/_maze.pyx":895
 * 
 *     def create_kruskal(self, fClear, c2, c3,
 *                        fKruskalPic=False,             # <<<<<<<<<<<<<<
 *                        fTreeWall=False,
 *                        nEntrancePos=ENTRANCE_RANDOM,
 */
    values[3] = ((PyObject *)Py_False);

    /* "daedalus/_maze.pyx":896
 *     def create_kruskal(self, fClear, c2, c3,
 *                        fKruskalPic=False,
 *                        fTreeWall=False,             # <<<<<<<<<<<<<<
 *                        nEntrancePos=ENTRANCE_RANDOM,
 *                        nRndBias=0,
 */
    values[4] = ((PyObject *)Py_False);
    values[5] = __pyx_k__15;
    values[6] = ((PyObject *)__pyx_int_0);
    values[7] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fClear)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_c2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_kruskal", 0, 3, 8, 1); __PYX_ERR(0, 894, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_c3)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_kruskal", 0, 3, 8, 2); __PYX_ERR(0, 894, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fKruskalPic);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTreeWall);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_kruskal") < 0)) __PYX_ERR(0, 894, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fClear = values[0];
    __pyx_v_c2 = values[1];
    __pyx_v_c3 = values[2];
    __pyx_v_fKruskalPic = values[3];
    __pyx_v_fTreeWall = values[4];
    __pyx_v_nEntrancePos = values[5];
    __pyx_v_nRndBias = values[6];
    __pyx_v_nRndRun = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_kruskal", 0, 3, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 894, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_kruskal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_41create_kruskal(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_fClear, __pyx_v_c2, __pyx_v_c3, __pyx_v_fKruskalPic, __pyx_v_fTreeWall, __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* "daedalus/_maze.pyx":894
 *         self._store_back_globals()
 * 
 *     def create_kruskal(self, fClear, c2, c3,             # <<<<<<<<<<<<<<
 *                        fKruskalPic=False,
 *                        fTreeWall=False,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_41create_kruskal(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fClear, CYTHON_UNUSED PyObject *__pyx_v_c2, CYTHON_UNUSED PyObject *__pyx_v_c3, PyObject *__pyx_v_fKruskalPic, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  CCol *__pyx_v_cpp_c2;
  CCol *__pyx_v_cpp_c3;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  bool __pyx_t_2;
  bool __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("create_kruskal", 0);

  /* "daedalus/_maze.pyx":945
 *         '''
 *         # TODO
 *         cdef CCol *cpp_c2 = NULL             # <<<<<<<<<<<<<<
 *         cdef CCol *cpp_c3 = NULL
 *         if not cpp_CreateMazeKruskal(self._maze, fClear, cpp_c2, cpp_c3,
 */
  __pyx_v_cpp_c2 = NULL;

  /* "daedalus/_maze.pyx":946
 *         # TODO
 *         cdef CCol *cpp_c2 = NULL
 *         cdef CCol *cpp_c3 = NULL             # <<<<<<<<<<<<<<
 *         if not cpp_CreateMazeKruskal(self._maze, fClear, cpp_c2, cpp_c3,
 *                                      fKruskalPic,
 */
  __pyx_v_cpp_c3 = NULL;

  /* "daedalus/_maze.pyx":947
 *         cdef CCol *cpp_c2 = NULL
 *         cdef CCol *cpp_c3 = NULL
 *         if not cpp_CreateMazeKruskal(self._maze, fClear, cpp_c2, cpp_c3,             # <<<<<<<<<<<<<<
 *                                      fKruskalPic,
 *                                      False,
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_fClear); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 947, __pyx_L1_error)

  /* "daedalus/_maze.pyx":948
 *         cdef CCol *cpp_c3 = NULL
 *         if not cpp_CreateMazeKruskal(self._maze, fClear, cpp_c2, cpp_c3,
 *                                      fKruskalPic,             # <<<<<<<<<<<<<<
 *                                      False,
 *                                      fTreeWall,
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_fKruskalPic); if (unlikely((__pyx_t_2 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 948, __pyx_L1_error)

  /* "daedalus/_maze.pyx":950
 *                                      fKruskalPic,
 *                                      False,
 *                                      fTreeWall,             # <<<<<<<<<<<<<<
 *                                      nEntrancePos,
 *                                      nRndBias,
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_fTreeWall); if (unlikely((__pyx_t_3 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 950, __pyx_L1_error)

  /* "daedalus/_maze.pyx":951
 *                                      False,
 *                                      fTreeWall,
 *                                      nEntrancePos,             # <<<<<<<<<<<<<<
 *                                      nRndBias,
 *                                      nRndRun):
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 951, __pyx_L1_error)

  /* "daedalus/_maze.pyx":952
 *                                      fTreeWall,
 *                                      nEntrancePos,
 *                                      nRndBias,             # <<<<<<<<<<<<<<
 *                                      nRndRun):
 *             raise MazeError('Could not create Kruskal Maze.')
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 952, __pyx_L1_error)

  /* "daedalus/_maze.pyx":953
 *                                      nEntrancePos,
 *                                      nRndBias,
 *                                      nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Kruskal Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 953, __pyx_L1_error)

  /* "daedalus/_maze.pyx":947
 *         cdef CCol *cpp_c2 = NULL
 *         cdef CCol *cpp_c3 = NULL
 *         if not cpp_CreateMazeKruskal(self._maze, fClear, cpp_c2, cpp_c3,             # <<<<<<<<<<<<<<
 *                                      fKruskalPic,
 *                                      False,
 */
  __pyx_t_7 = ((!(cpp_CreateMazeKruskal(__pyx_v_self->_maze, __pyx_t_1, __pyx_v_cpp_c2, __pyx_v_cpp_c3, __pyx_t_2, 0, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6) != 0)) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "daedalus/_maze.pyx":954
 *                                      nRndBias,
 *                                      nRndRun):
 *             raise MazeError('Could not create Kruskal Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 954, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_kp_s_Could_not_create_Kruskal_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_kp_s_Could_not_create_Kruskal_Maze);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 954, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_Raise(__pyx_t_8, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_ERR(0, 954, __pyx_L1_error)

    /* "daedalus/_maze.pyx":947
 *         cdef CCol *cpp_c2 = NULL
 *         cdef CCol *cpp_c3 = NULL
 *         if not cpp_CreateMazeKruskal(self._maze, fClear, cpp_c2, cpp_c3,             # <<<<<<<<<<<<<<
 *                                      fKruskalPic,
 *                                      False,
 */
  }

  /* "daedalus/_maze.pyx":955
 *                                      nRndRun):
 *             raise MazeError('Could not create Kruskal Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_tree(self,
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "daedalus/_maze.pyx":894
 *         self._store_back_globals()
 * 
 *     def create_kruskal(self, fClear, c2, c3,             # <<<<<<<<<<<<<<
 *                        fKruskalPic=False,
 *                        fTreeWall=False,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_kruskal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":957
 *         self._store_back_globals()
 * 
 *     def create_tree(self,             # <<<<<<<<<<<<<<
 *                     fTreeRandom=True,
 *                     fTreeWall=False,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_44create_tree(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_43create_tree[] = "\n        Overwrites the contents of this Maze to create a Tree maze.\n\n        Create a new perfect Maze in the bitmap using the Growing Tree\n        algorithm.  This can carve passages or add walls.\n\n        This algorithm creates a Maze by continually adding onto what\n        has already been generated. The behaviour of determining what\n        section to add onto next is controlled by the parameters\n        fTreeRandom and nTreeRiver. A Tree Maze can have many dead\n        ends or few dead ends, its solution can be direct or windy,\n        and it can be made by carving passages or adding walls.\n\n        :param bool fTreeRandom: If False, the algorithm will tend to\n            create Mazes with long and windy solutions; if True, the\n            solution will tend to be shorter and more direct.\n            Defaults to True.\n        :param bool fTreeWall: If true, creates the Maze by adding\n            walls; if false, creates the Maze by carving passages.\n            Defaults to False.\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaul""ts to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :param int nTreeRiver: with fTreeRandom set to False, this\n            determines how many of the most recent sections to choose\n            from; higher values give less windy solutions.  With\n            fTreeRandom set to True, higher values of this param give\n            more windy solutions.  Defaults to 10.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_44create_tree(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fTreeRandom = 0;
  PyObject *__pyx_v_fTreeWall = 0;
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_v_nTreeRiver = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_tree (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fTreeRandom,&__pyx_n_s_fTreeWall,&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,&__pyx_n_s_nTreeRiver,0};
    PyObject* values[6] = {0,0,0,0,0,0};

    /* "daedalus/_maze.pyx":958
 * 
 *     def create_tree(self,
 *                     fTreeRandom=True,             # <<<<<<<<<<<<<<
 *                     fTreeWall=False,
 *                     nEntrancePos=ENTRANCE_RANDOM,
 */
    values[0] = ((PyObject *)Py_True);

    /* "daedalus/_maze.pyx":959
 *     def create_tree(self,
 *                     fTreeRandom=True,
 *                     fTreeWall=False,             # <<<<<<<<<<<<<<
 *                     nEntrancePos=ENTRANCE_RANDOM,
 *                     nRndBias=0,
 */
    values[1] = ((PyObject *)Py_False);
    values[2] = __pyx_k__16;
    values[3] = ((PyObject *)__pyx_int_0);
    values[4] = ((PyObject *)__pyx_int_0);
    values[5] = ((PyObject *)__pyx_int_10);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTreeRandom);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTreeWall);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nTreeRiver);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_tree") < 0)) __PYX_ERR(0, 957, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fTreeRandom = values[0];
    __pyx_v_fTreeWall = values[1];
    __pyx_v_nEntrancePos = values[2];
    __pyx_v_nRndBias = values[3];
    __pyx_v_nRndRun = values[4];
    __pyx_v_nTreeRiver = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_tree", 0, 0, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 957, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_tree", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_43create_tree(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_fTreeRandom, __pyx_v_fTreeWall, __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun, __pyx_v_nTreeRiver);

  /* "daedalus/_maze.pyx":957
 *         self._store_back_globals()
 * 
 *     def create_tree(self,             # <<<<<<<<<<<<<<
 *                     fTreeRandom=True,
 *                     fTreeWall=False,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_43create_tree(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fTreeRandom, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun, PyObject *__pyx_v_nTreeRiver) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  bool __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("create_tree", 0);

  /* "daedalus/_maze.pyx":1017
 *         if not cpp_CreateMazeTree(self._maze,
 *                                   False,
 *                                   fTreeRandom,             # <<<<<<<<<<<<<<
 *                                   fTreeWall,
 *                                   nEntrancePos,
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_fTreeRandom); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1017, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1018
 *                                   False,
 *                                   fTreeRandom,
 *                                   fTreeWall,             # <<<<<<<<<<<<<<
 *                                   nEntrancePos,
 *                                   nRndBias,
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_fTreeWall); if (unlikely((__pyx_t_2 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1018, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1019
 *                                   fTreeRandom,
 *                                   fTreeWall,
 *                                   nEntrancePos,             # <<<<<<<<<<<<<<
 *                                   nRndBias,
 *                                   nRndRun,
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1019, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1020
 *                                   fTreeWall,
 *                                   nEntrancePos,
 *                                   nRndBias,             # <<<<<<<<<<<<<<
 *                                   nRndRun,
 *                                   nTreeRiver):
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1020, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1021
 *                                   nEntrancePos,
 *                                   nRndBias,
 *                                   nRndRun,             # <<<<<<<<<<<<<<
 *                                   nTreeRiver):
 *             raise MazeError('Could not create Tree Maze.')
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1021, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1022
 *                                   nRndBias,
 *                                   nRndRun,
 *                                   nTreeRiver):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Tree Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_nTreeRiver); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1022, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1015
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeTree(self._maze,             # <<<<<<<<<<<<<<
 *                                   False,
 *                                   fTreeRandom,
 */
  __pyx_t_7 = ((!(cpp_CreateMazeTree(__pyx_v_self->_maze, 0, __pyx_t_1, __pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6) != 0)) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "daedalus/_maze.pyx":1023
 *                                   nRndRun,
 *                                   nTreeRiver):
 *             raise MazeError('Could not create Tree Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_kp_s_Could_not_create_Tree_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_kp_s_Could_not_create_Tree_Maze);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1023, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_Raise(__pyx_t_8, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_ERR(0, 1023, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1015
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeTree(self._maze,             # <<<<<<<<<<<<<<
 *                                   False,
 *                                   fTreeRandom,
 */
  }

  /* "daedalus/_maze.pyx":1024
 *                                   nTreeRiver):
 *             raise MazeError('Could not create Tree Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_forest(self, fWall,
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1024, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1024, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "daedalus/_maze.pyx":957
 *         self._store_back_globals()
 * 
 *     def create_tree(self,             # <<<<<<<<<<<<<<
 *                     fTreeRandom=True,
 *                     fTreeWall=False,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_tree", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":1026
 *         self._store_back_globals()
 * 
 *     def create_forest(self, fWall,             # <<<<<<<<<<<<<<
 *                       fTreeRandom=True,
 *                       fTreeWall=False,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_46create_forest(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_45create_forest[] = "\n        Overwrites the contents of this Maze to create a Forest maze.\n\n        Create a new perfect Maze in the bitmap using the Growing\n        Forest algorithm.  This can carve passages or add walls.\n\n        The Growing Forest algorithm is basically multiple instances\n        of the Growing Tree algorithm running at the same time.\n\n        :param bool fTreeRandom: If False, the algorithm will tend to\n            create Mazes with long and windy solutions; if True, the\n            solution will tend to be shorter and more direct.\n            Defaults to True.\n        :param bool fTreeWall: If true, creates the Maze by adding\n            walls; if false, creates the Maze by carving passages.\n            Defaults to False.\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nForsAdd: Indicates the number of instances to add\n            each time a cell is added to the Maze: If positive, the\n            field indicates the exact number of instances to add, and,\n            if negative, indicates that one instance should be added\n            every X cells. Defaults to -100.\n        :param int nForsInit: The number of instances to begin with.\n            If positive, the field indicates the exact number of\n            instances, and, if negative, the field indicates that one\n            in X cells should start out as instances. Defaults to 1.\n            If both nForsInit is 1 and nForsAdd is 0, then this\n            function behaves like the Growing Tree algorithm.\n        :param int nRndBias: This allows most ""styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :param int nTreeRiver: with fTreeRandom set to False, this\n            determines how many of the most recent sections to choose\n            from; higher values give less windy solutions.  With\n            fTreeRandom set to True, higher values of this param give\n            more windy solutions.  Defaults to 10.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_46create_forest(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fWall = 0;
  PyObject *__pyx_v_fTreeRandom = 0;
  PyObject *__pyx_v_fTreeWall = 0;
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nForsAdd = 0;
  PyObject *__pyx_v_nForsInit = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_v_nTreeRiver = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_forest (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fWall,&__pyx_n_s_fTreeRandom,&__pyx_n_s_fTreeWall,&__pyx_n_s_nEntrancePos,&__pyx_n_s_nForsAdd,&__pyx_n_s_nForsInit,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,&__pyx_n_s_nTreeRiver,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};

    /* "daedalus/_maze.pyx":1027
 * 
 *     def create_forest(self, fWall,
 *                       fTreeRandom=True,             # <<<<<<<<<<<<<<
 *                       fTreeWall=False,
 *                       nEntrancePos=ENTRANCE_RANDOM,
 */
    values[1] = ((PyObject *)Py_True);

    /* "daedalus/_maze.pyx":1028
 *     def create_forest(self, fWall,
 *                       fTreeRandom=True,
 *                       fTreeWall=False,             # <<<<<<<<<<<<<<
 *                       nEntrancePos=ENTRANCE_RANDOM,
 *                       nForsAdd=-100,
 */
    values[2] = ((PyObject *)Py_False);
    values[3] = __pyx_k__17;
    values[4] = ((PyObject *)__pyx_int_neg_100);
    values[5] = ((PyObject *)__pyx_int_1);
    values[6] = ((PyObject *)__pyx_int_0);
    values[7] = ((PyObject *)__pyx_int_0);
    values[8] = ((PyObject *)__pyx_int_10);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fWall)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTreeRandom);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTreeWall);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nForsAdd);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nForsInit);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nTreeRiver);
          if (value) { values[8] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_forest") < 0)) __PYX_ERR(0, 1026, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fWall = values[0];
    __pyx_v_fTreeRandom = values[1];
    __pyx_v_fTreeWall = values[2];
    __pyx_v_nEntrancePos = values[3];
    __pyx_v_nForsAdd = values[4];
    __pyx_v_nForsInit = values[5];
    __pyx_v_nRndBias = values[6];
    __pyx_v_nRndRun = values[7];
    __pyx_v_nTreeRiver = values[8];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_forest", 0, 1, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1026, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_forest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_45create_forest(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_fWall, __pyx_v_fTreeRandom, __pyx_v_fTreeWall, __pyx_v_nEntrancePos, __pyx_v_nForsAdd, __pyx_v_nForsInit, __pyx_v_nRndBias, __pyx_v_nRndRun, __pyx_v_nTreeRiver);

  /* "daedalus/_maze.pyx":1026
 *         self._store_back_globals()
 * 
 *     def create_forest(self, fWall,             # <<<<<<<<<<<<<<
 *                       fTreeRandom=True,
 *                       fTreeWall=False,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_45create_forest(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fWall, PyObject *__pyx_v_fTreeRandom, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nForsAdd, PyObject *__pyx_v_nForsInit, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun, PyObject *__pyx_v_nTreeRiver) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  bool __pyx_t_2;
  bool __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  __Pyx_RefNannySetupContext("create_forest", 0);

  /* "daedalus/_maze.pyx":1093
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeForest(self._maze, fWall,             # <<<<<<<<<<<<<<
 *                                     True,
 *                                     False,
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_fWall); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1093, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1096
 *                                     True,
 *                                     False,
 *                                     fTreeRandom,             # <<<<<<<<<<<<<<
 *                                     fTreeWall,
 *                                     nEntrancePos,
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_fTreeRandom); if (unlikely((__pyx_t_2 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1096, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1097
 *                                     False,
 *                                     fTreeRandom,
 *                                     fTreeWall,             # <<<<<<<<<<<<<<
 *                                     nEntrancePos,
 *                                     nForsAdd,
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_fTreeWall); if (unlikely((__pyx_t_3 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1097, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1098
 *                                     fTreeRandom,
 *                                     fTreeWall,
 *                                     nEntrancePos,             # <<<<<<<<<<<<<<
 *                                     nForsAdd,
 *                                     nForsInit,
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1098, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1099
 *                                     fTreeWall,
 *                                     nEntrancePos,
 *                                     nForsAdd,             # <<<<<<<<<<<<<<
 *                                     nForsInit,
 *                                     nRndBias,
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_nForsAdd); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1099, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1100
 *                                     nEntrancePos,
 *                                     nForsAdd,
 *                                     nForsInit,             # <<<<<<<<<<<<<<
 *                                     nRndBias,
 *                                     nRndRun,
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_nForsInit); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1100, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1101
 *                                     nForsAdd,
 *                                     nForsInit,
 *                                     nRndBias,             # <<<<<<<<<<<<<<
 *                                     nRndRun,
 *                                     nTreeRiver):
 */
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1101, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1102
 *                                     nForsInit,
 *                                     nRndBias,
 *                                     nRndRun,             # <<<<<<<<<<<<<<
 *                                     nTreeRiver):
 *             raise MazeError('Could not create Forest Maze.')
 */
  __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1102, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1103
 *                                     nRndBias,
 *                                     nRndRun,
 *                                     nTreeRiver):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Forest Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_v_nTreeRiver); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1103, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1093
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeForest(self._maze, fWall,             # <<<<<<<<<<<<<<
 *                                     True,
 *                                     False,
 */
  __pyx_t_10 = ((!(cpp_CreateMazeForest(__pyx_v_self->_maze, __pyx_t_1, 1, 0, __pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9) != 0)) != 0);
  if (unlikely(__pyx_t_10)) {

    /* "daedalus/_maze.pyx":1104
 *                                     nRndRun,
 *                                     nTreeRiver):
 *             raise MazeError('Could not create Forest Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_13 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
      }
    }
    __pyx_t_11 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_13, __pyx_kp_s_Could_not_create_Forest_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_kp_s_Could_not_create_Forest_Maze);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_Raise(__pyx_t_11, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __PYX_ERR(0, 1104, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1093
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeForest(self._maze, fWall,             # <<<<<<<<<<<<<<
 *                                     True,
 *                                     False,
 */
  }

  /* "daedalus/_maze.pyx":1105
 *                                     nTreeRiver):
 *             raise MazeError('Could not create Forest Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_aldous_broder(self,
 */
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
    __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
    if (likely(__pyx_t_13)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
      __Pyx_INCREF(__pyx_t_13);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_12, function);
    }
  }
  __pyx_t_11 = (__pyx_t_13) ? __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_13) : __Pyx_PyObject_CallNoArg(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
  if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "daedalus/_maze.pyx":1026
 *         self._store_back_globals()
 * 
 *     def create_forest(self, fWall,             # <<<<<<<<<<<<<<
 *                       fTreeRandom=True,
 *                       fTreeWall=False,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_forest", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":1107
 *         self._store_back_globals()
 * 
 *     def create_aldous_broder(self,             # <<<<<<<<<<<<<<
 *                              fTreeWall=False,
 *                              nEntrancePos=ENTRANCE_RANDOM,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_48create_aldous_broder(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_47create_aldous_broder[] = "\n        Overwrites the contents of this Maze to create a AldousBroder maze.\n\n        Create a new perfect Maze in the bitmap using the\n        Aldous-Broder algorithm.  This can carve passages or add\n        walls. This is the simplest unbiased algorithm for creating\n        perfect Mazes.\n\n        This algorithm creates Mazes with the special property that\n        all possible Mazes of a given size are generated in equal\n        probability. It carves passages by moving the carving point\n        around totally randomly, relying on chance to finish the Maze,\n        hence this is slow. These Mazes have a low \"river\" factor,\n        only slightly higher than Kruskal's algorithm.\n\n        :param bool fTreeWall: If true, creates the Maze by adding\n            walls; if false, creates the Maze by carving passages.\n            Defaults to False.\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze mean""s\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_48create_aldous_broder(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fTreeWall = 0;
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_aldous_broder (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fTreeWall,&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[4] = {0,0,0,0};

    /* "daedalus/_maze.pyx":1108
 * 
 *     def create_aldous_broder(self,
 *                              fTreeWall=False,             # <<<<<<<<<<<<<<
 *                              nEntrancePos=ENTRANCE_RANDOM,
 *                              nRndBias=0,
 */
    values[0] = ((PyObject *)Py_False);
    values[1] = __pyx_k__18;
    values[2] = ((PyObject *)__pyx_int_0);
    values[3] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTreeWall);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_aldous_broder") < 0)) __PYX_ERR(0, 1107, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fTreeWall = values[0];
    __pyx_v_nEntrancePos = values[1];
    __pyx_v_nRndBias = values[2];
    __pyx_v_nRndRun = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_aldous_broder", 0, 0, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1107, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_aldous_broder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_47create_aldous_broder(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_fTreeWall, __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* "daedalus/_maze.pyx":1107
 *         self._store_back_globals()
 * 
 *     def create_aldous_broder(self,             # <<<<<<<<<<<<<<
 *                              fTreeWall=False,
 *                              nEntrancePos=ENTRANCE_RANDOM,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_47create_aldous_broder(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("create_aldous_broder", 0);

  /* "daedalus/_maze.pyx":1158
 *         if not cpp_CreateMazeAldousBroder(self._maze,
 *                                           False,
 *                                           fTreeWall,             # <<<<<<<<<<<<<<
 *                                           nEntrancePos,
 *                                           nRndBias,
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_fTreeWall); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1158, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1159
 *                                           False,
 *                                           fTreeWall,
 *                                           nEntrancePos,             # <<<<<<<<<<<<<<
 *                                           nRndBias,
 *                                           nRndRun):
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1159, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1160
 *                                           fTreeWall,
 *                                           nEntrancePos,
 *                                           nRndBias,             # <<<<<<<<<<<<<<
 *                                           nRndRun):
 *             raise MazeError('Could not create AldousBroder Maze.')
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1160, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1161
 *                                           nEntrancePos,
 *                                           nRndBias,
 *                                           nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create AldousBroder Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1161, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1156
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeAldousBroder(self._maze,             # <<<<<<<<<<<<<<
 *                                           False,
 *                                           fTreeWall,
 */
  __pyx_t_5 = ((!(cpp_CreateMazeAldousBroder(__pyx_v_self->_maze, 0, __pyx_t_1, __pyx_t_2, __pyx_t_3, __pyx_t_4) != 0)) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "daedalus/_maze.pyx":1162
 *                                           nRndBias,
 *                                           nRndRun):
 *             raise MazeError('Could not create AldousBroder Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_kp_s_Could_not_create_AldousBroder_Ma) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_s_Could_not_create_AldousBroder_Ma);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 1162, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1156
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeAldousBroder(self._maze,             # <<<<<<<<<<<<<<
 *                                           False,
 *                                           fTreeWall,
 */
  }

  /* "daedalus/_maze.pyx":1163
 *                                           nRndRun):
 *             raise MazeError('Could not create AldousBroder Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_wilson(self,
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "daedalus/_maze.pyx":1107
 *         self._store_back_globals()
 * 
 *     def create_aldous_broder(self,             # <<<<<<<<<<<<<<
 *                              fTreeWall=False,
 *                              nEntrancePos=ENTRANCE_RANDOM,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_aldous_broder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":1165
 *         self._store_back_globals()
 * 
 *     def create_wilson(self,             # <<<<<<<<<<<<<<
 *                       fTreeWall=False,
 *                       nEntrancePos=ENTRANCE_RANDOM,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_50create_wilson(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_49create_wilson[] = "\n        Overwrites the contents of this Maze to create a Wilson maze.\n\n        Create a new perfect Maze in the bitmap using Wilson's\n        algorithm. This can carve passages or add walls. Like the\n        Aldous-Broder algorithm, this generates all possible Mazes\n        with equal probability, however this runs about five times\n        faster on average.\n\n        This algorithm is an improved version of Aldous-Broder, in\n        that this also creates all Mazes with equal probability (so\n        Mazes created here are indistinguishable from those created\n        with Aldous-Broder). This carves passages at random too, but\n        jumps to an unfinished location each time the random walk hits\n        a previously carved section, so it quickly attaches passages\n        to the Maze until all parts have been created.\n\n        :param bool fTreeWall: If true, creates the Maze by adding\n            walls; if false, creates the Maze by carving passages.\n            Defaults to False.\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to"" 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_50create_wilson(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fTreeWall = 0;
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_wilson (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fTreeWall,&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[4] = {0,0,0,0};

    /* "daedalus/_maze.pyx":1166
 * 
 *     def create_wilson(self,
 *                       fTreeWall=False,             # <<<<<<<<<<<<<<
 *                       nEntrancePos=ENTRANCE_RANDOM,
 *                       nRndBias=0,
 */
    values[0] = ((PyObject *)Py_False);
    values[1] = __pyx_k__19;
    values[2] = ((PyObject *)__pyx_int_0);
    values[3] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTreeWall);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_wilson") < 0)) __PYX_ERR(0, 1165, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fTreeWall = values[0];
    __pyx_v_nEntrancePos = values[1];
    __pyx_v_nRndBias = values[2];
    __pyx_v_nRndRun = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_wilson", 0, 0, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1165, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_wilson", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_49create_wilson(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_fTreeWall, __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* "daedalus/_maze.pyx":1165
 *         self._store_back_globals()
 * 
 *     def create_wilson(self,             # <<<<<<<<<<<<<<
 *                       fTreeWall=False,
 *                       nEntrancePos=ENTRANCE_RANDOM,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_49create_wilson(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("create_wilson", 0);

  /* "daedalus/_maze.pyx":1218
 *         if not cpp_CreateMazeWilson(self._maze,
 *                                     False,
 *                                     fTreeWall,             # <<<<<<<<<<<<<<
 *                                     nEntrancePos,
 *                                     nRndBias,
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_fTreeWall); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1218, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1219
 *                                     False,
 *                                     fTreeWall,
 *                                     nEntrancePos,             # <<<<<<<<<<<<<<
 *                                     nRndBias,
 *                                     nRndRun):
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1219, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1220
 *                                     fTreeWall,
 *                                     nEntrancePos,
 *                                     nRndBias,             # <<<<<<<<<<<<<<
 *                                     nRndRun):
 *             raise MazeError('Could not create Wilson Maze.')
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1220, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1221
 *                                     nEntrancePos,
 *                                     nRndBias,
 *                                     nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Wilson Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1221, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1216
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeWilson(self._maze,             # <<<<<<<<<<<<<<
 *                                     False,
 *                                     fTreeWall,
 */
  __pyx_t_5 = ((!(cpp_CreateMazeWilson(__pyx_v_self->_maze, 0, __pyx_t_1, __pyx_t_2, __pyx_t_3, __pyx_t_4) != 0)) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "daedalus/_maze.pyx":1222
 *                                     nRndBias,
 *                                     nRndRun):
 *             raise MazeError('Could not create Wilson Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_kp_s_Could_not_create_Wilson_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_s_Could_not_create_Wilson_Maze);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 1222, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1216
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeWilson(self._maze,             # <<<<<<<<<<<<<<
 *                                     False,
 *                                     fTreeWall,
 */
  }

  /* "daedalus/_maze.pyx":1223
 *                                     nRndRun):
 *             raise MazeError('Could not create Wilson Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_eller(self,
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "daedalus/_maze.pyx":1165
 *         self._store_back_globals()
 * 
 *     def create_wilson(self,             # <<<<<<<<<<<<<<
 *                       fTreeWall=False,
 *                       nEntrancePos=ENTRANCE_RANDOM,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_wilson", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":1225
 *         self._store_back_globals()
 * 
 *     def create_eller(self,             # <<<<<<<<<<<<<<
 *                      fTreeWall=False,
 *                      nEntrancePos=ENTRANCE_RANDOM,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_52create_eller(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_51create_eller[] = "\n        Overwrites the contents of this Maze to create a Eller maze.\n\n        Create a new perfect Maze in the bitmap using Eller's\n        algorithm. This can carve passages or add walls. This is the\n        fastest algorithm for creating general perfect Mazes, and runs\n        over twice as fast as any of the others.\n\n        This algorithm creates Mazes by focusing on one row at a time,\n        keeping track of which paths connect so as to know where to\n        carve and not carve passages. These Mazes have a lowish\n        \"river\" factor, slightly higher than the Aldous-Broder\n        algorithm.\n\n        :param bool fTreeWall: If true, creates the Maze by adding\n            walls; if false, creates the Maze by carving passages.\n            Defaults to False.\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n        ""    present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_52create_eller(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fTreeWall = 0;
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_eller (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fTreeWall,&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[4] = {0,0,0,0};

    /* "daedalus/_maze.pyx":1226
 * 
 *     def create_eller(self,
 *                      fTreeWall=False,             # <<<<<<<<<<<<<<
 *                      nEntrancePos=ENTRANCE_RANDOM,
 *                      nRndBias=0,
 */
    values[0] = ((PyObject *)Py_False);
    values[1] = __pyx_k__20;
    values[2] = ((PyObject *)__pyx_int_0);
    values[3] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTreeWall);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_eller") < 0)) __PYX_ERR(0, 1225, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fTreeWall = values[0];
    __pyx_v_nEntrancePos = values[1];
    __pyx_v_nRndBias = values[2];
    __pyx_v_nRndRun = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_eller", 0, 0, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1225, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_eller", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_51create_eller(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_fTreeWall, __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* "daedalus/_maze.pyx":1225
 *         self._store_back_globals()
 * 
 *     def create_eller(self,             # <<<<<<<<<<<<<<
 *                      fTreeWall=False,
 *                      nEntrancePos=ENTRANCE_RANDOM,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_51create_eller(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("create_eller", 0);

  /* "daedalus/_maze.pyx":1275
 *         if not cpp_CreateMazeEller(self._maze,
 *                                    False,
 *                                    fTreeWall,             # <<<<<<<<<<<<<<
 *                                    nEntrancePos,
 *                                    nRndBias,
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_fTreeWall); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1275, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1276
 *                                    False,
 *                                    fTreeWall,
 *                                    nEntrancePos,             # <<<<<<<<<<<<<<
 *                                    nRndBias,
 *                                    nRndRun):
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1276, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1277
 *                                    fTreeWall,
 *                                    nEntrancePos,
 *                                    nRndBias,             # <<<<<<<<<<<<<<
 *                                    nRndRun):
 *             raise MazeError('Could not create Eller Maze.')
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1277, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1278
 *                                    nEntrancePos,
 *                                    nRndBias,
 *                                    nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Eller Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1278, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1273
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeEller(self._maze,             # <<<<<<<<<<<<<<
 *                                    False,
 *                                    fTreeWall,
 */
  __pyx_t_5 = ((!(cpp_CreateMazeEller(__pyx_v_self->_maze, 0, __pyx_t_1, __pyx_t_2, __pyx_t_3, __pyx_t_4) != 0)) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "daedalus/_maze.pyx":1279
 *                                    nRndBias,
 *                                    nRndRun):
 *             raise MazeError('Could not create Eller Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_kp_s_Could_not_create_Eller_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_s_Could_not_create_Eller_Maze);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 1279, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1273
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeEller(self._maze,             # <<<<<<<<<<<<<<
 *                                    False,
 *                                    fTreeWall,
 */
  }

  /* "daedalus/_maze.pyx":1280
 *                                    nRndRun):
 *             raise MazeError('Could not create Eller Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_braid_eller(self,
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "daedalus/_maze.pyx":1225
 *         self._store_back_globals()
 * 
 *     def create_eller(self,             # <<<<<<<<<<<<<<
 *                      fTreeWall=False,
 *                      nEntrancePos=ENTRANCE_RANDOM,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_eller", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":1282
 *         self._store_back_globals()
 * 
 *     def create_braid_eller(self,             # <<<<<<<<<<<<<<
 *                            nEntrancePos=ENTRANCE_RANDOM,
 *                            nRndBias=0,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_54create_braid_eller(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_53create_braid_eller[] = "\n        Overwrites the contents of this Maze to create a BraidEller maze.\n\n        Create a new braid Maze in a bitmap using a variation of\n        Eller's Algorithm.\n\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_54create_braid_eller(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_braid_eller (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[3] = {0,0,0};
    values[0] = __pyx_k__21;
    values[1] = ((PyObject *)__pyx_int_0);
    values[2] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_braid_eller") < 0)) __PYX_ERR(0, 1282, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_nEntrancePos = values[0];
    __pyx_v_nRndBias = values[1];
    __pyx_v_nRndRun = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_braid_eller", 0, 0, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1282, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_braid_eller", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_53create_braid_eller(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_53create_braid_eller(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("create_braid_eller", 0);

  /* "daedalus/_maze.pyx":1320
 *         if not cpp_CreateMazeBraidEller(self._maze,
 *                                         False,
 *                                         nEntrancePos,             # <<<<<<<<<<<<<<
 *                                         nRndBias,
 *                                         nRndRun):
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1320, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1321
 *                                         False,
 *                                         nEntrancePos,
 *                                         nRndBias,             # <<<<<<<<<<<<<<
 *                                         nRndRun):
 *             raise MazeError('Could not create BraidEller Maze.')
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1321, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1322
 *                                         nEntrancePos,
 *                                         nRndBias,
 *                                         nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create BraidEller Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1322, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1318
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeBraidEller(self._maze,             # <<<<<<<<<<<<<<
 *                                         False,
 *                                         nEntrancePos,
 */
  __pyx_t_4 = ((!(cpp_CreateMazeBraidEller(__pyx_v_self->_maze, 0, __pyx_t_1, __pyx_t_2, __pyx_t_3) != 0)) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "daedalus/_maze.pyx":1323
 *                                         nRndBias,
 *                                         nRndRun):
 *             raise MazeError('Could not create BraidEller Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_kp_s_Could_not_create_BraidEller_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_s_Could_not_create_BraidEller_Maze);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1323, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 1323, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1318
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeBraidEller(self._maze,             # <<<<<<<<<<<<<<
 *                                         False,
 *                                         nEntrancePos,
 */
  }

  /* "daedalus/_maze.pyx":1324
 *                                         nRndRun):
 *             raise MazeError('Could not create BraidEller Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_division(self,
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "daedalus/_maze.pyx":1282
 *         self._store_back_globals()
 * 
 *     def create_braid_eller(self,             # <<<<<<<<<<<<<<
 *                            nEntrancePos=ENTRANCE_RANDOM,
 *                            nRndBias=0,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_braid_eller", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":1326
 *         self._store_back_globals()
 * 
 *     def create_division(self,             # <<<<<<<<<<<<<<
 *                         nEntrancePos=ENTRANCE_RANDOM,
 *                         nRndBias=0,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_56create_division(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_55create_division[] = "\n        Overwrites the contents of this Maze to create a Division maze.\n\n        Create a new perfect Maze in the bitmap using recursive\n        division. This always adds walls, recursively dividing the\n        Maze into smaller rectangles.\n\n        The algorithm creates a Maze by adding walls, where the area\n        within the Maze is divided by a randomly positioned horizontal\n        or vertical wall with a passage opening within it. Each\n        subarea is then recursively divided with more walls until all\n        areas are filled.\n\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across"" a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_56create_division(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_division (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[3] = {0,0,0};
    values[0] = __pyx_k__22;
    values[1] = ((PyObject *)__pyx_int_0);
    values[2] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_division") < 0)) __PYX_ERR(0, 1326, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_nEntrancePos = values[0];
    __pyx_v_nRndBias = values[1];
    __pyx_v_nRndRun = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_division", 0, 0, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1326, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_division", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_55create_division(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_55create_division(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("create_division", 0);

  /* "daedalus/_maze.pyx":1371
 *         if not cpp_CreateMazeDivision(self._maze,
 *                                       False,
 *                                       nEntrancePos,             # <<<<<<<<<<<<<<
 *                                       nRndBias,
 *                                       nRndRun):
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1371, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1372
 *                                       False,
 *                                       nEntrancePos,
 *                                       nRndBias,             # <<<<<<<<<<<<<<
 *                                       nRndRun):
 *             raise MazeError('Could not create Division Maze.')
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1372, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1373
 *                                       nEntrancePos,
 *                                       nRndBias,
 *                                       nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Division Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1373, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1369
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeDivision(self._maze,             # <<<<<<<<<<<<<<
 *                                       False,
 *                                       nEntrancePos,
 */
  __pyx_t_4 = ((!(cpp_CreateMazeDivision(__pyx_v_self->_maze, 0, __pyx_t_1, __pyx_t_2, __pyx_t_3) != 0)) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "daedalus/_maze.pyx":1374
 *                                       nRndBias,
 *                                       nRndRun):
 *             raise MazeError('Could not create Division Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_7, __pyx_kp_s_Could_not_create_Division_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_s_Could_not_create_Division_Maze);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 1374, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1369
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeDivision(self._maze,             # <<<<<<<<<<<<<<
 *                                       False,
 *                                       nEntrancePos,
 */
  }

  /* "daedalus/_maze.pyx":1375
 *                                       nRndRun):
 *             raise MazeError('Could not create Division Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_binary(self,
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "daedalus/_maze.pyx":1326
 *         self._store_back_globals()
 * 
 *     def create_division(self,             # <<<<<<<<<<<<<<
 *                         nEntrancePos=ENTRANCE_RANDOM,
 *                         nRndBias=0,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_division", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":1377
 *         self._store_back_globals()
 * 
 *     def create_binary(self,             # <<<<<<<<<<<<<<
 *                       cRandomAdd=0,
 *                       fTreeWall=False,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_58create_binary(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_57create_binary[] = "\n        Overwrites the contents of this Maze to create a Binary maze.\n\n        Create a new perfect Maze in the bitmap using the Binary Tree\n        algorithm.  This can carve passages or add walls. This is the\n        simplest algorithm of any type for creating perfect\n        Mazes.\n\n        The algorithm creates Mazes with the special property that\n        each cell has a passage that leads either up or left, but\n        never both or neither. This creates a biased texture, where\n        one can always easily travel diagonally up and to the left\n        without hitting barriers or having to make choices. Moving\n        down and to the right is when the Maze becomes a challenge, so\n        the Maze is hard to solve but easy to solve backwards. The\n        Maze forms a binary tree, with the upper left corner the root,\n        where each node or cell has one or two children, and one\n        unique parent which is the cell above or to the left of it.\n\n        :param int cRandomAdd: The number of random walls and passages\n            to add to the maze.  A high number will make the maze more\n            random and less like its base design, while a large negative\n            number will make the maze be its base design with longer\n            passages and dead ends.  Defaults to 0.\n        :param bool fTreeWall: If true, creates the Maze by adding\n            walls; if false, creates the Maze by carving passages.\n            Defaults to False.\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This a""llows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_58create_binary(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cRandomAdd = 0;
  PyObject *__pyx_v_fTreeWall = 0;
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_binary (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cRandomAdd,&__pyx_n_s_fTreeWall,&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[0] = ((PyObject *)__pyx_int_0);

    /* "daedalus/_maze.pyx":1379
 *     def create_binary(self,
 *                       cRandomAdd=0,
 *                       fTreeWall=False,             # <<<<<<<<<<<<<<
 *                       nEntrancePos=ENTRANCE_RANDOM,
 *                       nRndBias=0,
 */
    values[1] = ((PyObject *)Py_False);
    values[2] = __pyx_k__23;
    values[3] = ((PyObject *)__pyx_int_0);
    values[4] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cRandomAdd);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTreeWall);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_binary") < 0)) __PYX_ERR(0, 1377, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cRandomAdd = values[0];
    __pyx_v_fTreeWall = values[1];
    __pyx_v_nEntrancePos = values[2];
    __pyx_v_nRndBias = values[3];
    __pyx_v_nRndRun = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_binary", 0, 0, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1377, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_binary", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_57create_binary(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_cRandomAdd, __pyx_v_fTreeWall, __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* "daedalus/_maze.pyx":1377
 *         self._store_back_globals()
 * 
 *     def create_binary(self,             # <<<<<<<<<<<<<<
 *                       cRandomAdd=0,
 *                       fTreeWall=False,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_57create_binary(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_cRandomAdd, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  bool __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("create_binary", 0);

  /* "daedalus/_maze.pyx":1437
 *         '''
 *         if not cpp_CreateMazeBinary(self._maze,
 *                                     cRandomAdd,             # <<<<<<<<<<<<<<
 *                                     False,
 *                                     fTreeWall,
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_cRandomAdd); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1437, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1439
 *                                     cRandomAdd,
 *                                     False,
 *                                     fTreeWall,             # <<<<<<<<<<<<<<
 *                                     nEntrancePos,
 *                                     nRndBias,
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_fTreeWall); if (unlikely((__pyx_t_2 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1439, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1440
 *                                     False,
 *                                     fTreeWall,
 *                                     nEntrancePos,             # <<<<<<<<<<<<<<
 *                                     nRndBias,
 *                                     nRndRun):
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1440, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1441
 *                                     fTreeWall,
 *                                     nEntrancePos,
 *                                     nRndBias,             # <<<<<<<<<<<<<<
 *                                     nRndRun):
 *             raise MazeError('Could not create Binary Maze.')
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1441, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1442
 *                                     nEntrancePos,
 *                                     nRndBias,
 *                                     nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Binary Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1442, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1436
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeBinary(self._maze,             # <<<<<<<<<<<<<<
 *                                     cRandomAdd,
 *                                     False,
 */
  __pyx_t_6 = ((!(cpp_CreateMazeBinary(__pyx_v_self->_maze, __pyx_t_1, 0, __pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5) != 0)) != 0);
  if (unlikely(__pyx_t_6)) {

    /* "daedalus/_maze.pyx":1443
 *                                     nRndBias,
 *                                     nRndRun):
 *             raise MazeError('Could not create Binary Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1443, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_kp_s_Could_not_create_Binary_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_kp_s_Could_not_create_Binary_Maze);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1443, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 1443, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1436
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeBinary(self._maze,             # <<<<<<<<<<<<<<
 *                                     cRandomAdd,
 *                                     False,
 */
  }

  /* "daedalus/_maze.pyx":1444
 *                                     nRndRun):
 *             raise MazeError('Could not create Binary Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_sidewinder(self,
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "daedalus/_maze.pyx":1377
 *         self._store_back_globals()
 * 
 *     def create_binary(self,             # <<<<<<<<<<<<<<
 *                       cRandomAdd=0,
 *                       fTreeWall=False,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_binary", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":1446
 *         self._store_back_globals()
 * 
 *     def create_sidewinder(self,             # <<<<<<<<<<<<<<
 *                           fTreeWall=False,
 *                           nEntrancePos=ENTRANCE_RANDOM,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_60create_sidewinder(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_59create_sidewinder[] = "\n        Overwrites the contents of this Maze to create a Sidewinder maze.\n\n        Create a new perfect Maze in the bitmap using the Sidewinder\n        algorithm.  This can carve passages or add walls.\n\n        This algorithm creates a Maze with the property that every\n        horizontal passage has exactly one passage leading up from it,\n        resulting in a Maze with no backtracking passages where the\n        solution path just snakes back and forth from entrance to\n        exit. The Maze is just as hard to solve forwards, however it's\n        easy to solve backwards since there's always exactly one way\n        to get to the previous row.\n\n        :param bool fTreeWall: If true, creates the Maze by adding\n            walls; if false, creates the Maze by carving passages.\n            Defaults to False.\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" factor. The run of a Maze means\n            how long straightaways t""end to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_60create_sidewinder(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fTreeWall = 0;
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_sidewinder (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fTreeWall,&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[4] = {0,0,0,0};

    /* "daedalus/_maze.pyx":1447
 * 
 *     def create_sidewinder(self,
 *                           fTreeWall=False,             # <<<<<<<<<<<<<<
 *                           nEntrancePos=ENTRANCE_RANDOM,
 *                           nRndBias=0,
 */
    values[0] = ((PyObject *)Py_False);
    values[1] = __pyx_k__24;
    values[2] = ((PyObject *)__pyx_int_0);
    values[3] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTreeWall);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_sidewinder") < 0)) __PYX_ERR(0, 1446, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fTreeWall = values[0];
    __pyx_v_nEntrancePos = values[1];
    __pyx_v_nRndBias = values[2];
    __pyx_v_nRndRun = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_sidewinder", 0, 0, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1446, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_sidewinder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_59create_sidewinder(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_fTreeWall, __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* "daedalus/_maze.pyx":1446
 *         self._store_back_globals()
 * 
 *     def create_sidewinder(self,             # <<<<<<<<<<<<<<
 *                           fTreeWall=False,
 *                           nEntrancePos=ENTRANCE_RANDOM,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_59create_sidewinder(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("create_sidewinder", 0);

  /* "daedalus/_maze.pyx":1496
 *         if not cpp_CreateMazeSidewinder(self._maze,
 *                                         False,
 *                                         fTreeWall,             # <<<<<<<<<<<<<<
 *                                         nEntrancePos,
 *                                         nRndBias,
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_fTreeWall); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1496, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1497
 *                                         False,
 *                                         fTreeWall,
 *                                         nEntrancePos,             # <<<<<<<<<<<<<<
 *                                         nRndBias,
 *                                         nRndRun):
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1497, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1498
 *                                         fTreeWall,
 *                                         nEntrancePos,
 *                                         nRndBias,             # <<<<<<<<<<<<<<
 *                                         nRndRun):
 *             raise MazeError('Could not create Sidewinder Maze.')
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1498, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1499
 *                                         nEntrancePos,
 *                                         nRndBias,
 *                                         nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Sidewinder Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1499, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1494
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeSidewinder(self._maze,             # <<<<<<<<<<<<<<
 *                                         False,
 *                                         fTreeWall,
 */
  __pyx_t_5 = ((!(cpp_CreateMazeSidewinder(__pyx_v_self->_maze, 0, __pyx_t_1, __pyx_t_2, __pyx_t_3, __pyx_t_4) != 0)) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "daedalus/_maze.pyx":1500
 *                                         nRndBias,
 *                                         nRndRun):
 *             raise MazeError('Could not create Sidewinder Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_kp_s_Could_not_create_Sidewinder_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_s_Could_not_create_Sidewinder_Maze);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 1500, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1494
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeSidewinder(self._maze,             # <<<<<<<<<<<<<<
 *                                         False,
 *                                         fTreeWall,
 */
  }

  /* "daedalus/_maze.pyx":1501
 *                                         nRndRun):
 *             raise MazeError('Could not create Sidewinder Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def create_unicursal(self,
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
    }
  }
  __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "daedalus/_maze.pyx":1446
 *         self._store_back_globals()
 * 
 *     def create_sidewinder(self,             # <<<<<<<<<<<<<<
 *                           fTreeWall=False,
 *                           nEntrancePos=ENTRANCE_RANDOM,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_sidewinder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":1503
 *         self._store_back_globals()
 * 
 *     def create_unicursal(self,             # <<<<<<<<<<<<<<
 *                          cRandomAdd=0,
 *                          fRiver=True,
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_62create_unicursal(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_61create_unicursal[] = "\n        Overwrites the contents of this Maze to create a Unicursal maze.\n\n        Create a new Unicursal Maze in the bitmap, i.e. a Maze without\n        junctions.\n\n        :param int cRandomAdd: The number of random walls and passages\n            to add to the maze.  A high number will make the maze more\n            random and less like its base design, while a large negative\n            number will make the maze be its base design with longer\n            passages and dead ends.  Defaults to 0.\n        :param bool fRiver: If set, creates a Maze with a relatively\n            high \"river\" factor (with relatively longer passages between\n            junctions and fewer but longer dead ends).  Defaults to True.\n        :param bool fTreeWall: If true, creates the Maze by adding\n            walls; if false, creates the Maze by carving passages.\n            Defaults to False.\n        :param int nEntrancePos: One of ENTRANCE_CORNER (entrance is\n            in the upper-left corner, and the exit in the lower-right),\n            ENTRANCE_MIDDLE (entrance and exit are in the middle of the\n            top and bottom edges), ENTRANCE_BALANCED (entrance and exit\n            are placed randomly, but are horizontally \"balanced\"),\n            ENTRANCE_RANDOM (entrance and exit are placed randomly).\n            Defaults to ENTRANCE_RANDOM.\n        :param int nRndBias: This allows most styles of random Mazes\n            created to be passage biased. A biased Maze has\n            straightaways that tend to go along one axis more often\n            than at other angles. When this value is positive, Mazes\n            will be horizontally biased, and when this is negative,\n            Mazes will be vertically biased. The farther the value\n            from zero, the more extreme the bias. A value of zero\n            means no bias.  Defaults to 0.\n        :param int nRndRun: This allows some styles of random Mazes\n            created to have a \"run\" fa""ctor. The run of a Maze means\n            how long straightaways tend to go before forced turnings\n            present themselves. A Maze with a low run won't have\n            straight passages for more than three or four cells, and\n            will look very random. A Maze with a high run will have\n            long passages going across a good percentage of the Maze,\n            and will look similar to a microchip. The more this value\n            is greater than zero, the higher the run.  Defaults to 0.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_62create_unicursal(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cRandomAdd = 0;
  PyObject *__pyx_v_fRiver = 0;
  PyObject *__pyx_v_fTreeWall = 0;
  PyObject *__pyx_v_nEntrancePos = 0;
  PyObject *__pyx_v_nRndBias = 0;
  PyObject *__pyx_v_nRndRun = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_unicursal (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cRandomAdd,&__pyx_n_s_fRiver,&__pyx_n_s_fTreeWall,&__pyx_n_s_nEntrancePos,&__pyx_n_s_nRndBias,&__pyx_n_s_nRndRun,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    values[0] = ((PyObject *)__pyx_int_0);

    /* "daedalus/_maze.pyx":1505
 *     def create_unicursal(self,
 *                          cRandomAdd=0,
 *                          fRiver=True,             # <<<<<<<<<<<<<<
 *                          fTreeWall=False,
 *                          nEntrancePos=ENTRANCE_RANDOM,
 */
    values[1] = ((PyObject *)Py_True);

    /* "daedalus/_maze.pyx":1506
 *                          cRandomAdd=0,
 *                          fRiver=True,
 *                          fTreeWall=False,             # <<<<<<<<<<<<<<
 *                          nEntrancePos=ENTRANCE_RANDOM,
 *                          nRndBias=0,
 */
    values[2] = ((PyObject *)Py_False);
    values[3] = __pyx_k__25;
    values[4] = ((PyObject *)__pyx_int_0);
    values[5] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cRandomAdd);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fRiver);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTreeWall);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nEntrancePos);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndBias);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nRndRun);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_unicursal") < 0)) __PYX_ERR(0, 1503, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cRandomAdd = values[0];
    __pyx_v_fRiver = values[1];
    __pyx_v_fTreeWall = values[2];
    __pyx_v_nEntrancePos = values[3];
    __pyx_v_nRndBias = values[4];
    __pyx_v_nRndRun = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_unicursal", 0, 0, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1503, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.create_unicursal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_61create_unicursal(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_cRandomAdd, __pyx_v_fRiver, __pyx_v_fTreeWall, __pyx_v_nEntrancePos, __pyx_v_nRndBias, __pyx_v_nRndRun);

  /* "daedalus/_maze.pyx":1503
 *         self._store_back_globals()
 * 
 *     def create_unicursal(self,             # <<<<<<<<<<<<<<
 *                          cRandomAdd=0,
 *                          fRiver=True,
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_61create_unicursal(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_cRandomAdd, PyObject *__pyx_v_fRiver, PyObject *__pyx_v_fTreeWall, PyObject *__pyx_v_nEntrancePos, PyObject *__pyx_v_nRndBias, PyObject *__pyx_v_nRndRun) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  bool __pyx_t_2;
  bool __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("create_unicursal", 0);

  /* "daedalus/_maze.pyx":1554
 *         '''
 *         if not cpp_CreateMazeUnicursal(self._maze,
 *                                        cRandomAdd,             # <<<<<<<<<<<<<<
 *                                        fRiver,
 *                                        True,
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_cRandomAdd); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1554, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1555
 *         if not cpp_CreateMazeUnicursal(self._maze,
 *                                        cRandomAdd,
 *                                        fRiver,             # <<<<<<<<<<<<<<
 *                                        True,
 *                                        True,
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_fRiver); if (unlikely((__pyx_t_2 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1555, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1559
 *                                        True,
 *                                        False,
 *                                        fTreeWall,             # <<<<<<<<<<<<<<
 *                                        nEntrancePos,
 *                                        nRndBias,
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_fTreeWall); if (unlikely((__pyx_t_3 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1559, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1560
 *                                        False,
 *                                        fTreeWall,
 *                                        nEntrancePos,             # <<<<<<<<<<<<<<
 *                                        nRndBias,
 *                                        nRndRun):
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_nEntrancePos); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1560, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1561
 *                                        fTreeWall,
 *                                        nEntrancePos,
 *                                        nRndBias,             # <<<<<<<<<<<<<<
 *                                        nRndRun):
 *             raise MazeError('Could not create Unicursal Maze.')
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_nRndBias); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1561, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1562
 *                                        nEntrancePos,
 *                                        nRndBias,
 *                                        nRndRun):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not create Unicursal Maze.')
 *         self._store_back_globals()
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_nRndRun); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1562, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1553
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeUnicursal(self._maze,             # <<<<<<<<<<<<<<
 *                                        cRandomAdd,
 *                                        fRiver,
 */
  __pyx_t_7 = ((!(cpp_CreateMazeUnicursal(__pyx_v_self->_maze, __pyx_t_1, __pyx_t_2, 1, 1, 0, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6) != 0)) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "daedalus/_maze.pyx":1563
 *                                        nRndBias,
 *                                        nRndRun):
 *             raise MazeError('Could not create Unicursal Maze.')             # <<<<<<<<<<<<<<
 *         self._store_back_globals()
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_kp_s_Could_not_create_Unicursal_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_kp_s_Could_not_create_Unicursal_Maze);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_Raise(__pyx_t_8, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_ERR(0, 1563, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1553
 *         :rtype: None
 *         '''
 *         if not cpp_CreateMazeUnicursal(self._maze,             # <<<<<<<<<<<<<<
 *                                        cRandomAdd,
 *                                        fRiver,
 */
  }

  /* "daedalus/_maze.pyx":1564
 *                                        nRndRun):
 *             raise MazeError('Could not create Unicursal Maze.')
 *         self._store_back_globals()             # <<<<<<<<<<<<<<
 * 
 *     def resize(self, width, height):
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_store_back_globals); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "daedalus/_maze.pyx":1503
 *         self._store_back_globals()
 * 
 *     def create_unicursal(self,             # <<<<<<<<<<<<<<
 *                          cRandomAdd=0,
 *                          fRiver=True,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("daedalus._maze.Maze.create_unicursal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":1566
 *         self._store_back_globals()
 * 
 *     def resize(self, width, height):             # <<<<<<<<<<<<<<
 *         '''
 *         Truncates or pads out this Maze to attain the given size.
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_64resize(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_63resize[] = "\n        Truncates or pads out this Maze to attain the given size.\n\n        :param int width: Odd integer, 3 or greater\n        :param int height: Odd integer, 3 or greater\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_64resize(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_width = 0;
  PyObject *__pyx_v_height = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("resize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_width,&__pyx_n_s_height,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_width)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_height)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("resize", 1, 2, 2, 1); __PYX_ERR(0, 1566, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "resize") < 0)) __PYX_ERR(0, 1566, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_width = values[0];
    __pyx_v_height = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("resize", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1566, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.resize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_63resize(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_width, __pyx_v_height);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_63resize(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_width, PyObject *__pyx_v_height) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("resize", 0);

  /* "daedalus/_maze.pyx":1574
 *         :rtype: None
 *         '''
 *         if width < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError('negative width')
 *         if height < 0:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_width, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1574, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1574, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "daedalus/_maze.pyx":1575
 *         '''
 *         if width < 0:
 *             raise ValueError('negative width')             # <<<<<<<<<<<<<<
 *         if height < 0:
 *             raise ValueError('negative height')
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1575, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1575, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1574
 *         :rtype: None
 *         '''
 *         if width < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError('negative width')
 *         if height < 0:
 */
  }

  /* "daedalus/_maze.pyx":1576
 *         if width < 0:
 *             raise ValueError('negative width')
 *         if height < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError('negative height')
 *         # if width % 2 != 1:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_height, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1576, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1576, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "daedalus/_maze.pyx":1577
 *             raise ValueError('negative width')
 *         if height < 0:
 *             raise ValueError('negative height')             # <<<<<<<<<<<<<<
 *         # if width % 2 != 1:
 *         #     raise ValueError('width must be odd')
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1577, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1576
 *         if width < 0:
 *             raise ValueError('negative width')
 *         if height < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError('negative height')
 *         # if width % 2 != 1:
 */
  }

  /* "daedalus/_maze.pyx":1582
 *         # if height % 2 != 1:
 *         #     raise ValueError('height must be odd')
 *         if not cpp_Resize(self._maze, width, height):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not resize the Maze.')
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_As_unsigned_int(__pyx_v_width); if (unlikely((__pyx_t_3 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1582, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyInt_As_unsigned_int(__pyx_v_height); if (unlikely((__pyx_t_4 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1582, __pyx_L1_error)
  __pyx_t_2 = ((!(cpp_Resize(__pyx_v_self->_maze, __pyx_t_3, __pyx_t_4) != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "daedalus/_maze.pyx":1583
 *         #     raise ValueError('height must be odd')
 *         if not cpp_Resize(self._maze, width, height):
 *             raise MazeError('Could not resize the Maze.')             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1583, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_kp_s_Could_not_resize_the_Maze) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_s_Could_not_resize_the_Maze);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1583, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1583, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1582
 *         # if height % 2 != 1:
 *         #     raise ValueError('height must be odd')
 *         if not cpp_Resize(self._maze, width, height):             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not resize the Maze.')
 * 
 */
  }

  /* "daedalus/_maze.pyx":1566
 *         self._store_back_globals()
 * 
 *     def resize(self, width, height):             # <<<<<<<<<<<<<<
 *         '''
 *         Truncates or pads out this Maze to attain the given size.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("daedalus._maze.Maze.resize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":1586
 * 
 *     @staticmethod
 *     def _handle_save_retval(rv):             # <<<<<<<<<<<<<<
 *         '''
 *         Helper method to handle return values from the Save* methods.
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_66_handle_save_retval(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_65_handle_save_retval[] = "\n        Helper method to handle return values from the Save* methods.\n\n        :param int rv: the value returned from the Save* method.\n        :rtype: None\n        ";
static PyMethodDef __pyx_mdef_8daedalus_5_maze_4Maze_66_handle_save_retval = {"_handle_save_retval", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_66_handle_save_retval, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_65_handle_save_retval};
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_66_handle_save_retval(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_rv = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_handle_save_retval (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_rv,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rv)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_handle_save_retval") < 0)) __PYX_ERR(0, 1586, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_rv = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_handle_save_retval", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1586, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze._handle_save_retval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_65_handle_save_retval(__pyx_v_rv);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_65_handle_save_retval(PyObject *__pyx_v_rv) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_handle_save_retval", 0);

  /* "daedalus/_maze.pyx":1593
 *         :rtype: None
 *         '''
 *         if rv == SAVE_NO_MAZE_ERROR:             # <<<<<<<<<<<<<<
 *             raise MazeError('Maze not properly constructed.')
 *         elif rv == SAVE_NO_FILENAME_GIVEN_ERROR:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(SAVE_NO_MAZE_ERROR); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_rv, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1593, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1593, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "daedalus/_maze.pyx":1594
 *         '''
 *         if rv == SAVE_NO_MAZE_ERROR:
 *             raise MazeError('Maze not properly constructed.')             # <<<<<<<<<<<<<<
 *         elif rv == SAVE_NO_FILENAME_GIVEN_ERROR:
 *             raise MazeError('No filename given.')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1594, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_kp_s_Maze_not_properly_constructed) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_s_Maze_not_properly_constructed);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1594, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1594, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1593
 *         :rtype: None
 *         '''
 *         if rv == SAVE_NO_MAZE_ERROR:             # <<<<<<<<<<<<<<
 *             raise MazeError('Maze not properly constructed.')
 *         elif rv == SAVE_NO_FILENAME_GIVEN_ERROR:
 */
  }

  /* "daedalus/_maze.pyx":1595
 *         if rv == SAVE_NO_MAZE_ERROR:
 *             raise MazeError('Maze not properly constructed.')
 *         elif rv == SAVE_NO_FILENAME_GIVEN_ERROR:             # <<<<<<<<<<<<<<
 *             raise MazeError('No filename given.')
 *         elif rv == SAVE_FILE_OPEN_ERROR:
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(SAVE_NO_FILENAME_GIVEN_ERROR); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_rv, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1595, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1595, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "daedalus/_maze.pyx":1596
 *             raise MazeError('Maze not properly constructed.')
 *         elif rv == SAVE_NO_FILENAME_GIVEN_ERROR:
 *             raise MazeError('No filename given.')             # <<<<<<<<<<<<<<
 *         elif rv == SAVE_FILE_OPEN_ERROR:
 *             raise MazeError('Could not open file.')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_kp_s_No_filename_given) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_No_filename_given);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1596, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1595
 *         if rv == SAVE_NO_MAZE_ERROR:
 *             raise MazeError('Maze not properly constructed.')
 *         elif rv == SAVE_NO_FILENAME_GIVEN_ERROR:             # <<<<<<<<<<<<<<
 *             raise MazeError('No filename given.')
 *         elif rv == SAVE_FILE_OPEN_ERROR:
 */
  }

  /* "daedalus/_maze.pyx":1597
 *         elif rv == SAVE_NO_FILENAME_GIVEN_ERROR:
 *             raise MazeError('No filename given.')
 *         elif rv == SAVE_FILE_OPEN_ERROR:             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not open file.')
 *         elif rv == SAVE_UNKNOWN_CMD_ERROR:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(SAVE_FILE_OPEN_ERROR); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_rv, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1597, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1597, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "daedalus/_maze.pyx":1598
 *             raise MazeError('No filename given.')
 *         elif rv == SAVE_FILE_OPEN_ERROR:
 *             raise MazeError('Could not open file.')             # <<<<<<<<<<<<<<
 *         elif rv == SAVE_UNKNOWN_CMD_ERROR:
 *             raise MazeError('Unknown save command.')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1598, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_kp_s_Could_not_open_file) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_s_Could_not_open_file);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1598, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1598, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1597
 *         elif rv == SAVE_NO_FILENAME_GIVEN_ERROR:
 *             raise MazeError('No filename given.')
 *         elif rv == SAVE_FILE_OPEN_ERROR:             # <<<<<<<<<<<<<<
 *             raise MazeError('Could not open file.')
 *         elif rv == SAVE_UNKNOWN_CMD_ERROR:
 */
  }

  /* "daedalus/_maze.pyx":1599
 *         elif rv == SAVE_FILE_OPEN_ERROR:
 *             raise MazeError('Could not open file.')
 *         elif rv == SAVE_UNKNOWN_CMD_ERROR:             # <<<<<<<<<<<<<<
 *             raise MazeError('Unknown save command.')
 *         elif rv == SAVE_SUCCESS:
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(SAVE_UNKNOWN_CMD_ERROR); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_rv, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1599, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1599, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "daedalus/_maze.pyx":1600
 *             raise MazeError('Could not open file.')
 *         elif rv == SAVE_UNKNOWN_CMD_ERROR:
 *             raise MazeError('Unknown save command.')             # <<<<<<<<<<<<<<
 *         elif rv == SAVE_SUCCESS:
 *             pass
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_kp_s_Unknown_save_command) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_Unknown_save_command);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1600, __pyx_L1_error)

    /* "daedalus/_maze.pyx":1599
 *         elif rv == SAVE_FILE_OPEN_ERROR:
 *             raise MazeError('Could not open file.')
 *         elif rv == SAVE_UNKNOWN_CMD_ERROR:             # <<<<<<<<<<<<<<
 *             raise MazeError('Unknown save command.')
 *         elif rv == SAVE_SUCCESS:
 */
  }

  /* "daedalus/_maze.pyx":1601
 *         elif rv == SAVE_UNKNOWN_CMD_ERROR:
 *             raise MazeError('Unknown save command.')
 *         elif rv == SAVE_SUCCESS:             # <<<<<<<<<<<<<<
 *             pass
 *         else:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(SAVE_SUCCESS); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_rv, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1601, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1601, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(__pyx_t_3)) {
    goto __pyx_L3;
  }

  /* "daedalus/_maze.pyx":1604
 *             pass
 *         else:
 *             raise MazeError('Unknown return value from Maze save function.')             # <<<<<<<<<<<<<<
 * 
 *     def _store_back_globals(self):
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_MazeError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1604, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_kp_s_Unknown_return_value_from_Maze_s) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_s_Unknown_return_value_from_Maze_s);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1604, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1604, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "daedalus/_maze.pyx":1586
 * 
 *     @staticmethod
 *     def _handle_save_retval(rv):             # <<<<<<<<<<<<<<
 *         '''
 *         Helper method to handle return values from the Save* methods.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("daedalus._maze.Maze._handle_save_retval", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":1606
 *             raise MazeError('Unknown return value from Maze save function.')
 * 
 *     def _store_back_globals(self):             # <<<<<<<<<<<<<<
 *         '''
 *         Helper function to retrieve global variable values and store them
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_68_store_back_globals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_67_store_back_globals[] = "\n        Helper function to retrieve global variable values and store them\n        into this instance after creating a maze.\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_68_store_back_globals(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_store_back_globals (wrapper)", 0);
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_67_store_back_globals(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_67_store_back_globals(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_store_back_globals", 0);

  /* "daedalus/_maze.pyx":1611
 *         into this instance after creating a maze.
 *         '''
 *         self._xEntrance = ms.xEntrance             # <<<<<<<<<<<<<<
 *         self._yEntrance = ms.yEntrance
 *         self._xExit     = ms.xExit
 */
  __pyx_t_1 = ms.xEntrance;
  __pyx_v_self->_xEntrance = __pyx_t_1;

  /* "daedalus/_maze.pyx":1612
 *         '''
 *         self._xEntrance = ms.xEntrance
 *         self._yEntrance = ms.yEntrance             # <<<<<<<<<<<<<<
 *         self._xExit     = ms.xExit
 *         self._yExit     = ms.yExit
 */
  __pyx_t_1 = ms.yEntrance;
  __pyx_v_self->_yEntrance = __pyx_t_1;

  /* "daedalus/_maze.pyx":1613
 *         self._xEntrance = ms.xEntrance
 *         self._yEntrance = ms.yEntrance
 *         self._xExit     = ms.xExit             # <<<<<<<<<<<<<<
 *         self._yExit     = ms.yExit
 * 
 */
  __pyx_t_1 = ms.xExit;
  __pyx_v_self->_xExit = __pyx_t_1;

  /* "daedalus/_maze.pyx":1614
 *         self._yEntrance = ms.yEntrance
 *         self._xExit     = ms.xExit
 *         self._yExit     = ms.yExit             # <<<<<<<<<<<<<<
 * 
 *     def save_bitmap(self, filename, kvOn=COLOR_WHITE, kvOff=COLOR_BLACK):
 */
  __pyx_t_1 = ms.yExit;
  __pyx_v_self->_yExit = __pyx_t_1;

  /* "daedalus/_maze.pyx":1606
 *             raise MazeError('Unknown return value from Maze save function.')
 * 
 *     def _store_back_globals(self):             # <<<<<<<<<<<<<<
 *         '''
 *         Helper function to retrieve global variable values and store them
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":1616
 *         self._yExit     = ms.yExit
 * 
 *     def save_bitmap(self, filename, kvOn=COLOR_WHITE, kvOff=COLOR_BLACK):             # <<<<<<<<<<<<<<
 *         '''
 *         Saves this Maze object as a bitmap to the given path.
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_70save_bitmap(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_69save_bitmap[] = "\n        Saves this Maze object as a bitmap to the given path.\n\n        :param str filename: The name of the file to write the bitmap\n            to.\n        :param int kvOn: The colour to draw wall pixels.  Defaults to\n            white (COLOR_WHITE).\n        :param int kvOff: The colour to draw passage pixels.  Defaults\n            to black (COLOR_BLACK).\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_70save_bitmap(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_filename = 0;
  PyObject *__pyx_v_kvOn = 0;
  PyObject *__pyx_v_kvOff = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("save_bitmap (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_filename,&__pyx_n_s_kvOn,&__pyx_n_s_kvOff,0};
    PyObject* values[3] = {0,0,0};
    values[1] = __pyx_k__26;
    values[2] = __pyx_k__27;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kvOn);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kvOff);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "save_bitmap") < 0)) __PYX_ERR(0, 1616, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_filename = values[0];
    __pyx_v_kvOn = values[1];
    __pyx_v_kvOff = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("save_bitmap", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1616, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.save_bitmap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_69save_bitmap(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_filename, __pyx_v_kvOn, __pyx_v_kvOff);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_69save_bitmap(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_filename, PyObject *__pyx_v_kvOn, PyObject *__pyx_v_kvOff) {
  int __pyx_v_retval;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char const *__pyx_t_4;
  KV __pyx_t_5;
  KV __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("save_bitmap", 0);
  __Pyx_INCREF(__pyx_v_filename);

  /* "daedalus/_maze.pyx":1628
 *         :rtype: None
 *         '''
 *         filename = compat.py3_str_to_bytes(filename)             # <<<<<<<<<<<<<<
 *         retval = cpp_SaveBitmap(self._maze, filename, kvOn, kvOff)
 *         self._handle_save_retval(retval)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_compat); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_py3_str_to_bytes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_filename);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_filename, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "daedalus/_maze.pyx":1629
 *         '''
 *         filename = compat.py3_str_to_bytes(filename)
 *         retval = cpp_SaveBitmap(self._maze, filename, kvOn, kvOff)             # <<<<<<<<<<<<<<
 *         self._handle_save_retval(retval)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_AsString(__pyx_v_filename); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 1629, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyInt_As_KV(__pyx_v_kvOn); if (unlikely((__pyx_t_5 == ((KV)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1629, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyInt_As_KV(__pyx_v_kvOff); if (unlikely((__pyx_t_6 == ((KV)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1629, __pyx_L1_error)
  __pyx_v_retval = cpp_SaveBitmap(__pyx_v_self->_maze, __pyx_t_4, __pyx_t_5, __pyx_t_6);

  /* "daedalus/_maze.pyx":1630
 *         filename = compat.py3_str_to_bytes(filename)
 *         retval = cpp_SaveBitmap(self._maze, filename, kvOn, kvOff)
 *         self._handle_save_retval(retval)             # <<<<<<<<<<<<<<
 * 
 *     def save_text(self, filename, fTextClip=True, fLineChar=False, fTextTab=False):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_handle_save_retval); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_retval); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_7, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "daedalus/_maze.pyx":1616
 *         self._yExit     = ms.yExit
 * 
 *     def save_bitmap(self, filename, kvOn=COLOR_WHITE, kvOff=COLOR_BLACK):             # <<<<<<<<<<<<<<
 *         '''
 *         Saves this Maze object as a bitmap to the given path.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("daedalus._maze.Maze.save_bitmap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_filename);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "daedalus/_maze.pyx":1632
 *         self._handle_save_retval(retval)
 * 
 *     def save_text(self, filename, fTextClip=True, fLineChar=False, fTextTab=False):             # <<<<<<<<<<<<<<
 *         '''
 *         Saves this Maze object formatted in ASCII text to the given path.
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_72save_text(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8daedalus_5_maze_4Maze_71save_text[] = "\n        Saves this Maze object formatted in ASCII text to the given path.\n\n        :param str filename: the name of the file to write the ASCII\n            representation to\n        :param bool fTextClip: If this flag is set, lines in the file\n            will never end with spaces (trailing space characters are\n            dropped). defaults to True\n        :param bool fLineChar: If this flag is set, maze walls are\n            drawn using horizontal (-) and vertical (|) lines, and\n            intersections (+).  If not, walls are drawn using hash\n            characters (#). Defaults to False.\n        :param bool fTextTab: Insert a tab between each\n            character. Defaults to False.\n        :rtype: None\n        ";
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_72save_text(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_filename = 0;
  PyObject *__pyx_v_fTextClip = 0;
  PyObject *__pyx_v_fLineChar = 0;
  PyObject *__pyx_v_fTextTab = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("save_text (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_filename,&__pyx_n_s_fTextClip,&__pyx_n_s_fLineChar,&__pyx_n_s_fTextTab,0};
    PyObject* values[4] = {0,0,0,0};
    values[1] = ((PyObject *)Py_True);
    values[2] = ((PyObject *)Py_False);
    values[3] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTextClip);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fLineChar);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fTextTab);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "save_text") < 0)) __PYX_ERR(0, 1632, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_filename = values[0];
    __pyx_v_fTextClip = values[1];
    __pyx_v_fLineChar = values[2];
    __pyx_v_fTextTab = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("save_text", 0, 1, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1632, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("daedalus._maze.Maze.save_text", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_71save_text(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), __pyx_v_filename, __pyx_v_fTextClip, __pyx_v_fLineChar, __pyx_v_fTextTab);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_71save_text(struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, PyObject *__pyx_v_filename, PyObject *__pyx_v_fTextClip, PyObject *__pyx_v_fLineChar, PyObject *__pyx_v_fTextTab) {
  int __pyx_v_retval;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char const *__pyx_t_4;
  bool __pyx_t_5;
  bool __pyx_t_6;
  bool __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("save_text", 0);
  __Pyx_INCREF(__pyx_v_filename);

  /* "daedalus/_maze.pyx":1649
 *         :rtype: None
 *         '''
 *         filename = compat.py3_str_to_bytes(filename)             # <<<<<<<<<<<<<<
 *         retval = cpp_SaveText(self._maze, filename, fTextClip, fLineChar, fTextTab)
 *         self._handle_save_retval(retval)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_compat); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_py3_str_to_bytes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_filename);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_filename, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "daedalus/_maze.pyx":1650
 *         '''
 *         filename = compat.py3_str_to_bytes(filename)
 *         retval = cpp_SaveText(self._maze, filename, fTextClip, fLineChar, fTextTab)             # <<<<<<<<<<<<<<
 *         self._handle_save_retval(retval)
 */
  __pyx_t_4 = __Pyx_PyObject_AsString(__pyx_v_filename); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 1650, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_fTextClip); if (unlikely((__pyx_t_5 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1650, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_fLineChar); if (unlikely((__pyx_t_6 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1650, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_fTextTab); if (unlikely((__pyx_t_7 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1650, __pyx_L1_error)
  __pyx_v_retval = cpp_SaveText(__pyx_v_self->_maze, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7);

  /* "daedalus/_maze.pyx":1651
 *         filename = compat.py3_str_to_bytes(filename)
 *         retval = cpp_SaveText(self._maze, filename, fTextClip, fLineChar, fTextTab)
 *         self._handle_save_retval(retval)             # <<<<<<<<<<<<<<
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_handle_save_retval); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_retval); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_8, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "daedalus/_maze.pyx":1632
 *         self._handle_save_retval(retval)
 * 
 *     def save_text(self, filename, fTextClip=True, fLineChar=False, fTextTab=False):             # <<<<<<<<<<<<<<
 *         '''
 *         Saves this Maze object formatted in ASCII text to the given path.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("daedalus._maze.Maze.save_text", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_filename);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_74__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_74__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_73__reduce_cython__(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_73__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("daedalus._maze.Maze.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_76__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_8daedalus_5_maze_4Maze_76__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8daedalus_5_maze_4Maze_75__setstate_cython__(((struct __pyx_obj_8daedalus_5_maze_Maze *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8daedalus_5_maze_4Maze_75__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8daedalus_5_maze_Maze *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("daedalus._maze.Maze.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_8daedalus_5_maze_Maze(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  if (unlikely(__pyx_pw_8daedalus_5_maze_4Maze_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8daedalus_5_maze_Maze(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_8daedalus_5_maze_4Maze_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_8daedalus_5_maze_Maze(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_getprop_8daedalus_5_maze_4Maze_width(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8daedalus_5_maze_4Maze_5width_1__get__(o);
}

static PyObject *__pyx_getprop_8daedalus_5_maze_4Maze_height(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8daedalus_5_maze_4Maze_6height_1__get__(o);
}

static PyObject *__pyx_getprop_8daedalus_5_maze_4Maze_entrance(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8daedalus_5_maze_4Maze_8entrance_1__get__(o);
}

static PyObject *__pyx_getprop_8daedalus_5_maze_4Maze_exit(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_8daedalus_5_maze_4Maze_4exit_1__get__(o);
}

static PyMethodDef __pyx_methods_8daedalus_5_maze_Maze[] = {
  {"index", (PyCFunction)__pyx_pw_8daedalus_5_maze_4Maze_18index, METH_O, __pyx_doc_8daedalus_5_maze_4Maze_17index},
  {"count", (PyCFunction)__pyx_pw_8daedalus_5_maze_4Maze_20count, METH_O, __pyx_doc_8daedalus_5_maze_4Maze_19count},
  {"get", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_22get, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_21get},
  {"create_perfect", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_24create_perfect, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_23create_perfect},
  {"create_perfect2", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_26create_perfect2, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_25create_perfect2},
  {"create_braid", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_28create_braid, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_27create_braid},
  {"create_braid_tilt", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_30create_braid_tilt, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_29create_braid_tilt},
  {"create_spiral", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_32create_spiral, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_31create_spiral},
  {"create_diagonal", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_34create_diagonal, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_33create_diagonal},
  {"create_recursive", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_36create_recursive, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_35create_recursive},
  {"create_prim", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_38create_prim, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_37create_prim},
  {"create_prim2", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_40create_prim2, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_39create_prim2},
  {"create_kruskal", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_42create_kruskal, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_41create_kruskal},
  {"create_tree", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_44create_tree, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_43create_tree},
  {"create_forest", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_46create_forest, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_45create_forest},
  {"create_aldous_broder", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_48create_aldous_broder, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_47create_aldous_broder},
  {"create_wilson", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_50create_wilson, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_49create_wilson},
  {"create_eller", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_52create_eller, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_51create_eller},
  {"create_braid_eller", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_54create_braid_eller, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_53create_braid_eller},
  {"create_division", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_56create_division, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_55create_division},
  {"create_binary", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_58create_binary, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_57create_binary},
  {"create_sidewinder", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_60create_sidewinder, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_59create_sidewinder},
  {"create_unicursal", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_62create_unicursal, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_61create_unicursal},
  {"resize", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_64resize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_63resize},
  {"_handle_save_retval", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_66_handle_save_retval, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_65_handle_save_retval},
  {"_store_back_globals", (PyCFunction)__pyx_pw_8daedalus_5_maze_4Maze_68_store_back_globals, METH_NOARGS, __pyx_doc_8daedalus_5_maze_4Maze_67_store_back_globals},
  {"save_bitmap", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_70save_bitmap, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_69save_bitmap},
  {"save_text", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8daedalus_5_maze_4Maze_72save_text, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8daedalus_5_maze_4Maze_71save_text},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_8daedalus_5_maze_4Maze_74__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_8daedalus_5_maze_4Maze_76__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_8daedalus_5_maze_Maze[] = {
  {(char *)"width", __pyx_getprop_8daedalus_5_maze_4Maze_width, 0, (char *)"\n        Returns this Maze's width.\n        ", 0},
  {(char *)"height", __pyx_getprop_8daedalus_5_maze_4Maze_height, 0, (char *)"\n        Returns this Maze's height.\n        ", 0},
  {(char *)"entrance", __pyx_getprop_8daedalus_5_maze_4Maze_entrance, 0, (char *)"\n        Returns this Maze's entrance location as a tuple (x,y).\n        ", 0},
  {(char *)"exit", __pyx_getprop_8daedalus_5_maze_4Maze_exit, 0, (char *)"\n        Returns this Maze's exit location as a tuple (x,y).\n        ", 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_Maze = {
  __pyx_pw_8daedalus_5_maze_4Maze_5__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_8daedalus_5_maze_Maze, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_8daedalus_5_maze_4Maze_9__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Maze = {
  __pyx_pw_8daedalus_5_maze_4Maze_5__len__, /*mp_length*/
  __pyx_pw_8daedalus_5_maze_4Maze_7__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_8daedalus_5_maze_Maze = {
  PyVarObject_HEAD_INIT(0, 0)
  "daedalus._maze.Maze", /*tp_name*/
  sizeof(struct __pyx_obj_8daedalus_5_maze_Maze), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8daedalus_5_maze_Maze, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_8daedalus_5_maze_4Maze_16__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Maze, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Maze, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_8daedalus_5_maze_4Maze_14__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "\n    A monochrome daedalus Maze object.\n\n    Even-numbered rows and columns in the Maze are walls, and\n    odd-numbered rows and columns are passages.\n\n    Walls are \"on\" pixels and passages are \"off\" pixels.\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_8daedalus_5_maze_4Maze_11__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8daedalus_5_maze_Maze, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_8daedalus_5_maze_Maze, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8daedalus_5_maze_Maze, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__ *__pyx_freelist_8daedalus_5_maze___pyx_scope_struct____iter__[8];
static int __pyx_freecount_8daedalus_5_maze___pyx_scope_struct____iter__ = 0;

static PyObject *__pyx_tp_new_8daedalus_5_maze___pyx_scope_struct____iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8daedalus_5_maze___pyx_scope_struct____iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__)))) {
    o = (PyObject*)__pyx_freelist_8daedalus_5_maze___pyx_scope_struct____iter__[--__pyx_freecount_8daedalus_5_maze___pyx_scope_struct____iter__];
    memset(o, 0, sizeof(struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8daedalus_5_maze___pyx_scope_struct____iter__(PyObject *o) {
  struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__ *p = (struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_i);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_v);
  Py_CLEAR(p->__pyx_t_0);
  Py_CLEAR(p->__pyx_t_1);
  Py_CLEAR(p->__pyx_t_2);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8daedalus_5_maze___pyx_scope_struct____iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__)))) {
    __pyx_freelist_8daedalus_5_maze___pyx_scope_struct____iter__[__pyx_freecount_8daedalus_5_maze___pyx_scope_struct____iter__++] = ((struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8daedalus_5_maze___pyx_scope_struct____iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__ *p = (struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__ *)o;
  if (p->__pyx_v_i) {
    e = (*v)(p->__pyx_v_i, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  if (p->__pyx_t_1) {
    e = (*v)(p->__pyx_t_1, a); if (e) return e;
  }
  if (p->__pyx_t_2) {
    e = (*v)(p->__pyx_t_2, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_8daedalus_5_maze___pyx_scope_struct____iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "daedalus._maze.__pyx_scope_struct____iter__", /*tp_name*/
  sizeof(struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct____iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8daedalus_5_maze___pyx_scope_struct____iter__, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8daedalus_5_maze___pyx_scope_struct____iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8daedalus_5_maze___pyx_scope_struct____iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count *__pyx_freelist_8daedalus_5_maze___pyx_scope_struct_1_count[8];
static int __pyx_freecount_8daedalus_5_maze___pyx_scope_struct_1_count = 0;

static PyObject *__pyx_tp_new_8daedalus_5_maze___pyx_scope_struct_1_count(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8daedalus_5_maze___pyx_scope_struct_1_count > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count)))) {
    o = (PyObject*)__pyx_freelist_8daedalus_5_maze___pyx_scope_struct_1_count[--__pyx_freecount_8daedalus_5_maze___pyx_scope_struct_1_count];
    memset(o, 0, sizeof(struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8daedalus_5_maze___pyx_scope_struct_1_count(PyObject *o) {
  struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count *p = (struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_value);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8daedalus_5_maze___pyx_scope_struct_1_count < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count)))) {
    __pyx_freelist_8daedalus_5_maze___pyx_scope_struct_1_count[__pyx_freecount_8daedalus_5_maze___pyx_scope_struct_1_count++] = ((struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8daedalus_5_maze___pyx_scope_struct_1_count(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count *p = (struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_value) {
    e = (*v)(p->__pyx_v_value, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8daedalus_5_maze___pyx_scope_struct_1_count(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count *p = (struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_8daedalus_5_maze_Maze *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_value);
  p->__pyx_v_value = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_8daedalus_5_maze___pyx_scope_struct_1_count = {
  PyVarObject_HEAD_INIT(0, 0)
  "daedalus._maze.__pyx_scope_struct_1_count", /*tp_name*/
  sizeof(struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_1_count), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8daedalus_5_maze___pyx_scope_struct_1_count, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8daedalus_5_maze___pyx_scope_struct_1_count, /*tp_traverse*/
  __pyx_tp_clear_8daedalus_5_maze___pyx_scope_struct_1_count, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8daedalus_5_maze___pyx_scope_struct_1_count, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr *__pyx_freelist_8daedalus_5_maze___pyx_scope_struct_2_genexpr[8];
static int __pyx_freecount_8daedalus_5_maze___pyx_scope_struct_2_genexpr = 0;

static PyObject *__pyx_tp_new_8daedalus_5_maze___pyx_scope_struct_2_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8daedalus_5_maze___pyx_scope_struct_2_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr)))) {
    o = (PyObject*)__pyx_freelist_8daedalus_5_maze___pyx_scope_struct_2_genexpr[--__pyx_freecount_8daedalus_5_maze___pyx_scope_struct_2_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8daedalus_5_maze___pyx_scope_struct_2_genexpr(PyObject *o) {
  struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_v);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8daedalus_5_maze___pyx_scope_struct_2_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr)))) {
    __pyx_freelist_8daedalus_5_maze___pyx_scope_struct_2_genexpr[__pyx_freecount_8daedalus_5_maze___pyx_scope_struct_2_genexpr++] = ((struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8daedalus_5_maze___pyx_scope_struct_2_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_8daedalus_5_maze___pyx_scope_struct_2_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "daedalus._maze.__pyx_scope_struct_2_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_8daedalus_5_maze___pyx_scope_struct_2_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8daedalus_5_maze___pyx_scope_struct_2_genexpr, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8daedalus_5_maze___pyx_scope_struct_2_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8daedalus_5_maze___pyx_scope_struct_2_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__maze(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__maze},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "_maze",
    __pyx_k_Wrapper_for_the_daedalus_monoch, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_An_error_while_performing_an_act, __pyx_k_An_error_while_performing_an_act, sizeof(__pyx_k_An_error_while_performing_an_act), 0, 0, 1, 0},
  {&__pyx_n_s_COLOR_BLACK, __pyx_k_COLOR_BLACK, sizeof(__pyx_k_COLOR_BLACK), 0, 0, 1, 1},
  {&__pyx_n_s_COLOR_WHITE, __pyx_k_COLOR_WHITE, sizeof(__pyx_k_COLOR_WHITE), 0, 0, 1, 1},
  {&__pyx_kp_s_Could_not_construct_Maze_object, __pyx_k_Could_not_construct_Maze_object, sizeof(__pyx_k_Could_not_construct_Maze_object), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_AldousBroder_Ma, __pyx_k_Could_not_create_AldousBroder_Ma, sizeof(__pyx_k_Could_not_create_AldousBroder_Ma), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Binary_Maze, __pyx_k_Could_not_create_Binary_Maze, sizeof(__pyx_k_Could_not_create_Binary_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_BraidEller_Maze, __pyx_k_Could_not_create_BraidEller_Maze, sizeof(__pyx_k_Could_not_create_BraidEller_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_BraidTilt_Maze, __pyx_k_Could_not_create_BraidTilt_Maze, sizeof(__pyx_k_Could_not_create_BraidTilt_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Braid_Maze, __pyx_k_Could_not_create_Braid_Maze, sizeof(__pyx_k_Could_not_create_Braid_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Diagonal_Maze, __pyx_k_Could_not_create_Diagonal_Maze, sizeof(__pyx_k_Could_not_create_Diagonal_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Division_Maze, __pyx_k_Could_not_create_Division_Maze, sizeof(__pyx_k_Could_not_create_Division_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Eller_Maze, __pyx_k_Could_not_create_Eller_Maze, sizeof(__pyx_k_Could_not_create_Eller_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Forest_Maze, __pyx_k_Could_not_create_Forest_Maze, sizeof(__pyx_k_Could_not_create_Forest_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Kruskal_Maze, __pyx_k_Could_not_create_Kruskal_Maze, sizeof(__pyx_k_Could_not_create_Kruskal_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Perfect2_Maze, __pyx_k_Could_not_create_Perfect2_Maze, sizeof(__pyx_k_Could_not_create_Perfect2_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Perfect_Maze, __pyx_k_Could_not_create_Perfect_Maze, sizeof(__pyx_k_Could_not_create_Perfect_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Prim2_Maze, __pyx_k_Could_not_create_Prim2_Maze, sizeof(__pyx_k_Could_not_create_Prim2_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Prim_Maze, __pyx_k_Could_not_create_Prim_Maze, sizeof(__pyx_k_Could_not_create_Prim_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Recursive_Maze, __pyx_k_Could_not_create_Recursive_Maze, sizeof(__pyx_k_Could_not_create_Recursive_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Sidewinder_Maze, __pyx_k_Could_not_create_Sidewinder_Maze, sizeof(__pyx_k_Could_not_create_Sidewinder_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Spiral_Maze, __pyx_k_Could_not_create_Spiral_Maze, sizeof(__pyx_k_Could_not_create_Spiral_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Tree_Maze, __pyx_k_Could_not_create_Tree_Maze, sizeof(__pyx_k_Could_not_create_Tree_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Unicursal_Maze, __pyx_k_Could_not_create_Unicursal_Maze, sizeof(__pyx_k_Could_not_create_Unicursal_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_Wilson_Maze, __pyx_k_Could_not_create_Wilson_Maze, sizeof(__pyx_k_Could_not_create_Wilson_Maze), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_open_file, __pyx_k_Could_not_open_file, sizeof(__pyx_k_Could_not_open_file), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_resize_the_Maze, __pyx_k_Could_not_resize_the_Maze, sizeof(__pyx_k_Could_not_resize_the_Maze), 0, 0, 1, 0},
  {&__pyx_n_s_ENTRANCE_BALANCED, __pyx_k_ENTRANCE_BALANCED, sizeof(__pyx_k_ENTRANCE_BALANCED), 0, 0, 1, 1},
  {&__pyx_n_s_ENTRANCE_CORNER, __pyx_k_ENTRANCE_CORNER, sizeof(__pyx_k_ENTRANCE_CORNER), 0, 0, 1, 1},
  {&__pyx_n_s_ENTRANCE_MIDDLE, __pyx_k_ENTRANCE_MIDDLE, sizeof(__pyx_k_ENTRANCE_MIDDLE), 0, 0, 1, 1},
  {&__pyx_n_s_ENTRANCE_RANDOM, __pyx_k_ENTRANCE_RANDOM, sizeof(__pyx_k_ENTRANCE_RANDOM), 0, 0, 1, 1},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_n_s_Maze, __pyx_k_Maze, sizeof(__pyx_k_Maze), 0, 0, 1, 1},
  {&__pyx_n_s_MazeError, __pyx_k_MazeError, sizeof(__pyx_k_MazeError), 0, 0, 1, 1},
  {&__pyx_n_s_Maze___iter, __pyx_k_Maze___iter, sizeof(__pyx_k_Maze___iter), 0, 0, 1, 1},
  {&__pyx_kp_s_Maze_not_properly_constructed, __pyx_k_Maze_not_properly_constructed, sizeof(__pyx_k_Maze_not_properly_constructed), 0, 0, 1, 0},
  {&__pyx_kp_s_Maze_x, __pyx_k_Maze_x, sizeof(__pyx_k_Maze_x), 0, 0, 1, 0},
  {&__pyx_kp_s_No_filename_given, __pyx_k_No_filename_given, sizeof(__pyx_k_No_filename_given), 0, 0, 1, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unknown_return_value_from_Maze_s, __pyx_k_Unknown_return_value_from_Maze_s, sizeof(__pyx_k_Unknown_return_value_from_Maze_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Unknown_save_command, __pyx_k_Unknown_save_command, sizeof(__pyx_k_Unknown_save_command), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s__30, __pyx_k__30, sizeof(__pyx_k__30), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_c2, __pyx_k_c2, sizeof(__pyx_k_c2), 0, 0, 1, 1},
  {&__pyx_n_s_c3, __pyx_k_c3, sizeof(__pyx_k_c3), 0, 0, 1, 1},
  {&__pyx_n_s_cRandomAdd, __pyx_k_cRandomAdd, sizeof(__pyx_k_cRandomAdd), 0, 0, 1, 1},
  {&__pyx_n_s_cSpiral, __pyx_k_cSpiral, sizeof(__pyx_k_cSpiral), 0, 0, 1, 1},
  {&__pyx_n_s_cSpiralWall, __pyx_k_cSpiralWall, sizeof(__pyx_k_cSpiralWall), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_compat, __pyx_k_compat, sizeof(__pyx_k_compat), 0, 0, 1, 1},
  {&__pyx_n_s_count_locals_genexpr, __pyx_k_count_locals_genexpr, sizeof(__pyx_k_count_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_daedalus__maze, __pyx_k_daedalus__maze, sizeof(__pyx_k_daedalus__maze), 0, 0, 1, 1},
  {&__pyx_kp_s_daedalus__maze_pyx, __pyx_k_daedalus__maze_pyx, sizeof(__pyx_k_daedalus__maze_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_fClear, __pyx_k_fClear, sizeof(__pyx_k_fClear), 0, 0, 1, 1},
  {&__pyx_n_s_fKruskalPic, __pyx_k_fKruskalPic, sizeof(__pyx_k_fKruskalPic), 0, 0, 1, 1},
  {&__pyx_n_s_fLineChar, __pyx_k_fLineChar, sizeof(__pyx_k_fLineChar), 0, 0, 1, 1},
  {&__pyx_n_s_fRiver, __pyx_k_fRiver, sizeof(__pyx_k_fRiver), 0, 0, 1, 1},
  {&__pyx_n_s_fTextClip, __pyx_k_fTextClip, sizeof(__pyx_k_fTextClip), 0, 0, 1, 1},
  {&__pyx_n_s_fTextTab, __pyx_k_fTextTab, sizeof(__pyx_k_fTextTab), 0, 0, 1, 1},
  {&__pyx_n_s_fTiltDiamond, __pyx_k_fTiltDiamond, sizeof(__pyx_k_fTiltDiamond), 0, 0, 1, 1},
  {&__pyx_n_s_fTreeRandom, __pyx_k_fTreeRandom, sizeof(__pyx_k_fTreeRandom), 0, 0, 1, 1},
  {&__pyx_n_s_fTreeWall, __pyx_k_fTreeWall, sizeof(__pyx_k_fTreeWall), 0, 0, 1, 1},
  {&__pyx_n_s_fWall, __pyx_k_fWall, sizeof(__pyx_k_fWall), 0, 0, 1, 1},
  {&__pyx_n_s_filename, __pyx_k_filename, sizeof(__pyx_k_filename), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_handle_save_retval, __pyx_k_handle_save_retval, sizeof(__pyx_k_handle_save_retval), 0, 0, 1, 1},
  {&__pyx_n_s_height, __pyx_k_height, sizeof(__pyx_k_height), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_init_random, __pyx_k_init_random, sizeof(__pyx_k_init_random), 0, 0, 1, 1},
  {&__pyx_n_s_iter, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
  {&__pyx_kp_s_key_must_be_an_integer_or_a_tupl, __pyx_k_key_must_be_an_integer_or_a_tupl, sizeof(__pyx_k_key_must_be_an_integer_or_a_tupl), 0, 0, 1, 0},
  {&__pyx_n_s_kvOff, __pyx_k_kvOff, sizeof(__pyx_k_kvOff), 0, 0, 1, 1},
  {&__pyx_n_s_kvOn, __pyx_k_kvOn, sizeof(__pyx_k_kvOn), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_map, __pyx_k_map, sizeof(__pyx_k_map), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_nEntrancePos, __pyx_k_nEntrancePos, sizeof(__pyx_k_nEntrancePos), 0, 0, 1, 1},
  {&__pyx_n_s_nForsAdd, __pyx_k_nForsAdd, sizeof(__pyx_k_nForsAdd), 0, 0, 1, 1},
  {&__pyx_n_s_nForsInit, __pyx_k_nForsInit, sizeof(__pyx_k_nForsInit), 0, 0, 1, 1},
  {&__pyx_n_s_nRndBias, __pyx_k_nRndBias, sizeof(__pyx_k_nRndBias), 0, 0, 1, 1},
  {&__pyx_n_s_nRndRun, __pyx_k_nRndRun, sizeof(__pyx_k_nRndRun), 0, 0, 1, 1},
  {&__pyx_n_s_nTreeRiver, __pyx_k_nTreeRiver, sizeof(__pyx_k_nTreeRiver), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_kp_s_negative_height, __pyx_k_negative_height, sizeof(__pyx_k_negative_height), 0, 0, 1, 0},
  {&__pyx_kp_s_negative_width, __pyx_k_negative_width, sizeof(__pyx_k_negative_width), 0, 0, 1, 0},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_py3_str_to_bytes, __pyx_k_py3_str_to_bytes, sizeof(__pyx_k_py3_str_to_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_rv, __pyx_k_rv, sizeof(__pyx_k_rv), 0, 0, 1, 1},
  {&__pyx_n_s_seed, __pyx_k_seed, sizeof(__pyx_k_seed), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_staticmethod, __pyx_k_staticmethod, sizeof(__pyx_k_staticmethod), 0, 0, 1, 1},
  {&__pyx_n_s_store_back_globals, __pyx_k_store_back_globals, sizeof(__pyx_k_store_back_globals), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_width, __pyx_k_width, sizeof(__pyx_k_width), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_kp_s_x_coordinate_out_of_range, __pyx_k_x_coordinate_out_of_range, sizeof(__pyx_k_x_coordinate_out_of_range), 0, 0, 1, 0},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_kp_s_y_coordinate_out_of_range, __pyx_k_y_coordinate_out_of_range, sizeof(__pyx_k_y_coordinate_out_of_range), 0, 0, 1, 0},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_staticmethod = __Pyx_GetBuiltinName(__pyx_n_s_staticmethod); if (!__pyx_builtin_staticmethod) __PYX_ERR(0, 1585, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 258, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 285, __pyx_L1_error)
  __pyx_builtin_map = __Pyx_GetBuiltinName(__pyx_n_s_map); if (!__pyx_builtin_map) __PYX_ERR(0, 287, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 305, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 324, __pyx_L1_error)
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) __PYX_ERR(0, 338, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "daedalus/_maze.pyx":258
 *         '''
 *         if width < 0:
 *             raise ValueError('negative width')             # <<<<<<<<<<<<<<
 *         if height < 0:
 *             raise ValueError('negative height')
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_negative_width); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "daedalus/_maze.pyx":260
 *             raise ValueError('negative width')
 *         if height < 0:
 *             raise ValueError('negative height')             # <<<<<<<<<<<<<<
 *         # if width % 2 != 1:
 *         #     raise ValueError('width must be odd')
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_negative_height); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "daedalus/_maze.pyx":288
 *         else:
 *             if not isinstance(key, tuple) and len(key) == 2 and list(set(map(type, key))) == [int]:
 *                 raise ValueError('key must be an integer or a tuple of two integers')             # <<<<<<<<<<<<<<
 *             (x,y) = key
 *             return self.get(x,y)
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_key_must_be_an_integer_or_a_tupl); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "daedalus/_maze.pyx":383
 *         '''
 *         if x < 0 or self.width <= x:
 *             raise IndexError('x coordinate out of range')             # <<<<<<<<<<<<<<
 *         if y < 0 or self.height <= y:
 *             raise IndexError('y coordinate out of range')
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_x_coordinate_out_of_range); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "daedalus/_maze.pyx":385
 *             raise IndexError('x coordinate out of range')
 *         if y < 0 or self.height <= y:
 *             raise IndexError('y coordinate out of range')             # <<<<<<<<<<<<<<
 *         return cpp_Get(self._maze, x, y)
 * 
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_y_coordinate_out_of_range); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "daedalus/_maze.pyx":219
 * cpp_Init()
 * 
 * def init_random(seed):             # <<<<<<<<<<<<<<
 *     '''
 *     Seeds the random number generator used internally by Daedalus to
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_n_s_seed); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_daedalus__maze_pyx, __pyx_n_s_init_random, 219, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 219, __pyx_L1_error)

  /* "daedalus/_maze.pyx":1586
 * 
 *     @staticmethod
 *     def _handle_save_retval(rv):             # <<<<<<<<<<<<<<
 *         '''
 *         Helper method to handle return values from the Save* methods.
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_n_s_rv); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 1586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_daedalus__maze_pyx, __pyx_n_s_handle_save_retval, 1586, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 1586, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_15 = PyInt_FromLong(15); if (unlikely(!__pyx_int_15)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_100 = PyInt_FromLong(-100); if (unlikely(!__pyx_int_neg_100)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_8daedalus_5_maze_Maze) < 0) __PYX_ERR(0, 234, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_8daedalus_5_maze_Maze.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8daedalus_5_maze_Maze.tp_dictoffset && __pyx_type_8daedalus_5_maze_Maze.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8daedalus_5_maze_Maze.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Maze, (PyObject *)&__pyx_type_8daedalus_5_maze_Maze) < 0) __PYX_ERR(0, 234, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_8daedalus_5_maze_Maze) < 0) __PYX_ERR(0, 234, __pyx_L1_error)
  __pyx_ptype_8daedalus_5_maze_Maze = &__pyx_type_8daedalus_5_maze_Maze;
  if (PyType_Ready(&__pyx_type_8daedalus_5_maze___pyx_scope_struct____iter__) < 0) __PYX_ERR(0, 298, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_8daedalus_5_maze___pyx_scope_struct____iter__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8daedalus_5_maze___pyx_scope_struct____iter__.tp_dictoffset && __pyx_type_8daedalus_5_maze___pyx_scope_struct____iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8daedalus_5_maze___pyx_scope_struct____iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_8daedalus_5_maze___pyx_scope_struct____iter__ = &__pyx_type_8daedalus_5_maze___pyx_scope_struct____iter__;
  if (PyType_Ready(&__pyx_type_8daedalus_5_maze___pyx_scope_struct_1_count) < 0) __PYX_ERR(0, 329, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_8daedalus_5_maze___pyx_scope_struct_1_count.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8daedalus_5_maze___pyx_scope_struct_1_count.tp_dictoffset && __pyx_type_8daedalus_5_maze___pyx_scope_struct_1_count.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8daedalus_5_maze___pyx_scope_struct_1_count.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_8daedalus_5_maze___pyx_scope_struct_1_count = &__pyx_type_8daedalus_5_maze___pyx_scope_struct_1_count;
  if (PyType_Ready(&__pyx_type_8daedalus_5_maze___pyx_scope_struct_2_genexpr) < 0) __PYX_ERR(0, 338, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_8daedalus_5_maze___pyx_scope_struct_2_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8daedalus_5_maze___pyx_scope_struct_2_genexpr.tp_dictoffset && __pyx_type_8daedalus_5_maze___pyx_scope_struct_2_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8daedalus_5_maze___pyx_scope_struct_2_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_8daedalus_5_maze___pyx_scope_struct_2_genexpr = &__pyx_type_8daedalus_5_maze___pyx_scope_struct_2_genexpr;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_maze(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_maze(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__maze(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__maze(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__maze(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_maze' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__maze(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_maze", __pyx_methods, __pyx_k_Wrapper_for_the_daedalus_monoch, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_daedalus___maze) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "daedalus._maze")) {
      if (unlikely(PyDict_SetItemString(modules, "daedalus._maze", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "daedalus/_maze.pyx":17
 * from libcpp cimport bool
 * 
 * from . import compat             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "pdconfig.h":
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_compat);
  __Pyx_GIVEREF(__pyx_n_s_compat);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_compat);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s__30, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_compat); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_compat, __pyx_t_1) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":207
 * 
 * # color value constants
 * COLOR_WHITE = kvWhite             # <<<<<<<<<<<<<<
 * COLOR_BLACK = kvBlack
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_KV(kvWhite); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_COLOR_WHITE, __pyx_t_2) < 0) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":208
 * # color value constants
 * COLOR_WHITE = kvWhite
 * COLOR_BLACK = kvBlack             # <<<<<<<<<<<<<<
 * 
 * # nEntrancePos enum constants
 */
  __pyx_t_2 = __Pyx_PyInt_From_KV(kvBlack); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_COLOR_BLACK, __pyx_t_2) < 0) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":211
 * 
 * # nEntrancePos enum constants
 * ENTRANCE_CORNER   = epCorner             # <<<<<<<<<<<<<<
 * ENTRANCE_MIDDLE   = epMiddle
 * ENTRANCE_BALANCED = epBalanced
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(epCorner); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ENTRANCE_CORNER, __pyx_t_2) < 0) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":212
 * # nEntrancePos enum constants
 * ENTRANCE_CORNER   = epCorner
 * ENTRANCE_MIDDLE   = epMiddle             # <<<<<<<<<<<<<<
 * ENTRANCE_BALANCED = epBalanced
 * ENTRANCE_RANDOM   = epRandom
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(epMiddle); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ENTRANCE_MIDDLE, __pyx_t_2) < 0) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":213
 * ENTRANCE_CORNER   = epCorner
 * ENTRANCE_MIDDLE   = epMiddle
 * ENTRANCE_BALANCED = epBalanced             # <<<<<<<<<<<<<<
 * ENTRANCE_RANDOM   = epRandom
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(epBalanced); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ENTRANCE_BALANCED, __pyx_t_2) < 0) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":214
 * ENTRANCE_MIDDLE   = epMiddle
 * ENTRANCE_BALANCED = epBalanced
 * ENTRANCE_RANDOM   = epRandom             # <<<<<<<<<<<<<<
 * 
 * # global initialisation
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(epRandom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ENTRANCE_RANDOM, __pyx_t_2) < 0) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":217
 * 
 * # global initialisation
 * cpp_Init()             # <<<<<<<<<<<<<<
 * 
 * def init_random(seed):
 */
  cpp_Init();

  /* "daedalus/_maze.pyx":219
 * cpp_Init()
 * 
 * def init_random(seed):             # <<<<<<<<<<<<<<
 *     '''
 *     Seeds the random number generator used internally by Daedalus to
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_8daedalus_5_maze_1init_random, NULL, __pyx_n_s_daedalus__maze); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_init_random, __pyx_t_2) < 0) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":230
 *     InitRndL(seed)
 * 
 * class MazeError(Exception):             # <<<<<<<<<<<<<<
 *     '''An error while performing an action on a Maze object.'''
 *     pass
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __Pyx_GIVEREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_t_2, __pyx_n_s_MazeError, __pyx_n_s_MazeError, (PyObject *) NULL, __pyx_n_s_daedalus__maze, __pyx_kp_s_An_error_while_performing_an_act); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_MazeError, __pyx_t_2, __pyx_t_3, NULL, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MazeError, __pyx_t_4) < 0) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":391
 *                        fRiver=True,
 *                        fTreeWall=False,
 *                        nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                        nRndBias=0,
 *                        nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__6 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":458
 *     def create_perfect2(self,
 *                         fRiver=True,
 *                         nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                         nRndBias=0,
 *                         nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__7 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":506
 * 
 *     def create_braid(self,
 *                      nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                      nRndBias=0,
 *                      nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__8 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":551
 *     def create_braid_tilt(self,
 *                           fTiltDiamond=False,
 *                           nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                           nRndBias=0,
 *                           nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__9 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":604
 *                       cSpiral=15,
 *                       cSpiralWall=15,
 *                       nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                       nRndBias=0,
 *                       nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__10 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":667
 *     def create_diagonal(self,
 *                         cRandomAdd=0,
 *                         nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                         nRndBias=0,
 *                         nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__11 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":717
 * 
 *     def create_recursive(self,
 *                          nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                          nRndBias=0,
 *                          nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__12 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":770
 *     def create_prim(self,
 *                     fTreeWall=False,
 *                     nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                     nRndBias=0,
 *                     nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__13 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":830
 *                      fTreeRandom=True,
 *                      fTreeWall=False,
 *                      nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                      nRndBias=0,
 *                      nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__14 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":897
 *                        fKruskalPic=False,
 *                        fTreeWall=False,
 *                        nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                        nRndBias=0,
 *                        nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__15 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":960
 *                     fTreeRandom=True,
 *                     fTreeWall=False,
 *                     nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                     nRndBias=0,
 *                     nRndRun=0,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__16 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":1029
 *                       fTreeRandom=True,
 *                       fTreeWall=False,
 *                       nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                       nForsAdd=-100,
 *                       nForsInit=1,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1029, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__17 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":1109
 *     def create_aldous_broder(self,
 *                              fTreeWall=False,
 *                              nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                              nRndBias=0,
 *                              nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__18 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":1167
 *     def create_wilson(self,
 *                       fTreeWall=False,
 *                       nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                       nRndBias=0,
 *                       nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__19 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":1227
 *     def create_eller(self,
 *                      fTreeWall=False,
 *                      nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                      nRndBias=0,
 *                      nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__20 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":1283
 * 
 *     def create_braid_eller(self,
 *                            nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                            nRndBias=0,
 *                            nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__21 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":1327
 * 
 *     def create_division(self,
 *                         nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                         nRndBias=0,
 *                         nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__22 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":1380
 *                       cRandomAdd=0,
 *                       fTreeWall=False,
 *                       nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                       nRndBias=0,
 *                       nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__23 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":1448
 *     def create_sidewinder(self,
 *                           fTreeWall=False,
 *                           nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                           nRndBias=0,
 *                           nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__24 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":1507
 *                          fRiver=True,
 *                          fTreeWall=False,
 *                          nEntrancePos=ENTRANCE_RANDOM,             # <<<<<<<<<<<<<<
 *                          nRndBias=0,
 *                          nRndRun=0):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ENTRANCE_RANDOM); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__25 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "daedalus/_maze.pyx":1586
 * 
 *     @staticmethod
 *     def _handle_save_retval(rv):             # <<<<<<<<<<<<<<
 *         '''
 *         Helper method to handle return values from the Save* methods.
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_8daedalus_5_maze_4Maze_66_handle_save_retval, NULL, __pyx_n_s_daedalus__maze); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_8daedalus_5_maze_Maze->tp_dict, __pyx_n_s_handle_save_retval, __pyx_t_2) < 0) __PYX_ERR(0, 1586, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_8daedalus_5_maze_Maze);

  /* "daedalus/_maze.pyx":1585
 *             raise MazeError('Could not resize the Maze.')
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def _handle_save_retval(rv):
 *         '''
 */
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_8daedalus_5_maze_Maze, __pyx_n_s_handle_save_retval); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_8daedalus_5_maze_Maze->tp_dict, __pyx_n_s_handle_save_retval, __pyx_t_1) < 0) __PYX_ERR(0, 1586, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_8daedalus_5_maze_Maze);

  /* "daedalus/_maze.pyx":1616
 *         self._yExit     = ms.yExit
 * 
 *     def save_bitmap(self, filename, kvOn=COLOR_WHITE, kvOff=COLOR_BLACK):             # <<<<<<<<<<<<<<
 *         '''
 *         Saves this Maze object as a bitmap to the given path.
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_COLOR_WHITE); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_k__26 = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_COLOR_BLACK); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_k__27 = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "daedalus/_maze.pyx":1
 * #!/usr/bin/env python             # <<<<<<<<<<<<<<
 * # -*- coding: utf-8 -*-
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init daedalus._maze", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init daedalus._maze");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* GetNameInClass */
static PyObject *__Pyx_GetGlobalNameAfterAttributeLookup(PyObject *name) {
    PyObject *result;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    result = __Pyx_PyObject_GetAttrStr(nmspace, name);
    if (!result) {
        result = __Pyx_GetGlobalNameAfterAttributeLookup(name);
    }
    return result;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_KV(KV value) {
    const KV neg_one = (KV) ((KV) 0 - (KV) 1), const_zero = (KV) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(KV) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(KV) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(KV) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(KV) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(KV) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(KV),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE unsigned long __Pyx_PyInt_As_unsigned_long(PyObject *x) {
    const unsigned long neg_one = (unsigned long) ((unsigned long) 0 - (unsigned long) 1), const_zero = (unsigned long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned long, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) >= 2 * PyLong_SHIFT) {
                            return (unsigned long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) >= 3 * PyLong_SHIFT) {
                            return (unsigned long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) >= 4 * PyLong_SHIFT) {
                            return (unsigned long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned long) (((unsigned long)-1)*(((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned long) ((((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned long) (((unsigned long)-1)*(((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned long) ((((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned long) (((unsigned long)-1)*(((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned long) ((((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned long) -1;
        }
    } else {
        unsigned long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned long) -1;
        val = __Pyx_PyInt_As_unsigned_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned long");
    return (unsigned long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned long");
    return (unsigned long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *x) {
    const unsigned int neg_one = (unsigned int) ((unsigned int) 0 - (unsigned int) 1), const_zero = (unsigned int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 2 * PyLong_SHIFT) {
                            return (unsigned int) (((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 3 * PyLong_SHIFT) {
                            return (unsigned int) (((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 4 * PyLong_SHIFT) {
                            return (unsigned int) (((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned int) ((((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned int) ((((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned int) ((((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned int) -1;
        }
    } else {
        unsigned int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned int) -1;
        val = __Pyx_PyInt_As_unsigned_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned int");
    return (unsigned int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned int");
    return (unsigned int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE KV __Pyx_PyInt_As_KV(PyObject *x) {
    const KV neg_one = (KV) ((KV) 0 - (KV) 1), const_zero = (KV) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(KV) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(KV, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (KV) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (KV) 0;
                case  1: __PYX_VERIFY_RETURN_INT(KV, digit, digits[0])
                case 2:
                    if (8 * sizeof(KV) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(KV, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(KV) >= 2 * PyLong_SHIFT) {
                            return (KV) (((((KV)digits[1]) << PyLong_SHIFT) | (KV)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(KV) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(KV, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(KV) >= 3 * PyLong_SHIFT) {
                            return (KV) (((((((KV)digits[2]) << PyLong_SHIFT) | (KV)digits[1]) << PyLong_SHIFT) | (KV)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(KV) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(KV, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(KV) >= 4 * PyLong_SHIFT) {
                            return (KV) (((((((((KV)digits[3]) << PyLong_SHIFT) | (KV)digits[2]) << PyLong_SHIFT) | (KV)digits[1]) << PyLong_SHIFT) | (KV)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (KV) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(KV) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(KV, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(KV) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(KV, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (KV) 0;
                case -1: __PYX_VERIFY_RETURN_INT(KV, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(KV,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(KV) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(KV, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(KV) - 1 > 2 * PyLong_SHIFT) {
                            return (KV) (((KV)-1)*(((((KV)digits[1]) << PyLong_SHIFT) | (KV)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(KV) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(KV, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(KV) - 1 > 2 * PyLong_SHIFT) {
                            return (KV) ((((((KV)digits[1]) << PyLong_SHIFT) | (KV)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(KV) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(KV, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(KV) - 1 > 3 * PyLong_SHIFT) {
                            return (KV) (((KV)-1)*(((((((KV)digits[2]) << PyLong_SHIFT) | (KV)digits[1]) << PyLong_SHIFT) | (KV)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(KV) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(KV, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(KV) - 1 > 3 * PyLong_SHIFT) {
                            return (KV) ((((((((KV)digits[2]) << PyLong_SHIFT) | (KV)digits[1]) << PyLong_SHIFT) | (KV)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(KV) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(KV, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(KV) - 1 > 4 * PyLong_SHIFT) {
                            return (KV) (((KV)-1)*(((((((((KV)digits[3]) << PyLong_SHIFT) | (KV)digits[2]) << PyLong_SHIFT) | (KV)digits[1]) << PyLong_SHIFT) | (KV)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(KV) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(KV, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(KV) - 1 > 4 * PyLong_SHIFT) {
                            return (KV) ((((((((((KV)digits[3]) << PyLong_SHIFT) | (KV)digits[2]) << PyLong_SHIFT) | (KV)digits[1]) << PyLong_SHIFT) | (KV)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(KV) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(KV, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(KV) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(KV, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            KV val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (KV) -1;
        }
    } else {
        KV val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (KV) -1;
        val = __Pyx_PyInt_As_KV(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to KV");
    return (KV) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to KV");
    return (KV) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
